#include "c.h"

#undef DISABLE

size_t rc_adjust_code(void *x, int by)
{
  Code p = x;
  RC_ADJUST_PREAMBLE;

#ifndef DISABLE
  RC_ADJUSTP(p->prev, by);
  RC_ADJUST(p->next, by);
  switch (p->kind)
    {
    case Blockbeg:
      RC_ADJUSTS(p->u.block.locals, by);
      RC_ADJUSTP(p->u.block.identifiers, by);
      RC_ADJUSTP(p->u.block.types, by);
      break;
    case Blockend:
      RC_ADJUSTS(p->u.begin, by);
      break;
    case Local:
      RC_ADJUSTS(p->u.var, by);
      break;
    case Address:
      RC_ADJUSTS(p->u.addr.sym, by);
      RC_ADJUSTS(p->u.addr.base, by);
      break;
    case Defpoint:
      RC_ADJUST(p->u.point.src.file, by);
      break;
    case Label: case Gen: case Jump:
      RC_ADJUSTP(p->u.forest, by);
      break;
    case Start:
      break;
    case Switch:
      RC_ADJUSTS(p->u.swtch.sym, by);
      RC_ADJUSTP(p->u.swtch.table, by);
      RC_ADJUSTS(p->u.swtch.deflab, by);
      RC_ADJUSTS(p->u.swtch.values, by);
      RC_ADJUSTS(p->u.swtch.labels, by);
      break;
    default:
      abort();
    }
#endif
  return sizeof *p;
}

size_t rc_adjust_symbol(void *x, int by)
{
  Symbol p = x;
  RC_ADJUST_PREAMBLE;

#ifndef DISABLE
  RC_ADJUSTP(p->name, by);
  RC_ADJUST(p->src.file, by);
  RC_ADJUSTP(p->up, by);
  RC_ADJUSTP(p->uses, by);

  /* Icky */
  if (p->type && p->type->op == FUNCTION) /* not really legal, but ... */
    {
      RC_ADJUST(p->u.f.pt.file, by);
      RC_ADJUSTP(p->u.f.callee, by);
    }
#ifdef sameregion
  else if (p->scope == LABELS)
    {
      RC_ADJUST(p->u.l.equatedto, by);
    }
  else if (p->scope == CONSTANTS)
    {
      RC_ADJUST(p->u.c.loc, by);
    }
  else if (p->temporary)
    {
      RC_ADJUST(p->u.t.cse, by);
    }
  else if (p->astruct)
    {
      RC_ADJUST(p->u.s.ftab, by);
      RC_ADJUST(p->u.s.flist, by);
    }
  else if (p->anenum)
    {
      RC_ADJUST(p->u.idlist, by);
    }
#endif

  RC_ADJUST(p->type, by);
  RC_ADJUSTP(p->x.name, by);
  RC_ADJUSTP(p->x.lastuse, by);
  RC_ADJUSTP(p->x.regnode, by);
#endif
  return sizeof *p;
}

size_t rc_adjust_tree(void *x, int by)
{
  Tree p = x;
  RC_ADJUST_PREAMBLE;

#ifndef DISABLE
  RC_ADJUSTP(p->type, by);
  RC_ADJUSTS(p->kids[0], by);
  RC_ADJUSTS(p->kids[1], by);
  RC_ADJUSTP(p->node, by);
  switch (generic(p->op))
    {
    case FIELD:
      RC_ADJUSTP(p->u.field, by);
      break;
    case CNST:
      break;
    default:
      RC_ADJUSTP(p->u.sym, by);
      break;
    }
#endif
  return sizeof *p;
}

size_t rc_adjust_type(void *x, int by)
{
  Type p = x;
  RC_ADJUST_PREAMBLE;

#ifndef DISABLE
  RC_ADJUSTS(p->type, by);
  switch (p->op)
    {
    case FUNCTION:
      RC_ADJUSTP(p->u.f.proto, by);
      break;
    default:
      RC_ADJUSTS(p->u.sym, by);
      break;
    }
#endif
  return sizeof *p;
}

/* Some copy functions */
void copy_coord(Coordinate *to, Coordinate *from)
{
  to->x = from->x;
  to->y = from->y;
  to->file = from->file;
}

void copy_ptrarray(void @@to, void @@from, int n)
{
  while (n--) *to++ = *from++;
}

void copy_symbol(Symbol to, Symbol from)
{
  /* update rcs */
  to->name = from->name;
  copy_coord(&to->src, &from->src);
  to->up = from->up;
  to->type = from->type;

  /* Clear union in to */
  if (to->type && to->type->op == FUNCTION) /* not really legal, but ... */
    {
      to->u.f.pt.file = 0;
      to->u.f.callee = 0;
    }
#ifdef sameregion
  else if (to->scope == LABELS)
    to->u.l.equatedto = 0;
  else if (to->scope == CONSTANTS)
    to->u.c.loc = 0;
  else if (to->temporary)
    to->u.t.cse = 0;
  else if (to->astruct)
    {
      to->u.s.ftab = 0;
      to->u.s.flist = 0;
    }
  else if (to->anenum)
    to->u.idlist = 0;
#endif
  memset(&to->u, 0, sizeof to->u);

  /* Icky */
  if (from->type && from->type->op == FUNCTION)
    {
      to->u.f.pt.file = from->u.f.pt.file;
      to->u.f.callee = from->u.f.callee;
    }
#ifdef sameregion
 else if (from->scope == LABELS)
    to->u.l.equatedto = from->u.l.equatedto;
  else if (from->scope == CONSTANTS)
    to->u.c.loc = from->u.c.loc;
  else if (from->temporary)
    to->u.t.cse = from->u.t.cse;
  else if (from->astruct)
    {
      to->u.s.ftab = from->u.s.ftab;
      to->u.s.flist = from->u.s.flist;
    }
  else if (from->anenum)
    to->u.idlist = from->u.idlist;
#endif

  to->x.name = from->x.name;
  to->x.lastuse = from->x.lastuse;
  to->x.regnode = from->x.regnode;

  /* Then copy everything over again - might be better to copy fields ? */
  memcpy((void *)to, (void *)from, sizeof *to);
}

void copy_node(Node to, Node from)
{
  to->syms[0] = from->syms[0];
  to->syms[1] = from->syms[1];
  to->syms[2] = from->syms[2];
#ifdef sameregion
  to->kids[0] = from->kids[0];
  to->kids[1] = from->kids[1];
  to->link = from->link;
  to->x.kids[0] = from->x.kids[0];
  to->x.kids[1] = from->x.kids[1];
  to->x.kids[2] = from->x.kids[2];
  to->x.prev = from->x.prev;
  to->x.next = from->x.next;
  to->x.prevuse = from->x.prevuse;
  to->x.state = from->x.state;
#endif
  memcpy((void *)&to, (void *)&from, sizeof to);
}

void copy_tree(Tree to, Tree from)
{
  to->type = from->type;
#ifdef sameregion
  to->kids[0] = from->kids[0];
  to->kids[1] = from->kids[1];
#endif
  to->node = from->node;

  /* clear union */
  switch (generic(to->op))
    {
    case FIELD:
      to->u.field = 0;
      break;
    case CNST:
      break;
    default:
      to->u.sym = 0;
      break;
    }
  memset(&to->u, 0, sizeof to->u);

  switch (generic(from->op))
    {
    case FIELD:
      to->u.field = from->u.field;
      break;
    case CNST:
      break;
    default:
      to->u.sym = from->u.sym;
      break;
    }
  memcpy((void *)&to->u, (void *)&from->u, sizeof to->u);
}

void deallocate(unsigned a) deletes
{
#if 0
#define NS 100
  static int scnt;

  if (a == STMT)
    if (++scnt <= NS) return;
    else scnt = 0;
  if (a == FUNC)
    {
      scnt = NS;
      deallocate(STMT);
    }
#endif
#if 0
  if (a == STMT && ((int *)heap[STMT])[1] < 100)
    return;
  if (a == FUNC)
    {
      deleteregion_ptr(&heap[STMT]);
      heap[STMT] = newheap();
    }
#endif

  switch (a)
    {
    case FUNC:
      clear_kids();
      clear_dagbuckets();
      deleteregion_ptr(&heap[STMT]);
      deleteregion_ptr(&heap[FUNC]);
      heap[FUNC] = newsubregion(heap[PERM]);
      heap[STMT] = newsubregion(heap[FUNC]);
      break;
    case STMT:
      deleteregion_ptr(&heap[STMT]);
      heap[STMT] = newsubregion(heap[FUNC]);
      break;
    default:
      deleteregion_ptr(&heap[a]);
      heap[a] = newregion();
    }
}
