#include "c.h"
#include <float.h>
#include <errno.h>

static char rcsid[] = "$Id: nlex.c,v 1.1.1.1 2001/06/28 19:46:59 dgay Exp $";

#define MAXTOKEN 32

enum { BLANK=01,  NEWLINE=02, LETTER=04,
       DIGIT=010, HEX=020,    OTHER=040 };

static unsigned char map[256] = { /* 000 nul */   0,
				   /* 001 soh */   0,
				   /* 002 stx */   0,
				   /* 003 etx */   0,
				   /* 004 eot */   0,
				   /* 005 enq */   0,
				   /* 006 ack */   0,
				   /* 007 bel */   0,
				   /* 010 bs  */   0,
				   /* 011 ht  */   BLANK,
				   /* 012 nl  */   NEWLINE,
				   /* 013 vt  */   BLANK,
				   /* 014 ff  */   BLANK,
				   /* 015 cr  */   0,
				   /* 016 so  */   0,
				   /* 017 si  */   0,
				   /* 020 dle */   0,
				   /* 021 dc1 */   0,
				   /* 022 dc2 */   0,
				   /* 023 dc3 */   0,
				   /* 024 dc4 */   0,
				   /* 025 nak */   0,
				   /* 026 syn */   0,
				   /* 027 etb */   0,
				   /* 030 can */   0,
				   /* 031 em  */   0,
				   /* 032 sub */   0,
				   /* 033 esc */   0,
				   /* 034 fs  */   0,
				   /* 035 gs  */   0,
				   /* 036 rs  */   0,
				   /* 037 us  */   0,
				   /* 040 sp  */   BLANK,
				   /* 041 !   */   OTHER,
				   /* 042 "   */   OTHER,
				   /* 043 #   */   OTHER,
				   /* 044 $   */   0,
				   /* 045 %   */   OTHER,
				   /* 046 &   */   OTHER,
				   /* 047 '   */   OTHER,
				   /* 050 (   */   OTHER,
				   /* 051 )   */   OTHER,
				   /* 052 *   */   OTHER,
				   /* 053 +   */   OTHER,
				   /* 054 ,   */   OTHER,
				   /* 055 -   */   OTHER,
				   /* 056 .   */   OTHER,
				   /* 057 /   */   OTHER,
				   /* 060 0   */   DIGIT,
				   /* 061 1   */   DIGIT,
				   /* 062 2   */   DIGIT,
				   /* 063 3   */   DIGIT,
				   /* 064 4   */   DIGIT,
				   /* 065 5   */   DIGIT,
				   /* 066 6   */   DIGIT,
				   /* 067 7   */   DIGIT,
				   /* 070 8   */   DIGIT,
				   /* 071 9   */   DIGIT,
				   /* 072 :   */   OTHER,
				   /* 073 ;   */   OTHER,
				   /* 074 <   */   OTHER,
				   /* 075 =   */   OTHER,
				   /* 076 >   */   OTHER,
				   /* 077 ?   */   OTHER,
				   /* 100 @   */   OTHER,
				   /* 101 A   */   LETTER|HEX,
				   /* 102 B   */   LETTER|HEX,
				   /* 103 C   */   LETTER|HEX,
				   /* 104 D   */   LETTER|HEX,
				   /* 105 E   */   LETTER|HEX,
				   /* 106 F   */   LETTER|HEX,
				   /* 107 G   */   LETTER,
				   /* 110 H   */   LETTER,
				   /* 111 I   */   LETTER,
				   /* 112 J   */   LETTER,
				   /* 113 K   */   LETTER,
				   /* 114 L   */   LETTER,
				   /* 115 M   */   LETTER,
				   /* 116 N   */   LETTER,
				   /* 117 O   */   LETTER,
				   /* 120 P   */   LETTER,
				   /* 121 Q   */   LETTER,
				   /* 122 R   */   LETTER,
				   /* 123 S   */   LETTER,
				   /* 124 T   */   LETTER,
				   /* 125 U   */   LETTER,
				   /* 126 V   */   LETTER,
				   /* 127 W   */   LETTER,
				   /* 130 X   */   LETTER,
				   /* 131 Y   */   LETTER,
				   /* 132 Z   */   LETTER,
				   /* 133 [   */   OTHER,
				   /* 134 \   */   OTHER,
				   /* 135 ]   */   OTHER,
				   /* 136 ^   */   OTHER,
				   /* 137 _   */   LETTER,
				   /* 140 `   */   0,
				   /* 141 a   */   LETTER|HEX,
				   /* 142 b   */   LETTER|HEX,
				   /* 143 c   */   LETTER|HEX,
				   /* 144 d   */   LETTER|HEX,
				   /* 145 e   */   LETTER|HEX,
				   /* 146 f   */   LETTER|HEX,
				   /* 147 g   */   LETTER,
				   /* 150 h   */   LETTER,
				   /* 151 i   */   LETTER,
				   /* 152 j   */   LETTER,
				   /* 153 k   */   LETTER,
				   /* 154 l   */   LETTER,
				   /* 155 m   */   LETTER,
				   /* 156 n   */   LETTER,
				   /* 157 o   */   LETTER,
				   /* 160 p   */   LETTER,
				   /* 161 q   */   LETTER,
				   /* 162 r   */   LETTER,
				   /* 163 s   */   LETTER,
				   /* 164 t   */   LETTER,
				   /* 165 u   */   LETTER,
				   /* 166 v   */   LETTER,
				   /* 167 w   */   LETTER,
				   /* 170 x   */   LETTER,
				   /* 171 y   */   LETTER,
				   /* 172 z   */   LETTER,
				   /* 173 {   */   OTHER,
				   /* 174 |   */   OTHER,
				   /* 175 }   */   OTHER,
				   /* 176 ~   */   OTHER, };
static struct symbol *ptval;
Coordinate src;         /* current source coordinate */
int t;
const char @token;      /* current token */
Symbol tsym;            /* symbol table entry for current token */

static int backslash(int q);
static Symbol fcon(void);
static Symbol icon(unsigned long, int, int);
static void ppnumber(char *);

int gettok(void) {
  if (!ptval)
    NEW0(ptval, PERM);

        for (;;) {
                register unsigned char *startrcp;
                while (map[*cp]&BLANK)
		        cp++;
                if (limit - cp < MAXTOKEN)
                        fillbuf();
                src.file = file;
                src.x = (char *)cp - line;
                src.y = lineno;
                switch (*cp++) {
                case '/': if (*cp == '*') {
			          int c = 0;
			          for (cp++; *cp != '/' || c != '*'; )
			                  if (map[*cp]&NEWLINE) {
			                          if (cp < limit)
			                                  c = *cp;
			                          cp = cp + 1;
			                          nextline();
			                          if (cp == limit)
			                                  break;
			                  } else
			                          c = *cp++;
			          if (cp < limit)
			                  cp++;
			          else
			                  error("unclosed comment\n");
			          continue;
			  }
			  return '/';
		case 'L': if (*cp == '\'') {
			          int t = gettok();
			          assert(t == ICON);
			          src.x--;
			          ptval->type = unsignedchar;
			          ptval->u.c.v.u = (unsigned char)ptval->u.c.v.i;
			          return t;
			  }
			  if (*cp != '"')
			          goto id;
			  cp = cp + 1;
			  goto scon;
		case '<':
		        if (*cp == '=') return cp++, LEQ;
		        if (*cp == '<') return cp++, LSHIFT;
		        return '<';
		case '>':
		        if (*cp == '=') return cp++, GEQ;
		        if (*cp == '>') return cp++, RSHIFT;
		        return '>';
		case '-':
		        if (*cp == '>') return cp++, DEREF;
		        if (*cp == '-') return cp++, DECR;
		        return '-';
		case '=': return *cp == '=' ? cp++, EQL    : '=';
		case '!': return *cp == '=' ? cp++, NEQ    : '!';
		case '|': return *cp == '|' ? cp++, OROR   : '|';
		case '&': return *cp == '&' ? cp++, ANDAND : '&';
		case '+': return *cp == '+' ? cp++, INCR   : '+';
		case ';': case ',': case ':':
		case '@': case '*': case '~': case '%': case '^': case '?':
		case '[': case ']': case '{': case '}': case '(': case ')': 
		        return cp[-1];
		case '\n': case '\v': case '\r': case '\f':
		        nextline();
		        if (cp == limit) {
		                tsym = NULL;
		                return EOI;
		        }
		        continue;

		case 'i':
		        if (cp[0] == 'f'
		        && !(map[cp[1]]&(DIGIT|LETTER))) {
		                cp = cp + 1;
		                return IF;
		        }
		        if (cp[0] == 'n'
		        &&  cp[1] == 't'
		        && !(map[cp[2]]&(DIGIT|LETTER))) {
		                cp = cp + 2;
		                tsym = inttype->u.sym;
		                return INT;
		        }
		        goto id;
		case 'j': case 'k': case 'm': case 'n': case 'o':
		case 'p': case 'q': case 'x': case 'y': case 'z':
		case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
		case 'G': case 'H': case 'I': case 'J': case 'K':
		case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
		case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
		case 'Y': case 'Z':
		id:
		        if (limit - cp < MAXLINE) {
			        cp--;
			        fillbuf();
			        ++cp;
			}
			startrcp = cp - 1;
			while (map[*cp]&(DIGIT|LETTER))
		                cp++;
			token = stringn((char *)startrcp, cp - startrcp);
		        tsym = lookup(token, identifiers);
		        return ID;
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9': {
		        unsigned long n = 0;
		        if (limit - cp < MAXLINE) {
			        cp--;
			        fillbuf();
			        ++cp;
			}
		        token = (char *)cp - 1;
		        if (*token == '0' && (*cp == 'x' || *cp == 'X')) {
		                int d, overflow = 0;
				while (*++cp) {
				        if (map[*cp]&DIGIT)
				                d = *cp - '0';
				        else if (*cp >= 'a' && *cp <= 'f')
				                d = *cp - 'a' + 10;
				        else if (*cp >= 'A' && *cp <= 'F')
				                d = *cp - 'A' + 10;
				        else
				                break;
				        if (n&~(~0UL >> 4))
				                overflow = 1;
				        else
				                n = (n<<4) + d;
				}
				if ((char *)cp - token <= 2)
				        error("invalid hexadecimal constant `%S'\n", token, (char *)cp-token);
				tsym = icon(n, overflow, 16);
		        } else if (*token == '0') {
		                int err = 0, overflow = 0;
				for ( ; map[*cp]&DIGIT; cp++) {
				        if (*cp == '8' || *cp == '9')
				                err = 1;
				        if (n&~(~0UL >> 3))
				                overflow = 1;
				        else
				                n = (n<<3) + (*cp - '0');
				}
				if (*cp == '.' || *cp == 'e' || *cp == 'E') {
				        tsym = fcon();
				        return FCON;
				}
				tsym = icon(n, overflow, 8);
				if (err)
				        error("invalid octal constant `%S'\n", token, (char*)cp-token);
		        } else {
		                int overflow = 0;
				for (n = *token - '0'; map[*cp]&DIGIT; ) {
				        int d = *cp++ - '0';
				        if (n > (ULONG_MAX - d)/10)
				                overflow = 1;
				        else
				                n = 10*n + d;
				}
				if (*cp == '.' || *cp == 'e' || *cp == 'E') {
				        tsym = fcon();
				        return FCON;
				}
				tsym = icon(n, overflow, 10);
		        }
		        return ICON;
		}
		case '.':
		        if (cp[0] == '.' && cp[1] == '.') {
		                cp += 2;
		                return ELLIPSIS;
		        }
		        if ((map[*cp]&DIGIT) == 0)
		                return '.';
		        if (limit - cp < MAXLINE) {
			        cp--;
			        fillbuf();
			        ++cp;
			}
		        cp--;
		        token = (char *)cp;
		        tsym = fcon();
		        return FCON;
		scon:
		case '\'': case '"': {
		        static char cbuf[BUFSIZE+1];
		        char *s = cbuf;
		        int nbad = 0;
		        *s++ = *--cp;
		        do {
		                cp++;
		                while (*cp != cbuf[0]) {
				        int c;
				        if (map[*cp]&NEWLINE) {
				                if (cp < limit)
				                        break;
				                cp++;
				                nextline();
				                if (cp == limit)
				                        break;
				                continue;
				        }
				        c = *cp++;
				        if (c == '\\') {
				                if (map[*cp]&NEWLINE) {
				                        if (cp < limit)
				                                break;
				                        cp++;
				                        nextline();
				                }
				                if (limit - cp < MAXTOKEN)
				                        fillbuf();
				                c = backslash(cbuf[0]);
				        } else if (map[c] == 0)
				                nbad++;
				        if (s < &cbuf[sizeof cbuf] - 1)
				                *s++ = c;
				}
		                if (*cp == cbuf[0])
		                        cp++;
		                else
		                        error("missing %c\n", cbuf[0]);
		        } while (cbuf[0] == '"' && getchr() == '"');
		        *s++ = 0;
		        if (s >= &cbuf[sizeof cbuf])
		                error("%s literal too long\n",
		                        cbuf[0] == '"' ? "string" : "character");
		        if (Aflag >= 2 && cbuf[0] == '"' && s - cbuf - 1 > 509)
			        warning("more than 509 characters in a string literal\n");
			if (Aflag >= 2 && nbad)
			        warning("%s literal contains non-portable characters\n",
			                cbuf[0] == '"' ? "string" : "character");
		        token = cbuf;
			tsym = ptval;
			if (cbuf[0] == '"') {
			        ptval->type = array(chartype, s - cbuf - 1, 0);
			        ptval->u.c.v.p = cbuf + 1;
			        return SCON;
			} else {
			        if (s - cbuf > 3)
			                warning("excess characters in multibyte character literal `%S' ignored\n", token, (char*)cp-token);

			        else if (s - cbuf <= 2)
			                error("missing '\n");
			        ptval->type = inttype;
			        ptval->u.c.v.i = cbuf[1];
			        return ICON;
			}
		}
		case 'a':
		        if (cp[0] == 'u'
		        &&  cp[1] == 't'
		        &&  cp[2] == 'o'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                return AUTO;
		        }
		        goto id;
		case 'b':
		        if (cp[0] == 'r'
		        &&  cp[1] == 'e'
		        &&  cp[2] == 'a'
		        &&  cp[3] == 'k'
		        && !(map[cp[4]]&(DIGIT|LETTER))) {
		                cp = cp + 4;
		                return BREAK;
		        }
		        goto id;
		case 'c':
		        if (cp[0] == 'a'
		        &&  cp[1] == 's'
		        &&  cp[2] == 'e'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                return CASE;
		        }
		        if (cp[0] == 'h'
		        &&  cp[1] == 'a'
		        &&  cp[2] == 'r'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                tsym = chartype->u.sym;
		                return CHAR;
		        }
		        if (cp[0] == 'o'
		        &&  cp[1] == 'n'
		        &&  cp[2] == 's'
		        &&  cp[3] == 't'
		        && !(map[cp[4]]&(DIGIT|LETTER))) {
		                cp = cp + 4;
		                return CONST;
		        }
		        if (cp[0] == 'o'
		        &&  cp[1] == 'n'
		        &&  cp[2] == 't'
		        &&  cp[3] == 'i'
		        &&  cp[4] == 'n'
		        &&  cp[5] == 'u'
		        &&  cp[6] == 'e'
		        && !(map[cp[7]]&(DIGIT|LETTER))) {
		                cp = cp + 7;
		                return CONTINUE;
		        }
		        goto id;
		case 'd':
		        if (cp[0] == 'e'
		        &&  cp[1] == 'f'
		        &&  cp[2] == 'a'
		        &&  cp[3] == 'u'
		        &&  cp[4] == 'l'
		        &&  cp[5] == 't'
		        && !(map[cp[6]]&(DIGIT|LETTER))) {
		                cp = cp + 6;
		                return DEFAULT;
		        }
		        if (cp[0] == 'o'
		        &&  cp[1] == 'u'
		        &&  cp[2] == 'b'
		        &&  cp[3] == 'l'
		        &&  cp[4] == 'e'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                tsym = doubletype->u.sym;
		                return DOUBLE;
		        }
		        if (cp[0] == 'o'
		        && !(map[cp[1]]&(DIGIT|LETTER))) {
		                cp = cp + 1;
		                return DO;
		        }
		        goto id;
		case 'e':
		        if (cp[0] == 'l'
		        &&  cp[1] == 's'
		        &&  cp[2] == 'e'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                return ELSE;
		        }
		        if (cp[0] == 'n'
		        &&  cp[1] == 'u'
		        &&  cp[2] == 'm'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                return ENUM;
		        }
		        if (cp[0] == 'x'
		        &&  cp[1] == 't'
		        &&  cp[2] == 'e'
		        &&  cp[3] == 'r'
		        &&  cp[4] == 'n'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return EXTERN;
		        }
		        goto id;
		case 'f':
		        if (cp[0] == 'l'
		        &&  cp[1] == 'o'
		        &&  cp[2] == 'a'
		        &&  cp[3] == 't'
		        && !(map[cp[4]]&(DIGIT|LETTER))) {
		                cp = cp + 4;
		                tsym = floattype->u.sym;
		                return FLOAT;
		        }
		        if (cp[0] == 'o'
		        &&  cp[1] == 'r'
		        && !(map[cp[2]]&(DIGIT|LETTER))) {
		                cp = cp + 2;
		                return FOR;
		        }
		        goto id;
		case 'g':
		        if (cp[0] == 'o'
		        &&  cp[1] == 't'
		        &&  cp[2] == 'o'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                return GOTO;
		        }
		        goto id;
		case 'h':
		        if (cp[0] == 's'
		        &&  cp[1] == 't'
		        &&  cp[2] == 'r'
		        &&  cp[3] == 'u'
		        &&  cp[4] == 'c'
		        &&  cp[5] == 't'
		        && !(map[cp[6]]&(DIGIT|LETTER))) {
		                cp = cp + 6;
		                return HSTRUCT;
		        }
		        goto id;
		case 'l':
		        if (cp[0] == 'o'
		        &&  cp[1] == 'n'
		        &&  cp[2] == 'g'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                tsym = longtype->u.sym;
		                return LONG;
		        }
		        goto id;
		case 'r':
		        if (cp[0] == 'e'
		        &&  cp[1] == 'g'
		        &&  cp[2] == 'i'
		        &&  cp[3] == 's'
		        &&  cp[4] == 't'
		        &&  cp[5] == 'e'
		        &&  cp[6] == 'r'
		        && !(map[cp[7]]&(DIGIT|LETTER))) {
		                cp = cp + 7;
		                return REGISTER;
		        }
		        if (cp[0] == 'e'
		        &&  cp[1] == 't'
		        &&  cp[2] == 'u'
		        &&  cp[3] == 'r'
		        &&  cp[4] == 'n'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return RETURN;
		        }
		        goto id;
		case 's':
		        if (cp[0] == 'h'
		        &&  cp[1] == 'o'
		        &&  cp[2] == 'r'
		        &&  cp[3] == 't'
		        && !(map[cp[4]]&(DIGIT|LETTER))) {
		                cp = cp + 4;
		                tsym = shorttype->u.sym;
		                return SHORT;
		        }
		        if (cp[0] == 'i'
		        &&  cp[1] == 'g'
		        &&  cp[2] == 'n'
		        &&  cp[3] == 'e'
		        &&  cp[4] == 'd'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return SIGNED;
		        }
		        if (cp[0] == 'i'
		        &&  cp[1] == 'z'
		        &&  cp[2] == 'e'
		        &&  cp[3] == 'o'
		        &&  cp[4] == 'f'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return SIZEOF;
		        }
		        if (cp[0] == 't'
		        &&  cp[1] == 'a'
		        &&  cp[2] == 't'
		        &&  cp[3] == 'i'
		        &&  cp[4] == 'c'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return STATIC;
		        }
		        if (cp[0] == 't'
		        &&  cp[1] == 'r'
		        &&  cp[2] == 'u'
		        &&  cp[3] == 'c'
		        &&  cp[4] == 't'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return STRUCT;
		        }
		        if (cp[0] == 'w'
		        &&  cp[1] == 'i'
		        &&  cp[2] == 't'
		        &&  cp[3] == 'c'
		        &&  cp[4] == 'h'
		        && !(map[cp[5]]&(DIGIT|LETTER))) {
		                cp = cp + 5;
		                return SWITCH;
		        }
		        goto id;
		case 't':
		        if (cp[0] == 'y'
		        &&  cp[1] == 'p'
		        &&  cp[2] == 'e'
		        &&  cp[3] == 'd'
		        &&  cp[4] == 'e'
		        &&  cp[5] == 'f'
		        && !(map[cp[6]]&(DIGIT|LETTER))) {
		                cp = cp + 6;
		                return TYPEDEF;
		        }
		        goto id;
		case 'u':
		        if (cp[0] == 'n'
		        &&  cp[1] == 'i'
		        &&  cp[2] == 'o'
		        &&  cp[3] == 'n'
		        && !(map[cp[4]]&(DIGIT|LETTER))) {
		                cp = cp + 4;
		                return UNION;
		        }
		        if (cp[0] == 'n'
		        &&  cp[1] == 's'
		        &&  cp[2] == 'i'
		        &&  cp[3] == 'g'
		        &&  cp[4] == 'n'
		        &&  cp[5] == 'e'
		        &&  cp[6] == 'd'
		        && !(map[cp[7]]&(DIGIT|LETTER))) {
		                cp = cp + 7;
		                return UNSIGNED;
		        }
		        goto id;
		case 'v':
		        if (cp[0] == 'o'
		        &&  cp[1] == 'i'
		        &&  cp[2] == 'd'
		        && !(map[cp[3]]&(DIGIT|LETTER))) {
		                cp = cp + 3;
		                tsym = voidtype->u.sym;
		                return VOID;
		        }
		        if (cp[0] == 'o'
		        &&  cp[1] == 'l'
		        &&  cp[2] == 'a'
		        &&  cp[3] == 't'
		        &&  cp[4] == 'i'
		        &&  cp[5] == 'l'
		        &&  cp[6] == 'e'
		        && !(map[cp[7]]&(DIGIT|LETTER))) {
		                cp = cp + 7;
		                return VOLATILE;
		        }
		        goto id;
		case 'w':
		        if (cp[0] == 'h'
		        &&  cp[1] == 'i'
		        &&  cp[2] == 'l'
		        &&  cp[3] == 'e'
		        && !(map[cp[4]]&(DIGIT|LETTER))) {
		                cp = cp + 4;
		                return WHILE;
		        }
		        goto id;
		case '_':
		        if (cp[0] == '_'
		        &&  cp[1] == 't'
		        &&  cp[2] == 'y'
		        &&  cp[3] == 'p'
		        &&  cp[4] == 'e'
		        &&  cp[5] == 'c'
		        &&  cp[6] == 'o'
		        &&  cp[7] == 'd'
		        &&  cp[8] == 'e'
		        && !(map[cp[9]]&(DIGIT|LETTER))) {
		                cp = cp + 9;
		                return TYPECODE;
		        }
		        if (cp[0] == '_'
		        &&  cp[1] == 'f'
		        &&  cp[2] == 'i'
		        &&  cp[3] == 'r'
		        &&  cp[4] == 's'
		        &&  cp[5] == 't'
		        &&  cp[6] == 'a'
		        &&  cp[7] == 'r'
		        &&  cp[8] == 'g'
		        && !(map[cp[9]]&(DIGIT|LETTER))) {
		                cp = cp + 9;
		                return FIRSTARG;
		        }
		        goto id;
                default:
                        if ((map[cp[-1]]&BLANK) == 0)
                                if (cp[-1] < ' ' || cp[-1] >= 0177)
				        error("illegal character `\\0%o'\n", cp[-1]);
				else
				        error("illegal character `%c'\n", cp[-1]);
                }
        }
}
static Symbol icon(unsigned long n, int overflow, int base) {
        if ((*cp=='u'||*cp=='U') && (cp[1]=='l'||cp[1]=='L')
        ||  (*cp=='l'||*cp=='L') && (cp[1]=='u'||cp[1]=='U')) {
                ptval->type = unsignedlong;
                cp += 2;
        } else if (*cp == 'u' || *cp == 'U') {
                if (overflow || n > unsignedtype->u.sym->u.limits.max.i)
                        ptval->type = unsignedlong;
                else
                        ptval->type = unsignedtype;
                cp += 1;
        } else if (*cp == 'l' || *cp == 'L') {
                if (overflow || n > longtype->u.sym->u.limits.max.i)
                        ptval->type = unsignedlong;
                else
                        ptval->type = longtype;
                cp += 1;
        } else if (overflow || n > longtype->u.sym->u.limits.max.i)
                ptval->type = unsignedlong;
        else if (n > inttype->u.sym->u.limits.max.i)
                ptval->type = longtype;
        else if (base != 10 && n > inttype->u.sym->u.limits.max.i)
                ptval->type = unsignedtype;
        else
                ptval->type = inttype;
        switch (ptval->type->op) {
	case INT:
	        if (overflow || n > ptval->type->u.sym->u.limits.max.i) {
	                warning("overflow in constant `%S'\n", token,
	                        (char*)cp - token);
	                ptval->u.c.v.i = ptval->type->u.sym->u.limits.max.i;
	        } else
	                ptval->u.c.v.i = n;
	        break;
	case UNSIGNED:
	        if (overflow || n > ptval->type->u.sym->u.limits.max.u) {
	                warning("overflow in constant `%S'\n", token,
	                        (char*)cp - token);
	                ptval->u.c.v.u = ptval->type->u.sym->u.limits.max.u;
	        } else
	                ptval->u.c.v.u = n;
	        break;
	default: assert(0);
	}
        ppnumber("integer");
        return ptval;
}
static void ppnumber(char *which) {
        unsigned char *rcp = cp--;

        for ( ; (map[*cp]&(DIGIT|LETTER)) || *cp == '.'; cp++)
                if ((cp[0] == 'E' || cp[0] == 'e')
                &&  (cp[1] == '-' || cp[1] == '+'))
                        cp++;
        if (cp > rcp)
                error("`%S' is a preprocessing number but an invalid %s constant\n", token,

                        (char*)cp-token, which);
}
static Symbol fcon(void) {
        if (*cp == '.')
	        do
		        cp++;
		while (map[*cp]&DIGIT);
	if (*cp == 'e' || *cp == 'E') {
	        if (*++cp == '-' || *cp == '+')
	                cp++;
	        if (map[*cp]&DIGIT)
	                do
			        cp++;
			while (map[*cp]&DIGIT);
	        else
	                error("invalid floating constant `%S'\n", token,
	                        (char*)cp - token);
	}

        errno = 0;
        ptval->u.c.v.d = strtod(token, NULL);
        if (errno == ERANGE)
                warning("overflow in floating constant `%S'\n", token,
		        (char*)cp - token);
        if (*cp == 'f' || *cp == 'F') {
	        ++cp;
	        if (ptval->u.c.v.d > floattype->u.sym->u.limits.max.d)
	                warning("overflow in floating constant `%S'\n", token,
			        (char*)cp - token);
	        ptval->type = floattype;
	} else if (*cp == 'l' || *cp == 'L') {
	        cp++;
	        ptval->type = longdouble;
	} else {
	        if (ptval->u.c.v.d > doubletype->u.sym->u.limits.max.d)
	                warning("overflow in floating constant `%S'\n", token,
			        (char*)cp - token);
	        ptval->type = doubletype;
	}
        ppnumber("floating");
        return ptval;
}

int getchr(void) {
        for (;;) {
                while (map[*cp]&BLANK)
                        cp++;
                if (!(map[*cp]&NEWLINE))
                        return *cp;
                cp++;
                nextline();
                if (cp == limit)
                        return EOI;
        }
}
static int backslash(int q) {
        int c;

        switch (*cp++) {
        case 'a': return 7;
        case 'b': return '\b';
        case 'f': return '\f';
        case 'n': return '\n';
        case 'r': return '\r';
        case 't': return '\t';
        case 'v': return '\v';
        case '\'': case '"': case '\\': case '\?': break;
        case 'x': {
                int overflow = 0;
                if ((map[*cp]&(DIGIT|HEX)) == 0) {
                        if (*cp < ' ' || *cp == 0177)
                                error("ill-formed hexadecimal escape sequence\n");
                        else
                                error("ill-formed hexadecimal escape sequence `\\x%c'\n", *cp);
                        if (*cp != q)
                                cp++;
                        return 0;
                }
                for (c = 0; map[*cp]&(DIGIT|HEX); cp++) {
                        if (c&~(~0UL >> 4))
                                overflow++;
                        if (map[*cp]&DIGIT)
                                c = (c<<4) + *cp - '0';
                        else
                                c = (c<<4) + (*cp&~040) - 'A' + 10;
                }
                if (c&~0377 || overflow)
                        warning("overflow in hexadecimal escape sequence\n");
                return c&0377;
                }
        case '0': case '1': case '2': case '3':
        case '4': case '5': case '6': case '7':
                c = *(cp-1) - '0';
                if (*cp >= '0' && *cp <= '7') {
                        c = (c<<3) + *cp++ - '0';
                        if (*cp >= '0' && *cp <= '7')
                                c = (c<<3) + *cp++ - '0';
                }
                if (c&~0377)
                        warning("overflow in octal escape sequence\n");
                return c&0377;
        default:
                if (cp[-1] < ' ' || cp[-1] >= 0177)
                        warning("unrecognized character escape sequence\n");
                else
                        warning("unrecognized character escape sequence `\\%c'\n", cp[-1]);
        }
        return cp[-1];
}
