static void
PREFIXquery ()
{
  static GParamDef args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image (unused)" },
    { PARAM_DRAWABLE, "drawable", "Input drawable" },
    { PARAM_INT32, "dummy", "dummy" }
  };
  static GParamDef *return_vals = 0;
  static int nargs = sizeof (args) / sizeof (args[0]);
  static int nreturn_vals = 0;

  gimp_install_procedure ("plug_in_gfig",
			  "Create Geometrical shapes with the Gimp",
			  "More here later",
			  "Andy Thomas",
			  "Andy Thomas",
			  "1997",
			  "<Image>/Filters/Render/Gfig",
			  "RGB*, GRAY*",
			  PROC_PLUG_IN,
			  nargs, nreturn_vals,
			  args, return_vals);
}

static void
PREFIXrun    (gchar    *name,
	gint      nparams,
	GParam   *param,
	gint     *nreturn_vals,
	GParam  **return_vals)
{
  GParam * values =((GParam *) g_malloc ((unsigned) sizeof (GParam) * (1)));
  GDrawable *drawable;
  GRunModeType run_mode;
  GStatusType status = STATUS_SUCCESS;

  int           pwidth, pheight;



  run_mode = param[0].data.d_int32;
  gfig_image = param[1].data.d_image;
  gfig_drawable = param[2].data.d_drawable;

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = PARAM_STATUS;
  values[0].data.d_status = status;

  gfig_select_drawable =
    drawable =
    gimp_drawable_get (param[2].data.d_drawable);

  tile_width  = gimp_tile_width();
  tile_height = gimp_tile_height();


  gfig_clear_selection(gfig_image);

  gimp_drawable_mask_bounds(drawable->id, &sel_x1, &sel_y1, &sel_x2, &sel_y2);

  sel_width  = sel_x2 - sel_x1;
  sel_height = sel_y2 - sel_y1;



  if (sel_width > sel_height) {
    pwidth  =(((sel_width) < (650)) ? (sel_width) : (650));
    pheight = sel_height * pwidth / sel_width;
  } else {
    pheight =(((sel_height) < (650)) ? (sel_height) : (650));
    pwidth  = sel_width * pheight / sel_height;
  }

  preview_width  =(((pwidth) > (2)) ? (pwidth) : (2));
  preview_height =(((pheight) > (2)) ? (pheight) : (2));

  org_scale_x_factor = scale_x_factor = (gdouble)sel_width/(gdouble)preview_width;
  org_scale_y_factor = scale_y_factor = (gdouble)sel_height/(gdouble)preview_height;

  switch (run_mode)
    {
    case RUN_INTERACTIVE:

      if (!gfig_dialog())
	{
	  gimp_drawable_detach (drawable);
	  return;
	}
      break;

    case RUN_NONINTERACTIVE:
      status = STATUS_CALLING_ERROR;
      break;

    case RUN_WITH_LAST_VALS:

      break;

    default:
      break;
    }

  if (gimp_drawable_color (drawable->id) || gimp_drawable_gray (drawable->id))
    {


      gimp_tile_cache_ntiles((drawable->width + gimp_tile_width() - 1) / gimp_tile_width());

      do_gfig();

      if (run_mode != RUN_NONINTERACTIVE)
	gimp_displays_flush ();





    }
  else
    {
      status = STATUS_EXECUTION_ERROR;
    }

  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}


static void
PREFIXok_warn_window(GtkWidget * widget,
		    gpointer   data)
{
  gtk_widget_destroy ((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((data)), ((gtk_widget_get_type ()))))));
}

void
PREFIXcreate_warn_dialog (gchar *msg)
{
  GtkWidget *window = 0;
  GtkWidget *label;
  GtkWidget *button;

  window = gtk_dialog_new ();

  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), "Warning");
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_container_get_type ()))))), 0);

  button = gtk_button_new_with_label ("OK");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) ok_warn_window,
                      window);

 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);
  label = gtk_label_new(msg);
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 10, 10);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), label,(!(0)),(!(0)), 0);
  gtk_widget_show (label);
  gtk_widget_show (window);
}


static void
PREFIXgfig_clear_selection(gint32 image_ID)
{
  GParam *return_vals;
  int nreturn_vals;


#line 780 "gfig.c"
  return_vals = gimp_run_procedure ("gimp_selection_clear",
                                    &nreturn_vals,
                                    PARAM_IMAGE, image_ID,
                                    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);
}


#line 793 "gfig.c"

void
PREFIXplug_in_parse_gfig_path()
{
  GParam *return_vals;
  gint nreturn_vals;
  gchar *path_string;
  gchar *home;
  gchar *path;
  gchar *token;
  struct stat filestat;
  gint	err;
  gchar buf[256];

  if(gfig_path_list)
    g_list_free(gfig_path_list);

  gfig_path_list = 0;

  return_vals = gimp_run_procedure ("gimp_gimprc_query",
				    &nreturn_vals,
				    PARAM_STRING, "gfig-path",
				    PARAM_END);

  if (return_vals[0].data.d_status != STATUS_SUCCESS || return_vals[1].data.d_string == 0)
    {
      g_warning("No gfig-path in gimprc: gfig_path_list is NULL\nSee README file");
      create_warn_dialog("No gfig-path in gimprc: gfig_path_list is NULL. See README file\n");


#line 826 "gfig.c"

      gimp_destroy_params (return_vals, nreturn_vals);
      return;
    }

  path_string = g_strdup (return_vals[1].data.d_string);
  gimp_destroy_params (return_vals, nreturn_vals);


#line 837 "gfig.c"
  home = getenv ("HOME");



  token = strtok (path_string, ":");

  while (token)
    {
      if (*token == '\0')
	{
	  token = strtok (0, ":");
	  continue;
	}

      if (*token == '~')
	{
	  path = g_malloc (strlen (home) + strlen (token) + 2);
	  sprintf (path, "%s%s", home, token + 1);
	}
      else
	{
	  path = g_malloc (strlen (token) + 2);
	  strcpy (path, token);
	}


      err = stat (path, &filestat);

      if (!err &&(((filestat.st_mode)&0xF000) == 0x4000))
	{
	  if (path[strlen (path) - 1] != '/')
	    strcat (path, "/");




	  gfig_path_list = g_list_append (gfig_path_list, path);
	}
      else
	{
	  sprintf(buf,"gfig-path misconfigured - \nPath `%.100s' not found\n", path);
	  g_warning(buf);
	  create_warn_dialog(buf);
	  g_free (path);
	}
      token = strtok (0, ":");
    }
  g_free (path_string);
}



#line 892 "gfig.c"
void
PREFIXgfig_name_encode (gchar *dest, gchar *src)
{
  int	cnt = 256 - 1;

  while (*src && cnt--)
    {
      if (((__ctype + 1)[*src] & 040) ||((__ctype + 1)[*src] & 010) || *src == '\\')
	{
	  sprintf (dest, "\\%03o", *src++);
	  dest += 4;
	}
      else
	*dest++ = *src++;
    }
  *dest = '\0';
}


#line 913 "gfig.c"
void
PREFIXgfig_name_decode (gchar *dest, gchar *src)
{
  int	cnt = 256 - 1;
  int	tmp;

  while (*src && cnt--)
    {
      if (*src == '\\' && *(src+1) && *(src+2) && *(src+3))
	{
	  sscanf (src+1, "%3o", &tmp);
	  *dest++ = tmp;
	  src += 4;
	}
      else
	*dest++ = *src++;
    }
  *dest = '\0';
}



#line 939 "gfig.c"

gint
PREFIXgfig_list_pos(GFIGOBJ *gfig)
{
  GFIGOBJ *g;
  int n;
  GList *tmp;

  n = 0;
  tmp = gfig_list;

  while (tmp)
    {
      g = tmp->data;

      if (strcmp (gfig->draw_name, g->draw_name) <= 0)
	break;
      n++;
      tmp = tmp->next;
    }

  return(n);
}

gint
PREFIXgfig_list_insert (GFIGOBJ *gfig)
{
  int		n;


#line 971 "gfig.c"

  n = gfig_list_pos(gfig);

  gfig_list = g_list_insert (gfig_list, gfig, n);





  return n;
}

void
PREFIXgfig_free(GFIGOBJ * gfig)
{
 if (1){ if (!(gfig != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 986,"gfig != NULL"); }else (void)0;

  if(gfig->obj_list)
    free_all_objs(gfig->obj_list);
  if(gfig->name)
    g_free(gfig->name);
  if(gfig->filename)
    g_free(gfig->filename);
  if(gfig->draw_name)
    g_free(gfig->draw_name);
  g_free (gfig);
}

void
PREFIXgfig_free_everything(GFIGOBJ * gfig)
{
 if (1){ if (!(gfig != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 1002,"gfig != NULL"); }else (void)0;

  if(gfig->filename)
    {



      remove(gfig->filename);
    }
  gfig_free(gfig);
}

void
PREFIXgfig_list_free_all ()
{
  GList * list;
  GFIGOBJ * gfig;

  list = gfig_list;
  while (list)
    {
      gfig = (GFIGOBJ *) list->data;
      gfig_free (gfig);
      list = list->next;
    }

  g_list_free (gfig_list);
  gfig_list = 0;
}


void
PREFIXgfig_list_load_all(GList *plist)
{
  GFIGOBJ  * gfig;
  GList    * list;
  gchar	   * path;
  gchar	   * filename;
  DIR	   * dir;
  struct dirent *dir_ent;
  struct stat	filestat;
  gint		err;


  current_obj = pic_obj = 0;
  gfig_list_free_all ();

  list = plist;
  while (list)
    {
      path = list->data;
      list = list->next;


      dir = opendir (path);

      if (!dir)
	g_warning("error reading GFig directory \"%s\"", path);
      else
	{
	  while ((dir_ent = readdir (dir)))
	    {
	      filename = g_malloc (strlen(path) + strlen (dir_ent->d_name) + 1);

	      sprintf (filename, "%s%s", path, dir_ent->d_name);


	      err = stat (filename, &filestat);

	      if (!err &&(((filestat.st_mode)&0xF000) == 0x8000))
		{
		  gfig = gfig_load (filename, dir_ent->d_name);

		  if (gfig)
		    {

		      if(access(filename,2))
			gfig->obj_status |= 0x2;

		      gfig_list_insert (gfig);
		    }
		}

	      g_free (filename);
	    }
	  closedir (dir);
	}
    }

  if(!gfig_list)
    {

      gfig = gfig_new();
      gfig->draw_name = g_strdup("First gfig");
      gfig_list_insert(gfig);
    }
  pic_obj = current_obj = gfig_list->data;

}

GFIGOBJ *
PREFIXgfig_new(void)
{
  GFIGOBJ * new;

  new =((GFIGOBJ *) g_malloc0 ((unsigned) sizeof (GFIGOBJ) * (1)));
  return(new);
}

void
PREFIXgfig_load_objs(GFIGOBJ *gfig,gint load_count,FILE *fp)
{
  DOBJECT *obj;
  gchar load_buf[256];




  while(load_count-- > 0)
    {
      obj = 0;
      get_line(load_buf,256,fp,0);

      if(!strcmp(load_buf,"<LINE>"))
	{
	  obj = d_load_line(fp);
	}
      else if(!strcmp(load_buf,"<CIRCLE>"))
	{
	  obj = d_load_circle(fp);
	}
      else if(!strcmp(load_buf,"<ELLIPSE>"))
	{
	  obj = d_load_ellipse(fp);
	}
      else if(!strcmp(load_buf,"<POLY>"))
	{
	  obj = d_load_poly(fp);
	}
      else if(!strcmp(load_buf,"<STAR>"))
	{
	  obj = d_load_star(fp);
	}
      else if(!strcmp(load_buf,"<SPIRAL>"))
	{
	  obj = d_load_spiral(fp);
	}
      else if(!strcmp(load_buf,"<BEZIER>"))
	{
	  obj = d_load_bezier(fp);
	}
      else if(!strcmp(load_buf,"<ARC>"))
	{
	  obj = d_load_arc(fp);
	}
      else
	{
	  g_warning("Unknown obj type file %s line %d\n",gfig->filename,line_no);
	}

      if(obj)
	{
	  add_to_all_obj(gfig,obj);
	}
    }
}

GFIGOBJ *
PREFIXgfig_load (gchar *filename, gchar *name)
{
  GFIGOBJ * gfig;
  FILE * fp;
  gchar load_buf[256];
  gchar str_buf[256];
  gint chk_count;
  gint load_count = 0;

 if (1){ if (!(filename != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 1179,"filename != NULL"); }else (void)0;





  fp = fopen (filename, "r");
  if (!fp)
    {
      g_warning ("Error opening: %s", filename);
      return 0;
    }

  gfig = gfig_new();

  gfig->name = g_strdup(name);
  gfig->filename = g_strdup(filename);



#line 1203 "gfig.c"

  get_line(load_buf,256,fp,1);

  if(strncmp("GFIG Version 0.1\n",load_buf,strlen(load_buf)))
    {
      gchar err[256];
      sprintf(err,"File '%s' is not a gfig file",gfig->filename);
      create_warn_dialog(err);
      return(0);
    }

  get_line(load_buf,256,fp,0);
  sscanf(load_buf,"Name: %100s",str_buf);
  gfig_name_decode(load_buf,str_buf);
  gfig->draw_name = g_strdup(load_buf);

  get_line(load_buf,256,fp,0);
  sscanf(load_buf,"Version: %f",&gfig->version);

  get_line(load_buf,256,fp,0);
  sscanf(load_buf,"ObjCount: %d",&load_count);

  if(load_options(gfig,fp))
    {

      gchar err[256];
      sprintf(err,
	      "File '%s' corrupt file - Line %d Option section incorrect",
	      filename,
	      line_no);
      create_warn_dialog(err);
      return(0);
    }



  gfig_load_objs(gfig,load_count,fp);



  chk_count = gfig_obj_counts(gfig->obj_list);

  if(chk_count != load_count)
    {

      gchar err[256];
      sprintf(err,"File '%s' corrupt file - Line %d Object count to small",
	      filename,
	      line_no);
      create_warn_dialog(err);
      return(0);
    }

  fclose(fp);

  if(!pic_obj)
    pic_obj = gfig;

  gfig->obj_status = 0x0;

  return(gfig);
}

void
PREFIXsave_options(FILE *fp)
{

  fprintf(fp,"<OPTIONS>\n");
  fprintf(fp,"GridSpacing: %d\n",selvals.opts.gridspacing);
  if(selvals.opts.gridtype == RECT_GRID)
     fprintf(fp,"GridType: RECT_GRID\n");
  else if(selvals.opts.gridtype == POLAR_GRID)
     fprintf(fp,"GridType: POLAR_GRID\n");
  else if(selvals.opts.gridtype == ISO_GRID)
     fprintf(fp,"GridType: ISO_GRID\n");
  else fprintf(fp,"GridType: RECT_GRID\n");
  fprintf(fp,"DrawGrid: %s\n",(selvals.opts.drawgrid)?"TRUE":"FALSE");
  fprintf(fp,"Snap2Grid: %s\n",(selvals.opts.snap2grid)?"TRUE":"FALSE");
  fprintf(fp,"LockOnGrid: %s\n",(selvals.opts.lockongrid)?"TRUE":"FALSE");

  fprintf(fp,"ShowControl: %s\n",(selvals.opts.showcontrol)?"TRUE":"FALSE");
  fprintf(fp,"</OPTIONS>\n");
}

gint
PREFIXload_bool(gchar * opt_buf,gint *toset)
{
  if(!strcmp(opt_buf,"TRUE"))
    *toset = 1;
  else if(!strcmp(opt_buf,"FALSE"))
    *toset = 0;
  else
    return(-1);

  return(0);
}

static void
PREFIXupdate_options(GFIGOBJ *old_obj)
{

  if(selvals.opts.gridspacing != old_obj->opts.gridspacing)
    {
      old_obj->opts.gridspacing = selvals.opts.gridspacing;
    }
  if(selvals.opts.gridtype != old_obj->opts.gridtype)
    {
      old_obj->opts.gridtype = selvals.opts.gridtype;
    }
  if(selvals.opts.drawgrid != old_obj->opts.drawgrid)
    {
      old_obj->opts.drawgrid = selvals.opts.drawgrid;
    }
  if(selvals.opts.snap2grid != old_obj->opts.snap2grid)
    {
      old_obj->opts.snap2grid = selvals.opts.snap2grid;
    }
  if(selvals.opts.lockongrid != old_obj->opts.lockongrid)
    {
      old_obj->opts.lockongrid = selvals.opts.lockongrid;
    }
  if(selvals.opts.showcontrol != old_obj->opts.showcontrol)
    {
      old_obj->opts.showcontrol = selvals.opts.showcontrol;
    }


  if(selvals.opts.gridspacing != current_obj->opts.gridspacing)
    {

(((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.gridspacing)), ((gtk_adjustment_get_type ())))))->value = current_obj->opts.gridspacing;
      gtk_signal_emit_by_name((( (gfig_opt_widget.gridspacing) != 0 &&( (((GtkObject*) (gfig_opt_widget.gridspacing))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (gfig_opt_widget.gridspacing))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (gfig_opt_widget.gridspacing) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (gfig_opt_widget.gridspacing), GTK_TYPE_OBJECT) ), "value_changed");
    }
  if(selvals.opts.drawgrid != current_obj->opts.drawgrid)
    {
      gtk_toggle_button_set_active ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.drawgrid)), ((gtk_toggle_button_get_type ()))))),current_obj->opts.drawgrid);
    }
  if(selvals.opts.snap2grid != current_obj->opts.snap2grid)
    {
      gtk_toggle_button_set_active ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.snap2grid)), ((gtk_toggle_button_get_type ()))))),current_obj->opts.snap2grid);
    }
  if(selvals.opts.lockongrid != current_obj->opts.lockongrid)
    {
      gtk_toggle_button_set_active ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.lockongrid)), ((gtk_toggle_button_get_type ()))))),current_obj->opts.lockongrid);
    }
  if(selvals.opts.showcontrol != current_obj->opts.showcontrol)
    {
      gtk_toggle_button_set_active ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.showcontrol)), ((gtk_toggle_button_get_type ()))))),current_obj->opts.showcontrol);
    }
  if(selvals.opts.gridtype != current_obj->opts.gridtype)
    {
      gtk_option_menu_set_history ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.gridtypemenu)), ((gtk_option_menu_get_type ()))))),
				   current_obj->opts.gridtype);

      gridtype_menu_callback(
			     gtk_menu_get_active(
(((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gtk_option_menu_get_menu((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_opt_widget.gridtypemenu)), ((gtk_option_menu_get_type ())))))))), ((gtk_menu_get_type ())))))),(gpointer)1);
#line 1361 "gfig.c"










    }
}

gint
PREFIXload_options(GFIGOBJ *gfig,FILE *fp)
{
  gchar load_buf[256];
  gchar str_buf[256];
  gchar opt_buf[256];

  get_line(load_buf,256,fp,0);





  if(strcmp(load_buf,"<OPTIONS>"))
    return(-1);

  get_line(load_buf,256,fp,0);





  while(strcmp(load_buf,"</OPTIONS>"))
    {






      sscanf(load_buf,"%s %s",str_buf,opt_buf);


      if(!strcmp(str_buf,"GridSpacing:"))
	{

	  int sp = 0;
	  sp = atoi(opt_buf);
	  if(sp <= 0)
	    return(-1);
	  gfig->opts.gridspacing = sp;
	}
      else if(!strcmp(str_buf,"DrawGrid:"))
	{

	  if(load_bool(opt_buf,&gfig->opts.drawgrid))
	    return(-1);
	}
      else if(!strcmp(str_buf,"Snap2Grid:"))
	{

	  if(load_bool(opt_buf,&gfig->opts.snap2grid))
	    return(-1);
	}
      else if(!strcmp(str_buf,"LockOnGrid:"))
	{

	  if(load_bool(opt_buf,&gfig->opts.lockongrid))
	    return(-1);
	}
      else if(!strcmp(str_buf,"ShowControl:"))
	{

	  if(load_bool(opt_buf,&gfig->opts.showcontrol))
	    return(-1);
	}
      else if(!strcmp(str_buf,"GridType:"))
	{

	  if(!strcmp(opt_buf,"RECT_GRID"))
	    gfig->opts.gridtype = RECT_GRID;
	  else if(!strcmp(opt_buf,"POLAR_GRID"))
	    gfig->opts.gridtype = POLAR_GRID;
	  else if(!strcmp(opt_buf,"ISO_GRID"))
	    gfig->opts.gridtype = ISO_GRID;
	  else
	    return(-1);
	}

      get_line(load_buf,256,fp,0);



    }
  return(0);
}

gint
PREFIXgfig_obj_counts(DALLOBJS * objs)
{
  gint count = 0;

  while(objs)
    {
      count++;
      objs = objs->next;
    }

  return(count);
}

void
PREFIXgfig_save_callbk()
{
  FILE *fp;
  DALLOBJS * objs;
  gint count = 0;
  gchar * savename;
  gchar conv_buf[256 *3 +1];

  savename = current_obj->filename;

  fp = fopen (savename, "w+");

  if (!fp)
    {
      gchar errbuf[256];
      sprintf(errbuf,"Error opening '%.100s' could not save", savename);
      create_warn_dialog(errbuf);
      g_warning (errbuf);
      return;
    }


  fputs("GFIG Version 0.1\n",fp);


#line 1506 "gfig.c"

  gfig_name_encode(conv_buf,current_obj->draw_name);
  fprintf(fp,"Name: %s\n",conv_buf);
  fprintf(fp,"Version: %f\n",current_obj->version);
  objs = current_obj->obj_list;

  count = gfig_obj_counts(objs);

  fprintf(fp,"ObjCount: %d\n",count);

  save_options(fp);

  objs = current_obj->obj_list;
  while(objs)
    {
      objs->obj->savefunc(objs->obj,fp);
      objs = objs->next;
    }

  if(((fp)->_flag& 040))
    create_warn_dialog("Failed to write file\n");
  else
    {
      gfig_obj_modified(current_obj,0x0);
      current_obj->obj_status &= ~(0x1 | 0x2);
    }

  fclose(fp);

  gfig_update_stat_labels();
}

void
PREFIXfile_selection_ok (GtkWidget        *w,
		   GtkFileSelection *fs,
		   gpointer data)
{
  gchar *filenamebuf;
  struct stat filestat;
  gint	err;
  GFIGOBJ *obj = (GFIGOBJ *)gtk_object_get_user_data((( (fs) != 0 &&( (((GtkObject*) (fs))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (fs))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (fs) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (fs), GTK_TYPE_OBJECT) ));
  GFIGOBJ *real_current;

  filenamebuf = gtk_file_selection_get_filename ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((fs)), ((gtk_file_selection_get_type ()))))));





  if(strlen(filenamebuf) == 0)
    {
      create_warn_dialog("Save:- No filename given");
      return;
    }


  err = stat (filenamebuf, &filestat);

  if (!err &&(((filestat.st_mode)&0xF000) == 0x4000))
    {

      create_warn_dialog("Save:- Can't save to a directory");
      return;
    }

  obj->filename = g_strdup(filenamebuf);

  real_current = current_obj;
  current_obj = obj;
  gfig_save_callbk();
  current_obj = current_obj;

  gtk_widget_destroy((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((fs)), ((gtk_widget_get_type ()))))));

}

void
PREFIXdestroy_window (GtkWidget  *widget,
		GtkWidget **window)
{
  *window = 0;
}

void
PREFIXhide_file_sel(GtkWidget  *widget,
		gpointer w)
{
  gtk_widget_destroy((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((w)), ((gtk_widget_get_type ()))))));
}

void
PREFIXcreate_file_selection (GFIGOBJ *obj, gchar *tpath)
{
  static GtkWidget *window = 0;

  if (!window)
    {
      window = gtk_file_selection_new ("Save gfig drawing");
 gtk_window_set_position ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);

      gtk_signal_connect ((( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ), "destroy",
			  (GtkSignalFunc) destroy_window,
			  &window);

      gtk_object_set_user_data((( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ),obj);
      gtk_signal_connect ((( ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button) != 0 &&( (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button), GTK_TYPE_OBJECT) ),
			  "clicked", (GtkSignalFunc) file_selection_ok,
			  (gpointer)window);
      gtk_signal_connect_object((( ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button) != 0 &&( (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button), GTK_TYPE_OBJECT) ),
				 "clicked", (GtkSignalFunc) gtk_widget_destroy,
(( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ));
    }

  if(tpath)
    {
      gtk_file_selection_set_filename((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ()))))),tpath);
    }
  else

    if(gfig_path_list)
      {
	gtk_file_selection_set_filename((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ()))))),
					g_list_nth(gfig_path_list,
						   g_list_length(gfig_path_list)-1)->data);
      }
  else
    gtk_file_selection_set_filename((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ()))))),"/tmp");

  if (!(((((( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) )->flags)) & GTK_VISIBLE) != 0))
    gtk_widget_show (window);

}

void
PREFIXgfig_save(void)
{

  if(!current_obj->filename)
   {

     create_file_selection(current_obj,0);
     return;
    }
  gfig_save_callbk();
}


static gchar*
PREFIXmy_gdk_pixmap_skip_whitespaces (gchar *buffer)
{
  gint32 index = 0;

  while (buffer[index] != 0 && (buffer[index] == 0x20 || buffer[index] == 0x09))
    index++;

  return &buffer[index];
}

static gchar*
PREFIXmy_gdk_pixmap_skip_string (gchar *buffer)
{
  gint32 index = 0;

  while (buffer[index] != 0 && buffer[index] != 0x20 && buffer[index] != 0x09)
    index++;

  return &buffer[index];
}




static gchar*
PREFIXmy_gdk_pixmap_extract_color (gchar *buffer)
{
  gint counter, numnames;
  gchar *ptr = 0, ch, temp[128];
  gchar color[120], *retcol;
  gint space;

  counter = 0;
  while (ptr == 0)
    {
      if (buffer[counter] == 'c')
        {
          ch = buffer[counter + 1];
          if (ch == 0x20 || ch == 0x09)
            ptr = &buffer[counter + 1];
        }
      else if (buffer[counter] == 0)
        return 0;

      counter++;
    }

  ptr = my_gdk_pixmap_skip_whitespaces (ptr);

  if (ptr[0] == 0)
    return 0;
  else if (ptr[0] == '#')
    {
      counter = 1;
      while (ptr[counter] != 0 &&
             ((ptr[counter] >= '0' && ptr[counter] <= '9') ||
              (ptr[counter] >= 'a' && ptr[counter] <= 'f') ||
              (ptr[counter] >= 'A' && ptr[counter] <= 'F')))
        counter++;

      retcol =((gchar *) g_malloc ((unsigned) sizeof (gchar) * (counter+1)));
      strncpy (retcol, ptr, counter);

      retcol[counter] = 0;

      return retcol;
    }

  color[0] = 0;
  numnames = 0;

  space = 120 - 1;
  while (space > 0)
    {
      sscanf (ptr, "%127s", temp);

      if (((gint)ptr[0] == 0) ||
          (strcmp ("s", temp) == 0) || (strcmp ("m", temp) == 0) ||
          (strcmp ("g", temp) == 0) || (strcmp ("g4", temp) == 0))
        {
          break;
        }
      else
        {
          if (numnames > 0)
            {
              space -= 1;
              strcat (color, " ");
            }
          strncat (color, temp, space);
          space -=(((space) < (strlen (temp))) ? (space) : (strlen (temp)));
          ptr = my_gdk_pixmap_skip_string (ptr);
          ptr = my_gdk_pixmap_skip_whitespaces (ptr);
          numnames++;
        }
    }

  retcol = g_strdup (color);
  return retcol;
}

GdkPixmap*
PREFIXmy_gdk_pixmap_create_from_xpm_d (GdkWindow  *window,
			      GdkBitmap **mask,
			      GdkColor   *transparent_color,
			      gchar     **data)
{
  GdkPixmap *pixmap = 0;
  GdkImage *image = 0;
  GdkColormap *colormap;
  GdkVisual *visual;
  GdkGC *gc;
  gint width, height, num_cols, cpp, cnt, n, ns, xcnt, ycnt, i;
  gchar *buffer, *color_name = 0, pixel_str[32];
  _GdkPixmapColor *colors = 0, *color = 0;
  gulong index;
  extern void * gdk_root_parent;

  if (!window)
    window = (GdkWindow*) &gdk_root_parent;

  i = 0;
  buffer = data[i++];
  sscanf (buffer,"%d %d %d %d", &width, &height, &num_cols, &cpp);

  colors =((_GdkPixmapColor *) g_malloc ((unsigned) sizeof (_GdkPixmapColor) * (num_cols)));

  colormap = xxx;
  visual = yyy;

  for (cnt = 0; cnt < num_cols; cnt++)
    {
      buffer = data[i++];

      colors[cnt].color_string =((gchar *) g_malloc ((unsigned) sizeof (gchar) * (cpp + 1)));
      for (n = 0; n < cpp; n++)
	colors[cnt].color_string[n] = buffer[n];
      colors[cnt].color_string[n] = 0;
      colors[cnt].transparent =(0);

      if (color_name != 0)
	g_free (color_name);

      color_name = (gchar *)my_gdk_pixmap_extract_color (&buffer[cpp]);

      if (color_name != 0)
	{
	  if (gdk_color_parse (color_name, &colors[cnt].color) ==(0))
	    {
	      colors[cnt].color = *transparent_color;
	      colors[cnt].transparent =(!(0));
	    }
	}
      else
	{
	  colors[cnt].color = *transparent_color;
	  colors[cnt].transparent =(!(0));
	}

      gdk_color_alloc (colormap, &colors[cnt].color);
    }

  index = 0;
  image = gdk_image_new (GDK_IMAGE_FASTEST, visual, width, height);

  gc = 0;
  if (mask)
    {
      GdkColor tmp_color;

      *mask = gdk_pixmap_new (window, width, height, 1);
      gc = gdk_gc_new (*mask);
      gdk_draw_rectangle (*mask, gc,(!(0)), 0, 0, -1, -1);

      gdk_color_white (colormap, &tmp_color);
      gdk_gc_set_foreground (gc, &tmp_color);
    }

  for (ycnt = 0; ycnt < height; ycnt++)
    {
      buffer = data[i++];

      for (n = 0, cnt = 0, xcnt = 0; n < (width * cpp); n += cpp, xcnt++)
	{
	  strncpy (pixel_str, &buffer[n], cpp);
	  pixel_str[cpp] = 0;
	  color = 0;
	  ns = 0;

	  while (color == 0)
	    {
	      if (strcmp (pixel_str, colors[ns].color_string) == 0)
		color = &colors[ns];
	      else
		ns++;
	    }

	  gdk_image_put_pixel (image, xcnt, ycnt, color->color.pixel);

	  if (mask && color->transparent)
	    {
	      if (cnt < xcnt)
		gdk_draw_line (*mask, gc, cnt, ycnt, xcnt - 1, ycnt);
	      cnt = xcnt + 1;
	    }
	}

      if (mask && (cnt < xcnt))
	gdk_draw_line (*mask, gc, cnt, ycnt, xcnt - 1, ycnt);
    }

  if (mask)
    gdk_gc_destroy (gc);

  pixmap = gdk_pixmap_new (window, width, height, visual->depth);

  gc = gdk_gc_new (pixmap);
  gdk_gc_set_foreground (gc, transparent_color);
  gdk_draw_image (pixmap, gc, image, 0, 0, 0, 0, image->width, image->height);
  gdk_gc_destroy (gc);
  gdk_image_destroy (image);

  if (colors != 0)
    {
      for (cnt = 0; cnt < num_cols; cnt++)
	g_free (colors[cnt].color_string);
      g_free (colors);
    }

  return pixmap;
}







static void
PREFIXcache_preview()
{
  GPixelRgn src_rgn;
  int y,x;
  guchar *src_rows;
  guchar *p;
  int isgrey = 0;

  gimp_pixel_rgn_init(&src_rgn,gfig_select_drawable,sel_x1,sel_y1,sel_width,sel_height,(0),(0));

  src_rows =((guchar *) g_malloc ((unsigned) sizeof (guchar) * (sel_width*4)));
  p = pv_cache =((guchar *) g_malloc ((unsigned) sizeof (guchar) * (preview_width*preview_height*4)));

  real_img_bpp = gimp_drawable_bpp(gfig_select_drawable->id);

  has_alpha = gimp_drawable_has_alpha(gfig_select_drawable->id);

  if(real_img_bpp < 3)
    {
      img_bpp = 3 + has_alpha;
    }
  else
    {
      img_bpp = real_img_bpp;
    }

  switch ( gimp_drawable_type (gfig_select_drawable->id) )
    {
    case GRAYA_IMAGE:
    case GRAY_IMAGE:
      isgrey = 1;
    default:
      break;
    }



  for (y = 0; y < preview_height; y++) {

    gimp_pixel_rgn_get_row(&src_rgn,
			   src_rows,
			   sel_x1,
			   sel_y1 + (y*sel_height)/preview_height,
			   sel_width);

    for (x = 0; x < (preview_width); x ++) {

      int i;
      for (i = 0 ; i < 3; i++ )
	p[x*img_bpp+i] = src_rows[((x*sel_width)/preview_width)*src_rgn.bpp +((isgrey)?0:i)];
      if(has_alpha)
	p[x*img_bpp+3] = src_rows[((x*sel_width)/preview_width)*src_rgn.bpp + ((isgrey)?1:3)];
    }
    p += (preview_width*img_bpp);
  }
  g_free(src_rows);
}

static void
PREFIXrefill_cache()
{
  GdkCursorType ctype1 = GDK_WATCH;
  GdkCursorType ctype2 = GDK_TOP_LEFT_ARROW;
  static GdkCursor *preview_cursor1;
  static GdkCursor *preview_cursor2;

  if(!preview_cursor1)
    preview_cursor1 = gdk_cursor_new(ctype1);

  if(!preview_cursor2)
    preview_cursor2 = gdk_cursor_new(ctype2);

  gdk_window_set_cursor(gtk_widget_get_toplevel((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_preview)), ((gtk_widget_get_type ()))))))->window,
			preview_cursor1);

  gdk_window_set_cursor(gfig_preview->window,preview_cursor1);

  gdk_flush();

  cache_preview();

  gdk_window_set_cursor(gtk_widget_get_toplevel((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_preview)), ((gtk_widget_get_type ()))))))->window,
			preview_cursor2);

  toggle_obj_type(0,(gpointer)selvals.otype);

}

void
PREFIXgfig_set_pixmap(GFIGOBJ *obj,char **pixdata)
{
  GdkPixmap *pixmap;
  GdkColor transparent;
  GdkBitmap *mask;

  pixmap = my_gdk_pixmap_create_from_xpm_d(gfig_gtk_list->window,&mask,&transparent,pixdata);
  gtk_pixmap_set((((GtkPixmap *) gtk_type_check_object_cast ((GtkTypeObject*) ((obj->pixmap_widget)), ((gtk_pixmap_get_type ()))))),pixmap,mask);
}


GtkWidget*
PREFIXgfig_new_pixmap(GtkWidget *list, char **pixdata)
{
  GtkWidget *pixmap_widget;
  GdkPixmap *pixmap;
  GdkColor transparent;
  GdkBitmap *mask;

  pixmap = my_gdk_pixmap_create_from_xpm_d(list->window,&mask,&transparent,pixdata);
  pixmap_widget = gtk_pixmap_new(pixmap,mask);
  gtk_widget_show(pixmap_widget);
  return(pixmap_widget);
}

GtkWidget*
PREFIXgfig_list_item_new_with_label_and_pixmap (GFIGOBJ *obj, gchar *label, GtkWidget *pix_widget)
{
  GtkWidget *list_item;
  GtkWidget *label_widget;
  GtkWidget *alignment;
  GtkWidget *hbox;

  hbox = gtk_hbox_new((0),1);
  gtk_widget_show(hbox);

  list_item = gtk_list_item_new ();
  label_widget = gtk_label_new (label);
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label_widget)), ((gtk_misc_get_type ()))))), 0.0, 0.5);

  alignment = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((alignment)), ((gtk_container_get_type ()))))), 0);
  gtk_widget_show(alignment);

  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))),pix_widget,(0),(0),0);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_container_get_type ()))))), label_widget);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_item)), ((gtk_container_get_type ()))))), hbox);

  gtk_widget_show (obj->label_widget = label_widget);
  gtk_widget_show (obj->pixmap_widget = pix_widget);
  gtk_widget_show (obj->list_item = list_item);

  return list_item;
}

GtkWidget*
PREFIXgfig_menu_item_new_with_pixmap(GtkWidget *pix_widget)
{
  GtkWidget *menu_item;

  menu_item = gtk_menu_item_new ();
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu_item)), ((gtk_container_get_type ()))))), pix_widget);
  gtk_widget_show(menu_item);

  return menu_item;
}

void
PREFIXgfig_obj_modified(GFIGOBJ *obj,gint stat_type)
{
  GdkPixmap *gdk_pix;



 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 2067,"obj != NULL"); }else (void)0;

  if(obj->obj_status == stat_type)
    return;


  gtk_pixmap_get((((GtkPixmap *) gtk_type_check_object_cast ((GtkTypeObject*) ((obj->pixmap_widget)), ((gtk_pixmap_get_type ()))))),&gdk_pix,0);


  if(stat_type == 0x1)
    gfig_set_pixmap(obj,Floppy6_xpm);
  else
    gfig_set_pixmap(obj,blank_xpm);


  gdk_pixmap_unref(gdk_pix);
  gtk_widget_draw((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((obj->list_item)), ((gtk_widget_get_type ()))))),0);
}

static gint
PREFIXselect_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer data)
{
  gint type = (gint)data;
  gint count = 0;
  DALLOBJS * objs;

  if(current_obj)
    {
      objs = current_obj->obj_list;

      while(objs)
	{
	  objs = objs->next;
	  count++;
	}
    }

  switch(type)
    {
    case 2 :
      obj_show_single--;
      if(obj_show_single < 0)
	obj_show_single = count - 1;
      break;
    case 1 :
      obj_show_single++;
      if(obj_show_single >= count)
	obj_show_single = 0;
      break;
    case 4 :
      obj_show_single = -1;
      break;
    default:
      break;
    }

  draw_grid_clear(widget,data);

  return((0));
}

static GtkWidget *
PREFIXobj_select_buttons(void)
{
  GtkWidget *button;
  GtkWidget *hbox,*vbox;

  hbox = gtk_hbox_new((0), 0);
  vbox = gtk_vbox_new((0), 0);

  button = gtk_button_new_with_label ("<");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) select_button_press,
		      (gpointer)2);
  gtk_widget_show(button);

  button = gtk_button_new_with_label (">");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) select_button_press,
		      (gpointer)1);
  gtk_widget_show(button);

  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), hbox,(!(0)),(!(0)), 0);

  button = gtk_button_new_with_label ("==");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) select_button_press,
		      (gpointer)4);
  gtk_widget_show(button);

  gtk_widget_show(hbox);
  gtk_widget_show(vbox);
  return(vbox);
}

static GtkWidget *
PREFIXbut_with_pix(char **pixdata, GSList **group, gint baction)
{
  GtkWidget * button;
  GtkWidget * alignment;
  GtkWidget * pixmap_widget;

  button = gtk_radio_button_new(*group);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((button)), ((gtk_container_get_type ()))))), 0);
  gtk_toggle_button_set_mode ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((button)), ((gtk_toggle_button_get_type ()))))),(0));
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) toggle_obj_type,
		      (gpointer)baction);

  gtk_widget_show(button);

  *group = gtk_radio_button_group ((((GtkRadioButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((button)), ((gtk_radio_button_get_type ()))))));

  alignment = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((alignment)), ((gtk_container_get_type ()))))), 0);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((button)), ((gtk_container_get_type ()))))), alignment);

  pixmap_widget = gfig_new_pixmap(button,pixdata);
  gtk_widget_show(pixmap_widget);
  gtk_widget_show(alignment);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((alignment)), ((gtk_container_get_type ()))))), pixmap_widget);
  return(button);
}


static GtkWidget *
PREFIXsmall_preview(GtkWidget * list)
{
  GtkWidget * label;
  GtkWidget * frame;
  GtkWidget * button;
  GtkWidget * vbox;
  gint y;

  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 4);
  gtk_widget_show(vbox);

  label = gtk_label_new("Prev");
  gtk_widget_show(label);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), label);

  frame = gtk_frame_new (0);
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_IN);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), 0);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), frame);
  gtk_widget_show(frame);

  pic_preview = gtk_preview_new(GTK_PREVIEW_COLOR);
  gtk_widget_show(pic_preview);
  gtk_preview_size((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((pic_preview)), ((gtk_preview_get_type ()))))),48,48);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), pic_preview);


  for(y = 0; y < 48; y++)
    {
      guchar prow[48 *3];
      memset(prow,-1,48 *3);
      gtk_preview_draw_row((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((pic_preview)), ((gtk_preview_get_type ()))))), prow, 0, y,48);
    }

  gtk_signal_connect_after((( (pic_preview) != 0 &&( (((GtkObject*) (pic_preview))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (pic_preview))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (pic_preview) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (pic_preview), GTK_TYPE_OBJECT) ), "expose_event",
			    (GtkSignalFunc) pic_preview_expose,
 0);



  button = gtk_button_new_with_label ("Edit");
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) edit_button_press,
		      (gpointer) list);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Edit Gfig object collection",0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("Merge");
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) merge_button_press,
		      (gpointer)list);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Merge Gfig Object collection into the current edit session",0);
  gtk_widget_show(button);

  return(vbox);
}



static void
PREFIXnum_sides_dialog (gchar * d_title,
		  gint * num_sides,
		  gint * which_way,
		  gint adj_min,
		  gint adj_max)
{
  GtkWidget *window = 0;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *hbox;
  GtkObject *size_data;
  GtkWidget *slider;
  GtkWidget *entry;
  gchar buf[512];

  window = gtk_dialog_new ();

  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), d_title);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_container_get_type ()))))), 0);

  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) ok_warn_window,
                      window);

 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  hbox = gtk_hbox_new((0), 0);
  label = gtk_label_new("Number of sides/points/turns:-");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_widget_show (label);
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 1, 1);

  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(!(0)),(!(0)), 0);

  size_data = gtk_adjustment_new (*num_sides, adj_min, adj_max, 5, 1, 0);
  slider = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((size_data)), ((gtk_adjustment_get_type ()))))));
  gtk_widget_set_usize (slider, 100, 0);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))),slider,(!(0)),(!(0)),0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_scale_get_type ()))))), GTK_POS_LEFT);
  gtk_scale_set_digits ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_scale_get_type ()))))), 0);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_range_get_type ()))))),GTK_UPDATE_CONTINUOUS );
  gtk_signal_connect ((( (size_data) != 0 &&( (((GtkObject*) (size_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (size_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (size_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (size_data), GTK_TYPE_OBJECT) ), "value_changed",
		      (GtkSignalFunc) gfig_scale_update,
		      num_sides);
  gtk_widget_show (slider);

  entry = gtk_entry_new();
  gtk_object_set_user_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), size_data);
  gtk_object_set_user_data(size_data, entry);
  gtk_widget_set_usize(entry,28, 0);
  sprintf(buf, "%d", *num_sides);
  gtk_entry_set_text((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((entry)), ((gtk_entry_get_type ()))))), buf);
  gtk_signal_connect((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), "changed",
		     (GtkSignalFunc) gfig_entry_update,
		     num_sides);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), entry,(!(0)),(!(0)), 0);
  gtk_widget_show(entry);

  if(which_way)
    {
      GtkWidget *option_menu;
      GtkWidget *menu;
      GtkWidget *menuitem;


      option_menu = gtk_option_menu_new ();
      menu = gtk_menu_new();

      menuitem = gtk_menu_item_new_with_label("Clockwise");
      gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc)gfig_toggle_update,
			  (gpointer)which_way);

      gtk_widget_show (menuitem);
      gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

      menuitem = gtk_menu_item_new_with_label("Anti-Clockwise");
      gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc)gfig_toggle_update,
			  (gpointer)which_way);

      gtk_widget_show (menuitem);
      gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

      gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
      gtk_widget_show (option_menu);
      gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), option_menu,(!(0)),(!(0)),0);
    }

  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), hbox,(!(0)),(!(0)), 0);
  gtk_widget_show (label);
  gtk_widget_show (hbox);
  gtk_widget_show (window);
}

static void
PREFIXbezier_dialog (void)
{
  GtkWidget *window = 0;
  GtkWidget *vbox;
  GtkWidget *button;
  GtkWidget *toggle;

  window = gtk_dialog_new ();

  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), "Bezier settings");
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_container_get_type ()))))), 0);

  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) ok_warn_window,
                      window);

 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  vbox = gtk_vbox_new((0), 0);

  toggle = gtk_check_button_new_with_label ("Closed");
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&bezier_closed);
  gtk_tooltips_set_tip(gfig_tooltips,toggle,"Close curve on completion",0);
  gtk_toggle_button_set_active ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((toggle)), ((gtk_toggle_button_get_type ()))))),bezier_closed);
  gtk_widget_show(toggle);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))),toggle,(!(0)),(!(0)), 0);

  toggle = gtk_check_button_new_with_label ("Show line frame");
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&bezier_line_frame);
  gtk_tooltips_set_tip(gfig_tooltips,toggle,"Draws lines between the control points. Only during curve creation",0);
  gtk_toggle_button_set_active ((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((toggle)), ((gtk_toggle_button_get_type ()))))),bezier_line_frame);
  gtk_widget_show(toggle);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))),toggle,(!(0)),(!(0)), 0);

  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), vbox,(!(0)),(!(0)), 0);
  gtk_widget_show (vbox);
  gtk_widget_show (window);
}

static gint
PREFIXpoly_button_press (GtkWidget      *w,
                    GdkEventButton *event,
                    gpointer        data)
{
  if ((event->type == GDK_2BUTTON_PRESS) &&
      (event->button == 1))
    num_sides_dialog("Regular polygon number of sides",&poly_num_sides,0,3,200);
  return(0);
}

static gint
PREFIXstar_button_press (GtkWidget      *w,
                    GdkEventButton *event,
                    gpointer        data)
{
  if ((event->type == GDK_2BUTTON_PRESS) &&
      (event->button == 1))
    num_sides_dialog("Star number of points",&star_num_sides,0,3,200);
  return(0);
}

static gint
PREFIXspiral_button_press (GtkWidget      *w,
                    GdkEventButton *event,
                    gpointer        data)
{
  if ((event->type == GDK_2BUTTON_PRESS) &&
      (event->button == 1))
    num_sides_dialog("Spiral number of points",&spiral_num_turns,&spiral_toggle,1,20);
  return(0);
}

static gint
PREFIXbezier_button_press (GtkWidget      *w,
                    GdkEventButton *event,
                    gpointer        data)
{
  if ((event->type == GDK_2BUTTON_PRESS) &&
      (event->button == 1))
    bezier_dialog();
  return(0);
}

static GtkWidget *
PREFIXdraw_buttons(GtkWidget *ww)
{
  GtkWidget * frame;
  GtkWidget * button;
  GtkWidget * vbox;
  GSList * group;

  frame = gtk_frame_new ("Ops");
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), 1);


  group = 0;
  vbox = gtk_vbox_new((0), 0);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), vbox);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 2);


  button = but_with_pix(line_xpm,&group,LINE);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create line",0);

  button = but_with_pix(circle_xpm,&group,CIRCLE);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create circle",0);

  button = but_with_pix(ellipse_xpm,&group,ELLIPSE);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create ellipse",0);

  button = but_with_pix(curve_xpm,&group,ARC);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create arch",0);

  button = but_with_pix(poly_xpm,&group,POLY);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);

  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) poly_button_press,
 0);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create reg polygon",0);

  button = but_with_pix(star_xpm,&group,STAR);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) star_button_press,
 0);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create star",0);

  button = but_with_pix(spiral_xpm,&group,SPIRAL);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show(button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) spiral_button_press,
 0);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create spiral",0);

  button = but_with_pix(bezier_xpm,&group,BEZIER);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show(button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) bezier_button_press,
 0);

  gtk_tooltips_set_tip(gfig_tooltips,button,"Create bezier curve. Shift + Button ends object creation.",0);

  button = but_with_pix(move_obj_xpm,&group,MOVE_OBJ);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Move an object",0);

  button = but_with_pix(move_point_xpm,&group,MOVE_POINT);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Move a single point",0);

  button = but_with_pix(copy_obj_xpm,&group,COPY_OBJ);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Copy an object",0);

  button = but_with_pix(delete_xpm,&group,DEL_OBJ);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Delete an object",0);

  button = obj_select_buttons();
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), button);
  gtk_widget_show(button);








  gtk_widget_show(vbox);
  gtk_widget_show(frame);
  return(frame);
}


static gint
PREFIXgfig_brush_preview_events ( GtkWidget *widget,
			     GdkEvent *event )
{
  GdkEventButton *bevent;
  GdkEventMotion *mevent;
  static GdkPoint point;
  static int have_start = 0;

  switch (event->type)
    {
    case GDK_EXPOSE:
      break;

    case GDK_BUTTON_PRESS:
      bevent = (GdkEventButton *) event;
      point.x = bevent->x;
      point.y = bevent->y;
      have_start = 1;

      break;
    case GDK_BUTTON_RELEASE:
      bevent = (GdkEventButton *) event;
      have_start = 0;

      break;
    case GDK_MOTION_NOTIFY:
      mevent = (GdkEventMotion *) event;

      if(!have_start || !(mevent->state & GDK_BUTTON1_MASK))
	break;

      gfig_brush_fill_preview_xy(widget,point.x - mevent->x,point.y - mevent->y);
      gtk_widget_draw(widget,0);
      point.x = mevent->x;
      point.y = mevent->y;
      break;
    default:
      break;
    }
  return(0);
}

static void
PREFIXgfig_brush_update_preview(GtkWidget *widget,gpointer data)
{
  GtkWidget *pw = (GtkWidget*)data;
  BRUSHDESC *bdesc;



  bdesc = gtk_object_get_user_data ((( (pw) != 0 &&( (((GtkObject*) (pw))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (pw))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (pw) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (pw), GTK_TYPE_OBJECT) ));
  brush_list_button_press(0,0,bdesc);
}

static void
PREFIXgfig_brush_menu_callback(GtkWidget *widget, gpointer data)
{
  BRUSH_TYPE btype = (BRUSH_TYPE)data;

  switch(btype)
    {
    case BRUSH_BRUSH_TYPE:
      selvals.brshtype = btype;
      gtk_widget_hide(pressure_hbox);
      gtk_widget_hide(pencil_hbox);
      gtk_widget_show(fade_out_hbox);
      break;
    case BRUSH_PENCIL_TYPE:
      selvals.brshtype = btype;
      gtk_widget_hide(fade_out_hbox);
      gtk_widget_hide(pressure_hbox);
      gtk_widget_show(pencil_hbox);
      break;
    case BRUSH_AIRBRUSH_TYPE:
      selvals.brshtype = btype;
      gtk_widget_hide(fade_out_hbox);
      gtk_widget_hide(pencil_hbox);
      gtk_widget_show(pressure_hbox);
      break;
    case BRUSH_PATTERN_TYPE:
      selvals.brshtype = btype;
      gtk_widget_hide(fade_out_hbox);
      gtk_widget_hide(pressure_hbox);
      gtk_widget_show(pencil_hbox);
      break;
    default:
      create_warn_dialog("Internal error - invalid brush type");
      break;
    }
  gfig_brush_update_preview(widget,
			    (gpointer)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) )));
}


static GtkWidget *
PREFIXgfig_brush_preview(GtkWidget **pv)
{
  GtkWidget *option_menu;
  GtkWidget *menu;
  GtkWidget *menuitem;
  GtkWidget * frame;
  GtkWidget * hbox;
  GtkWidget * vbox;
  gint y;


  hbox = gtk_hbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_container_get_type ()))))), 4);
  gtk_widget_show(hbox);

  frame = gtk_frame_new (0);
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_IN);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), 0);
  gtk_widget_show(frame);

  *pv = gtk_preview_new(GTK_PREVIEW_COLOR);
  gtk_widget_show(*pv);
  gtk_widget_set_events((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((*pv)), ((gtk_widget_get_type ()))))),GDK_EXPOSURE_MASK | GDK_MOTION_NOTIFY | GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_MOTION_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK);
  gtk_signal_connect((( (*pv) != 0 &&( (((GtkObject*) (*pv))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (*pv))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (*pv) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (*pv), GTK_TYPE_OBJECT) ), "event",
		      (GtkSignalFunc) gfig_brush_preview_events,
 0);
  gtk_preview_size((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((*pv)), ((gtk_preview_get_type ()))))),32,32);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), *pv);


  for(y = 0; y < 32; y++)
    {
      guchar prow[32 *3];
      memset(prow,-1,32 *3);
      gtk_preview_draw_row((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((*pv)), ((gtk_preview_get_type ()))))), prow, 0, y,32);
    }


  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 4);
  gtk_widget_show(vbox);

  option_menu = gtk_option_menu_new ();
  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Brush");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) *pv);

  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc)gfig_brush_menu_callback,
		      (gpointer)BRUSH_BRUSH_TYPE);

  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);


  menuitem = gtk_menu_item_new_with_label("Airbrush");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) *pv);

  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc)gfig_brush_menu_callback,
		      (gpointer)BRUSH_AIRBRUSH_TYPE);

  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Pencil");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) *pv);

  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc)gfig_brush_menu_callback,
		      (gpointer)BRUSH_PENCIL_TYPE);

  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Pattern");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) *pv);

  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc)gfig_brush_menu_callback,
		      (gpointer)BRUSH_PATTERN_TYPE);

  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), option_menu);
  gtk_tooltips_set_tip(gfig_tooltips,option_menu,"Use the brush/pencil or the airbrush when drawing on the image. Pattern paints with currently selected brush with a pattern. Only applies to circles/ellipses if Approx. Circles/Ellipses toggle is set.",0);

  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_container_get_type ()))))), vbox);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_container_get_type ()))))), frame);

  return(hbox);
}

static void
PREFIXgfig_brush_fill_preview_xy(GtkWidget *pw,gint x1 ,gint y1)
{
  gint row_count;
  BRUSHDESC *bdesc = (BRUSHDESC*)gtk_object_get_user_data((( (pw) != 0 &&( (((GtkObject*) (pw))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (pw))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (pw) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (pw), GTK_TYPE_OBJECT) ));


  bdesc->x_off += x1;
  bdesc->y_off += y1;

  if(bdesc->y_off < 0)
    bdesc->y_off = 0;
  if(bdesc->y_off > (bdesc->height - 32))
    bdesc->y_off = bdesc->height - 32;

  if(bdesc->x_off < 0)
    bdesc->x_off = 0;
  if(bdesc->x_off > (bdesc->width - 32))
    bdesc->x_off = bdesc->width - 32;


  for(row_count = 0; row_count < 32; row_count++)
    gtk_preview_draw_row((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((pw)), ((gtk_preview_get_type ()))))),
			 &bdesc->pv_buf[bdesc->x_off*bdesc->bpp
				       + (bdesc->width
					  *bdesc->bpp
					  *(row_count + bdesc->y_off))],
			 0,
			 row_count,
 32);
}

static void
PREFIXgfig_brush_fill_preview(GtkWidget *pw,gint32 layer_ID, BRUSHDESC * bdesc)
{
  GPixelRgn src_rgn;
  GDrawable *brushdrawable;
  gint bcount = 3;

  if(bdesc->pv_buf)
    {
      g_free(bdesc->pv_buf);
    }

  brushdrawable = gimp_drawable_get(layer_ID);

  bdesc->bpp = bcount;


  gimp_pixel_rgn_init(&src_rgn,brushdrawable,0,0,bdesc->width,bdesc->height,(0),(0));

  bdesc->pv_buf =((guchar *) g_malloc ((unsigned) sizeof (guchar) * (bdesc->width*bdesc->height*bcount)));
  bdesc->x_off = bdesc->y_off = 0;

  gimp_pixel_rgn_get_rect(&src_rgn,bdesc->pv_buf,0,0,bdesc->width,bdesc->height);


  gfig_brush_fill_preview_xy(pw,0,0);
}

void
PREFIXmygimp_brush_set(gchar *bname)
{
  GParam *return_vals;
  int nreturn_vals;

  return_vals = gimp_run_procedure ("gimp_brushes_set_brush",
				    &nreturn_vals,
				    PARAM_STRING, bname,
				    PARAM_END);

  if (return_vals[0].data.d_status != STATUS_SUCCESS)
    {
      create_warn_dialog("Can't set brush...(1)");
    }

  gimp_destroy_params (return_vals, nreturn_vals);
}

static gchar *
PREFIXmygimp_brush_get (void)
{
  GParam *return_vals;
  int nreturn_vals;
  static gchar saved_bname[1024];

  return_vals = gimp_run_procedure ("gimp_brushes_get_brush",
                                    &nreturn_vals,
				    PARAM_END);

  if (return_vals[0].data.d_status == STATUS_SUCCESS)
    {
      strncpy(saved_bname,return_vals[1].data.d_string,sizeof(saved_bname));
    }
  else
    {
      saved_bname[0] = '\0';
    }

  gimp_destroy_params (return_vals, nreturn_vals);

  return(saved_bname);
}

static void
PREFIXmygimp_brush_info(gint32 *width,
		  gint32 *height)
{
  GParam *return_vals;
  int nreturn_vals;

  return_vals = gimp_run_procedure ("gimp_brushes_get_brush",
                                    &nreturn_vals,
				    PARAM_END);

  if (return_vals[0].data.d_status == STATUS_SUCCESS)
    {
      *width =(((return_vals[2].data.d_int32) > (32)) ? (return_vals[2].data.d_int32) : (32));
      *height =(((return_vals[3].data.d_int32) > (32)) ? (return_vals[3].data.d_int32) : (32));
    }
  else
    {
      create_warn_dialog("Failed to get brush info");
      *width = *height = 48;
    }

  gimp_destroy_params (return_vals, nreturn_vals);
}

static void
PREFIXgfig_brush_img_del(void)
{
  gimp_image_delete(brush_image_ID);
  brush_image_ID = -1;
}

static gint32
PREFIXgfig_gen_brush_preview(BRUSHDESC *bdesc)
{

#line 2898 "gfig.c"
  GParam *return_vals = 0;
  int nreturn_vals;
  static gint32 layer_ID = -1;
  guchar fR,fG,fB;
  guchar bR,bG,bB;
  gchar *saved_bname;
  gint32 width,height;
  gdouble line_pnts[2];

  if(brush_image_ID == -1)
    {

      brush_image_ID = gimp_image_new(48,48,0);
      if(brush_image_ID < 0)
	{
	  create_warn_dialog("Failed to generate brush preview");
	  return(-1);
	}
      if((layer_ID = gimp_layer_new(brush_image_ID,
				    "Brush preview",
				    48,
				    48,
				    0,
				    100.0,
				    0 )) < 0)
	{
	  create_warn_dialog("Error in creating layer for brush preview\n");
	  return(-1);
	}
      gimp_image_add_layer(brush_image_ID,layer_ID,-1);
    }




#line 2936 "gfig.c"

  gimp_palette_get_foreground(&fR,&fG,&fB);
  gimp_palette_get_background(&bR,&bG,&bB);
  saved_bname = mygimp_brush_get();

  gimp_palette_set_background((guchar)-1,(guchar)-1,(guchar)-1);
  gimp_palette_set_foreground(0,0,0);
  mygimp_brush_set(bdesc->bname);

  mygimp_brush_info(&width,&height);
  bdesc->width = width;
  bdesc->height = height;
  line_pnts[0] = (gdouble)width/2;
  line_pnts[1] = (gdouble)height/2;

  gimp_layer_resize(layer_ID,width,height,0,0);
  gimp_image_resize(brush_image_ID,width,height,0,0);

  gimp_drawable_fill(layer_ID,1);



  switch(selvals.brshtype)
	 {
	 case BRUSH_BRUSH_TYPE:
	   return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					     PARAM_DRAWABLE, layer_ID,
					     PARAM_FLOAT,0.0,
					     PARAM_INT32,2* 1,
					     PARAM_FLOATARRAY, &line_pnts[0],
					     PARAM_END);
	   break;
	 case BRUSH_PENCIL_TYPE:
	   return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					     PARAM_DRAWABLE, layer_ID,
					     PARAM_INT32,2,
					     PARAM_FLOATARRAY, &line_pnts[0],
					     PARAM_END);
	   break;
	 case BRUSH_AIRBRUSH_TYPE:
	   return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					     PARAM_DRAWABLE, layer_ID,
					     PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					     PARAM_INT32,2,
					     PARAM_FLOATARRAY, &line_pnts[0],
					     PARAM_END);
	   break;
	 case BRUSH_PATTERN_TYPE:
	   return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					     PARAM_DRAWABLE, layer_ID,
					     PARAM_DRAWABLE, layer_ID,
					     PARAM_INT32, 1,
					     PARAM_FLOAT,(gdouble)0.0,
					     PARAM_FLOAT,(gdouble)0.0,
					     PARAM_INT32,2,
					     PARAM_FLOATARRAY, &line_pnts[0],
					     PARAM_END);
	   break;
	 default:
	   break;
	 }

  gimp_destroy_params (return_vals, nreturn_vals);

  gimp_palette_set_background(bR,bG,bB);
  gimp_palette_set_foreground(fR,fG,fB);
  mygimp_brush_set(saved_bname);

  return(layer_ID);
}

static gint
PREFIXbrush_list_button_press(GtkWidget *widget,
			GdkEventButton *event,
			gpointer   data)
{
  gint32 layer_ID;

  BRUSHDESC *bdesc = (BRUSHDESC *)data;
  if((layer_ID = gfig_gen_brush_preview(bdesc)) != -1)
    {
      gtk_object_set_user_data ((( (brush_page_pw) != 0 &&( (((GtkObject*) (brush_page_pw))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (brush_page_pw))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (brush_page_pw) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (brush_page_pw), GTK_TYPE_OBJECT) ), (gpointer)bdesc);
      gfig_brush_fill_preview(brush_page_pw,layer_ID,bdesc);
      gtk_widget_draw(brush_page_pw,0);
    }

  return((0));
}




static void
PREFIXpaint_menu_callback (GtkWidget *widget, gpointer data)
{
  gint mtype = (gint)data;

  if(mtype == 1)
    {




      selvals.onlayers = (DRAWONLAYERS)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));

      if(selvals.onlayers == ORIGINAL_LAYER)
	gtk_widget_set_sensitive(page_menu_bg,(0));
      else
	gtk_widget_set_sensitive(page_menu_bg,(!(0)));
    }
  else if(mtype == 2)
    {




      selvals.onlayerbg = (DRAWLAYERBG)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));
    }
  else if(mtype == 3)
    {



      selvals.painttype = (PAINTTYPE)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));
      switch(selvals.painttype)
	{
	case PAINT_BRUSH_TYPE:
	  gtk_widget_set_sensitive(select_page_widget,(0));
	  gtk_widget_set_sensitive(brush_page_widget,(!(0)));
	  gtk_widget_set_sensitive(page_menu_layers,(!(0)));
	  if(selvals.onlayers == ORIGINAL_LAYER)
	    gtk_widget_set_sensitive(page_menu_bg,(0));
	  else
	    gtk_widget_set_sensitive(page_menu_bg,(!(0)));
	  break;
	case PAINT_SELECTION_TYPE:
	  gtk_widget_set_sensitive(select_page_widget,(!(0)));
	  gtk_widget_set_sensitive(brush_page_widget,(0));
	  gtk_widget_set_sensitive(page_menu_layers,(0));
	  gtk_widget_set_sensitive(page_menu_bg,(0));
	  break;
	case PAINT_SELECTION_FILL_TYPE:
	  gtk_widget_set_sensitive(select_page_widget,(!(0)));
	  gtk_widget_set_sensitive(brush_page_widget,(0));
	  gtk_widget_set_sensitive(page_menu_layers,(!(0)));
	  if(selvals.onlayers == ORIGINAL_LAYER)
	    gtk_widget_set_sensitive(page_menu_bg,(0));
	  else
	    gtk_widget_set_sensitive(page_menu_bg,(!(0)));
	  break;
	default:
	  break;
	}
    }
}

GtkWidget *
PREFIXpaint_page_menu_bgs(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Transparent");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) LAYER_TRANS_BG);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Background");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) LAYER_BG_BG);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("White");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) LAYER_WHITE_BG);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Copy");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) LAYER_COPY_BG);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}


GtkWidget *
PREFIXpaint_page_menu_type(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Brush");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) PAINT_BRUSH_TYPE);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)3);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Selection");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) PAINT_SELECTION_TYPE);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)3);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Selection+Fill");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) PAINT_SELECTION_FILL_TYPE);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)3);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);


  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}

GtkWidget *
PREFIXpaint_page_menu_layers(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Original");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) ORIGINAL_LAYER);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);


  menuitem = gtk_menu_item_new_with_label("New");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) SINGLE_LAYER);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Multiple");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) MULTI_LAYER);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) paint_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}

static GtkWidget *
PREFIXpaint_page()
{
  GtkWidget *table;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *label;
  GtkWidget *toggle;
  GtkWidget *page_menu_type;
  GtkWidget *scale_scale;
  GtkObject *scale_scale_data;

  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 4);

  table = gtk_table_new (6, 6,(0));
  gtk_table_set_row_spacings((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),6);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_container_get_type ()))))), 1);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), table,(!(0)),(!(0)), 0);
  gtk_widget_show(table);


  label = gtk_label_new ("Using:-");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show (label);

  page_menu_type = paint_page_menu_type();
  gtk_tooltips_set_tip(gfig_tooltips,page_menu_type,"Draw type. Either a brush or a selection. See brush page or selection page for more options",0);

  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), page_menu_type, 1, 2, 1, 2, GTK_FILL , GTK_FILL, 0, 0);

  label = gtk_label_new ("draw on:-");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 2, 3, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show (label);
  page_menu_layers = paint_page_menu_layers();
  gtk_tooltips_set_tip(gfig_tooltips,page_menu_layers,"Draw all objects on one layer (original or new) or one object per layer",0);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), page_menu_layers, 3, 4, 1, 2, GTK_FILL , GTK_FILL, 0, 0);
  if(gimp_drawable_channel(gfig_drawable))
      gtk_widget_set_sensitive(page_menu_layers,(0));


  label = gtk_label_new ("with BG of:-");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show (label);

  page_menu_bg = paint_page_menu_bgs();
  gtk_tooltips_set_tip(gfig_tooltips,page_menu_bg,"Layer background type. Copy causes previous layer to be copied before the draw is performed",0);

  gtk_widget_set_sensitive(page_menu_bg,(0));
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), page_menu_bg, 1, 2, 2, 3, GTK_FILL , GTK_FILL, 0, 0);


  toggle = gtk_check_button_new_with_label ("Reverse line ");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.reverselines);
  gtk_tooltips_set_tip(gfig_tooltips,toggle,"Draw lines in reverse order",0);
  gtk_widget_show(toggle);

  toggle = gtk_check_button_new_with_label ("Scale to image ");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 1, 2, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_toggle_button_set_active((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((toggle)), ((gtk_toggle_button_get_type ()))))),selvals.scaletoimage);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_scale2img_update,
		      (gpointer)&selvals.scaletoimage);
  gtk_tooltips_set_tip(gfig_tooltips,toggle,"Scale drawings to images size",0);
  gtk_widget_show(toggle);

  hbox = gtk_hbox_new ((0), 1);
  scale_scale_data = gtk_adjustment_new (1.0, 0.1, 5.0, 0.01, 0.01, 0.0);
  scale_scale = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_scale_data)), ((gtk_adjustment_get_type ()))))));
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), scale_scale,(!(0)),(!(0)), 0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_scale)), ((gtk_scale_get_type ()))))), GTK_POS_TOP);
  gtk_scale_set_digits((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_scale)), ((gtk_scale_get_type ()))))),2);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_scale)), ((gtk_range_get_type ()))))), GTK_UPDATE_CONTINUOUS);
  gtk_signal_connect ((( (scale_scale_data) != 0 &&( (((GtkObject*) (scale_scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (scale_scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (scale_scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (scale_scale_data), GTK_TYPE_OBJECT) ), "value_changed",
                      (GtkSignalFunc)gfig_scale_update_scale,
                      &selvals.scaletoimagefp);
  gtk_widget_show (scale_scale);
  gtk_widget_show (hbox);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 2, 4, 3, 4, GTK_FILL|GTK_EXPAND , GTK_FILL, 0, 0);
  gtk_widget_set_sensitive((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_scale)), ((gtk_widget_get_type ()))))),(0));
  gtk_object_set_user_data ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), (gpointer)scale_scale_data);
  gtk_object_set_user_data ((( (scale_scale_data) != 0 &&( (((GtkObject*) (scale_scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (scale_scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (scale_scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (scale_scale_data), GTK_TYPE_OBJECT) ), (gpointer)scale_scale);

  toggle = gtk_check_button_new_with_label ("Approx. Circles/Ellipses ");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 2, 4, 5, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.approxcircles);
  gtk_tooltips_set_tip(gfig_tooltips,toggle,"Approx. circles & ellipses using lines. Allows the use of brush fading with these types of objects.",0);
  gtk_widget_show(toggle);

  return(vbox);
}











































































































static void
PREFIXgfig_brush_invoker(gchar    *name,
		   gdouble  opacity,
		   gint     spacing,
		   gint     paint_mode,
		   gint     width,
		   gint     height,
		   gchar *  mask_data,
		   gint     closing,
		   gpointer udata)
{
  BRUSHDESC *bdesc = g_malloc0(sizeof(BRUSHDESC));

  bdesc->bpp = 3;
  bdesc->width = width;
  bdesc->height = height;
  bdesc->bname = g_strdup(name);

  brush_list_button_press(0,0,bdesc);

}

static gint
PREFIXselect_brush_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
   BRUSHDESC *bdesc = g_malloc0(sizeof(BRUSHDESC));
   gimp_interactive_selection_brush("Gfig brush selection",
				    mygimp_brush_get(),
				    1.0,
				    -1,
				    1,
				    gfig_brush_invoker,
 0);

   bdesc->bpp = 3;
   bdesc->bname = mygimp_brush_get();

   brush_list_button_press(0,0,bdesc);

   return((!(0)));
}

static GtkWidget *
PREFIXbrush_page()
{
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *pw;
  GtkWidget *scale;
  GtkObject *fade_out_scale_data;
  GtkObject *pressure_scale_data;
  GtkWidget *vbox;
  GtkWidget *button;
  BRUSHDESC *bdesc = g_malloc0(sizeof(BRUSHDESC));


  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 4);

  table = gtk_table_new (6, 6,(0));
  gtk_table_set_row_spacings((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),6);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_container_get_type ()))))), 1);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), table,(!(0)),(!(0)), 0);
  gtk_widget_show(table);










  fade_out_hbox = gtk_hbox_new ((0), 1);

  label = gtk_label_new ("Fade Out:");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((fade_out_hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show (label);

  fade_out_scale_data = gtk_adjustment_new (0.0, 0.0, 3000.0, 1.0, 1.0, 0.0);
  scale = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((fade_out_scale_data)), ((gtk_adjustment_get_type ()))))));
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((fade_out_hbox)), ((gtk_box_get_type ()))))), scale,(!(0)),(!(0)), 0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_scale_get_type ()))))), GTK_POS_TOP);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_range_get_type ()))))), GTK_UPDATE_DELAYED);
  gtk_signal_connect ((( (fade_out_scale_data) != 0 &&( (((GtkObject*) (fade_out_scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (fade_out_scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (fade_out_scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (fade_out_scale_data), GTK_TYPE_OBJECT) ), "value_changed",
                      (GtkSignalFunc)gfig_scale_update_fp,
                      &selvals.brushfade);
  gtk_widget_show (scale);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), fade_out_hbox, 1, 2, 0, 1, GTK_FILL|GTK_EXPAND , GTK_FILL, 0, 0);
  gtk_widget_show (fade_out_hbox);

  pressure_hbox = gtk_hbox_new ((0), 1);
  label = gtk_label_new ("Pressure:");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((pressure_hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show (label);

  pressure_scale_data = gtk_adjustment_new (20.0, 0.0, 100.0, 1.0, 1.0, 0.0);
  scale = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((pressure_scale_data)), ((gtk_adjustment_get_type ()))))));
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((pressure_hbox)), ((gtk_box_get_type ()))))), scale,(!(0)),(!(0)), 0);
  gtk_widget_show (scale);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_scale_get_type ()))))), GTK_POS_TOP);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_range_get_type ()))))), GTK_UPDATE_DELAYED);
  gtk_signal_connect ((( (pressure_scale_data) != 0 &&( (((GtkObject*) (pressure_scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (pressure_scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (pressure_scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (pressure_scale_data), GTK_TYPE_OBJECT) ), "value_changed",
                      (GtkSignalFunc)gfig_scale_update_fp,
                      &selvals.airbrushpressure);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), pressure_hbox, 1, 2, 0, 1, GTK_FILL|GTK_EXPAND , GTK_FILL, 0, 0);

  pencil_hbox = gtk_hbox_new ((0), 1);
  label = gtk_label_new ("No options...");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((pencil_hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show (label);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), pencil_hbox, 1, 2, 0, 1,GTK_FILL|GTK_EXPAND, GTK_FILL, 0, 0);



  pw = gfig_brush_preview(&brush_page_pw);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),pw, 0, 1, 0, 1, 0, 0, 0, 0);

  gtk_signal_connect ((( (pressure_scale_data) != 0 &&( (((GtkObject*) (pressure_scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (pressure_scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (pressure_scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (pressure_scale_data), GTK_TYPE_OBJECT) ), "value_changed",
                      (GtkSignalFunc)gfig_brush_update_preview,
                      (gpointer)brush_page_pw);

























  brush_sel_button = button = gtk_button_new_with_label ("Set brush...");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 0,4,1,5, 0, 0, 0, 0);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) select_brush_press,
 0);
  gtk_widget_show(button);


  bdesc->bpp = 3;
  bdesc->bname = mygimp_brush_get();
  brush_list_button_press(0,0,bdesc);

  return(vbox);
}

static void
PREFIXselect_menu_callback (GtkWidget *widget, gpointer data)
{
  gint mtype = (gint)data;

  if(mtype == 1)
    {
      SELECTION_TYPE type =
	(SELECTION_TYPE)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));

      selopt.type = type;
    }
  else if(mtype == 2)
    {
      ARC_TYPE type =
	(ARC_TYPE)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));

      selopt.as_pie = type;
    }
  else if(mtype == 3)
    {
      FILL_TYPE type =
	(FILL_TYPE)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));

      selopt.fill_type = type;
    }
  else if(mtype == 4)
    {
      FILL_WHEN type =
	(FILL_WHEN)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));
      selopt.fill_when = type;
    }
}

GtkWidget *
PREFIXselect_page_menu_fill_when(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Each selection");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) FILL_EACH);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)4);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("All selections");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) FILL_AFTER);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)4);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}



GtkWidget *
PREFIXselect_page_menu_fill_type(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Pattern");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) FILL_PATTERN);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)3);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Foreground");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) FILL_FOREGROUND);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)3);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Background");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) FILL_BACKGROUND);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)3);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}


GtkWidget *
PREFIXselect_page_menu_type(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Add");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) ADD);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Sub");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) SUBTRACT);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Replace");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) REPLACE);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Intersect");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) INTERSECT);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}

GtkWidget *
PREFIXselect_page_menu_arctype(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Segment");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) ARC_SEGMENT);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Sector");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) ARC_SECTOR);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) select_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}


static GtkWidget *
PREFIXselect_page()
{
  GtkWidget *menu;
  GtkWidget *label;
  GtkWidget *toggle;
  GtkWidget *hbox;
  GtkWidget *scale;
  GtkObject *scale_data;
  GtkWidget *table;
  GtkWidget *vbox;

  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 4);

  table = gtk_table_new (7, 7,(0));
  gtk_table_set_row_spacings((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),6);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_container_get_type ()))))), 1);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), table,(!(0)),(!(0)), 0);
  gtk_widget_show(table);


#line 3824 "gfig.c"



  hbox = gtk_hbox_new ((0), 1);
  label = gtk_label_new ("Selection type:");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  menu = select_page_menu_type();
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), menu,(!(0)),(!(0)), 0);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);  gtk_widget_show (hbox);



  toggle = gtk_check_button_new_with_label ("Antialiasing");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selopt.antia);
  gtk_widget_show(toggle);


  toggle = gtk_check_button_new_with_label ("Feather");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selopt.feather);
  gtk_widget_show(toggle);


  hbox = gtk_hbox_new ((0), 1);

  label = gtk_label_new ("Feather Radius:");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show (label);

  scale_data = gtk_adjustment_new (selopt.feather_radius, 0.0, 100.0, 1.0, 1.0, 0.0);
  scale = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_data)), ((gtk_adjustment_get_type ()))))));
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), scale,(!(0)),(!(0)), 0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_scale_get_type ()))))), GTK_POS_TOP);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_range_get_type ()))))), GTK_UPDATE_DELAYED);
  gtk_signal_connect ((( (scale_data) != 0 &&( (((GtkObject*) (scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (scale_data), GTK_TYPE_OBJECT) ), "value_changed",
                      (GtkSignalFunc)gfig_scale_update_fp,
                      &selopt.feather_radius);
  gtk_widget_show (scale);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 1, 3, 1, 2, GTK_FILL|GTK_EXPAND , GTK_FILL, 0, 0);
  gtk_widget_show (hbox);


  hbox = gtk_hbox_new ((0), 1);
  label = gtk_label_new ("Fill type:");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  menu = select_page_menu_fill_type();
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), menu,(!(0)),(!(0)), 0);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 0, 0);  gtk_widget_show (hbox);


  hbox = gtk_hbox_new ((0), 1);

  label = gtk_label_new ("Fill Opacity:");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show (label);

  scale_data = gtk_adjustment_new (selopt.fill_opacity, 0.0, 100.0, 1.0, 1.0, 0.0);
  scale = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale_data)), ((gtk_adjustment_get_type ()))))));
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), scale,(!(0)),(!(0)), 0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_scale_get_type ()))))), GTK_POS_TOP);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((scale)), ((gtk_range_get_type ()))))), GTK_UPDATE_DELAYED);
  gtk_signal_connect ((( (scale_data) != 0 &&( (((GtkObject*) (scale_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (scale_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (scale_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (scale_data), GTK_TYPE_OBJECT) ), "value_changed",
                      (GtkSignalFunc)gfig_scale_update_fp,
                      &selopt.fill_opacity);
  gtk_widget_show (scale);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 1, 3, 2, 3, GTK_FILL|GTK_EXPAND , GTK_FILL, 0, 0);
  gtk_widget_show (hbox);


  hbox = gtk_hbox_new ((0), 1);
  label = gtk_label_new ("Fill after: ");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  menu = select_page_menu_fill_when();
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), menu,(!(0)),(!(0)), 0);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 0, 1, 4, 5, GTK_FILL, GTK_FILL, 0, 0);  gtk_widget_show (hbox);


  hbox = gtk_hbox_new ((0), 1);
  label = gtk_label_new ("Arc as: ");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  menu = select_page_menu_arctype();
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), menu,(!(0)),(!(0)), 0);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), hbox, 0, 1, 5, 6, GTK_FILL, GTK_FILL, 0, 0);  gtk_widget_show (hbox);

  return(vbox);
}

static void
PREFIXgridtype_menu_callback (GtkWidget *widget, gpointer data)
{
  gint mtype = (gint)data;

  if(mtype == 1)
    {










      selvals.opts.gridtype = (GRIDTYPE)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));
    }
  else
    {
      grid_gc_type = (gint)gtk_object_get_user_data ((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));
    }

  draw_grid_clear(widget,0);
}

GtkWidget *
PREFIXoption_page_menu_gridtype(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  gfig_opt_widget.gridtypemenu = option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Rectangle");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) RECT_GRID);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Polar");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) POLAR_GRID);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Isometric");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer) ISO_GRID);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
		          (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)1);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}

GtkWidget *
PREFIXoption_page_menu_gridrender(void)
{
  GtkWidget	*option_menu;
  GtkWidget	*menu;
  GtkWidget	*menuitem;

  option_menu = gtk_option_menu_new ();

  menu = gtk_menu_new();

  menuitem = gtk_menu_item_new_with_label("Normal");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)GTK_STATE_NORMAL);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Black");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)-2);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("White");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)-3);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Grey");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)-4);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Darker");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)GTK_STATE_ACTIVE);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Lighter");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)GTK_STATE_PRELIGHT);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  menuitem = gtk_menu_item_new_with_label("Very Dark");
  gtk_object_set_user_data ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), (gpointer)GTK_STATE_SELECTED);
  gtk_signal_connect ((( (menuitem) != 0 &&( (((GtkObject*) (menuitem))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menuitem))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menuitem) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menuitem), GTK_TYPE_OBJECT) ), "activate",
			  (GtkSignalFunc) gridtype_menu_callback,
		      (gpointer)2);
  gtk_widget_show (menuitem);
  gtk_menu_append ((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((menu)), ((gtk_menu_get_type ()))))), menuitem);

  gtk_option_menu_set_menu ((((GtkOptionMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((option_menu)), ((gtk_option_menu_get_type ()))))), menu);
  gtk_widget_show (option_menu);

  return option_menu;
}

static GtkWidget *
PREFIXoptions_page()
{
  GtkWidget *table;
  GtkWidget *menu;
  GtkWidget *toggle;
  GtkWidget *slider;
  GtkWidget *entry;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *vbox;
  GtkObject *size_data;
  char buf[256];

  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 4);

  table = gtk_table_new (6, 6,(0));
  gtk_table_set_row_spacings((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),6);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_container_get_type ()))))), 1);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), table,(!(0)),(!(0)), 0);
  gtk_widget_show(table);


  toggle = gtk_check_button_new_with_label ("Show image");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.showimage);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) toggle_show_image,
		      (gpointer)1);
  gtk_widget_show(toggle);

  button = gtk_button_new_with_label ("Reload image");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 1, 2, 0, 1, 0, 0, 0, 0);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) reload_button_press,
 0);
  gtk_widget_show(button);

  toggle = gtk_check_button_new_with_label ("Hide cntr pnts ");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.opts.showcontrol);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) toggle_show_image,
		      (gpointer)1);

  sprintf(buf,"Grid type:");
  label = gtk_label_new (buf);
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 2, 3, 0, 1, GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_widget_show (label);
  menu = option_page_menu_gridtype();
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), menu, 3, 4, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

  sprintf(buf,"Grid Colour:");
  label = gtk_label_new (buf);
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 2, 3, 1, 2, GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_widget_show (label);
  menu = option_page_menu_gridrender();
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), menu, 3, 4, 1, 2, GTK_FILL, GTK_FILL, 0, 0);


  gtk_widget_show(toggle);
  gfig_opt_widget.showcontrol = toggle;

  sprintf(buf,"Max Undo:");
  label = gtk_label_new (buf);
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 0, 1, 2, 3, GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_widget_show (label);

  size_data = gtk_adjustment_new (selvals.maxundo,1,10,5, 1, 0);
  slider = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((size_data)), ((gtk_adjustment_get_type ()))))));
  gtk_widget_set_usize (slider,120, 0);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), slider, 1, 4, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_scale_get_type ()))))), GTK_POS_LEFT);
  gtk_scale_set_digits ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_scale_get_type ()))))), 0);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_range_get_type ()))))),GTK_UPDATE_CONTINUOUS );
  gtk_signal_connect ((( (size_data) != 0 &&( (((GtkObject*) (size_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (size_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (size_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (size_data), GTK_TYPE_OBJECT) ), "value_changed",
		      (GtkSignalFunc) gfig_scale_update,
		      &selvals.maxundo);
  gtk_widget_show (slider);

  entry = gtk_entry_new();
  gtk_object_set_user_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), size_data);
  gtk_object_set_user_data(size_data, entry);
  gtk_widget_set_usize(entry,28, 0);
  sprintf(buf, "%d", selvals.maxundo);
  gtk_entry_set_text((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((entry)), ((gtk_entry_get_type ()))))), buf);
  gtk_signal_connect((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), "changed",
		     (GtkSignalFunc) gfig_entry_update,
		     &selvals.maxundo);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), entry, 4, 5, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(entry);

  toggle = gtk_check_button_new_with_label ("Show tool tips ");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.showtooltips);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) toggle_tooltips,
		      (gpointer)&selvals.showtooltips);
  gtk_toggle_button_set_active((((GtkToggleButton *) gtk_type_check_object_cast ((GtkTypeObject*) ((toggle)), ((gtk_toggle_button_get_type ()))))),selvals.showtooltips);
  gtk_widget_show(toggle);

  toggle = gtk_check_button_new_with_label ("Show pos");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 1, 2, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.showpos);
  gtk_signal_connect_after ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_pos_enable,
		      (gpointer)1);
  gtk_widget_show(toggle);

  button = gtk_button_new_with_label ("About");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 3, 4, 4, 5, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) about_button_press,
 0);
  gtk_widget_show(button);

  return(vbox);
}

static GtkWidget *
PREFIXgrid_frame()
{
  GtkWidget *frame;
  GtkWidget *table;
  GtkWidget *toggle;
  GtkWidget *label;
  GtkWidget *slider;
  GtkObject *size_data;
  GtkWidget *entry;

  char buf[256];

  frame = gtk_frame_new ("Grid");
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), 1);

  table = gtk_table_new (7, 7,(0));
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), table);

  toggle = gtk_check_button_new_with_label ("Snap to grid");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 1, 2, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.opts.snap2grid);
  gtk_widget_show(toggle);
  gfig_opt_widget.snap2grid = toggle;

  toggle = gtk_check_button_new_with_label ("Display grid");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.opts.drawgrid);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) draw_grid_clear,
		      (gpointer)1);
  gtk_widget_show(toggle);
  gfig_opt_widget.drawgrid = toggle;


  toggle = gtk_check_button_new_with_label ("Lock on grid");
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), toggle, 2, 3, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL |GTK_EXPAND, 0, 0);
  gtk_signal_connect ((( (toggle) != 0 &&( (((GtkObject*) (toggle))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (toggle))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (toggle) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (toggle), GTK_TYPE_OBJECT) ), "toggled",
		      (GtkSignalFunc) gfig_toggle_update,
		      (gpointer)&selvals.opts.lockongrid);

  gfig_opt_widget.lockongrid = toggle;

  label = gtk_label_new ("Grid spacing ");
  gtk_misc_set_alignment ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.5);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 0, 1, 4, 5, GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_widget_show (label);

  size_data = gtk_adjustment_new (selvals.opts.gridspacing,10,50, (50 + 10)/2, 1, 0);
  slider = gtk_hscale_new ((((GtkAdjustment *) gtk_type_check_object_cast ((GtkTypeObject*) ((size_data)), ((gtk_adjustment_get_type ()))))));
  gtk_widget_set_usize (slider,120, 0);
  gtk_table_attach ((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), slider, 1,3 , 4, 5, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_scale_set_value_pos ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_scale_get_type ()))))), GTK_POS_LEFT);
  gtk_scale_set_digits ((((GtkScale *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_scale_get_type ()))))), 0);
  gtk_range_set_update_policy ((((GtkRange *) gtk_type_check_object_cast ((GtkTypeObject*) ((slider)), ((gtk_range_get_type ()))))),GTK_UPDATE_CONTINUOUS );
  gtk_signal_connect ((( (size_data) != 0 &&( (((GtkObject*) (size_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (size_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (size_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (size_data), GTK_TYPE_OBJECT) ), "value_changed",
		      (GtkSignalFunc) gfig_scale_update,
		      &selvals.opts.gridspacing);
  gtk_signal_connect ((( (size_data) != 0 &&( (((GtkObject*) (size_data))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (size_data))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (size_data) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (size_data), GTK_TYPE_OBJECT) ), "value_changed",
		      (GtkSignalFunc) draw_grid_clear,
		      (gpointer)0);
  gtk_widget_show (slider);
  gfig_opt_widget.gridspacing = size_data;

  entry = gtk_entry_new();
  gtk_object_set_user_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), size_data);
  gtk_object_set_user_data(size_data, entry);
  gtk_widget_set_usize(entry,28, 0);
  sprintf(buf, "%d", selvals.opts.gridspacing);
  gtk_entry_set_text((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((entry)), ((gtk_entry_get_type ()))))), buf);
  gtk_signal_connect((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), "changed",
		     (GtkSignalFunc) gfig_entry_update,
		     &selvals.opts.gridspacing);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), entry, 3, 4, 4, 5, GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(entry);
  gtk_widget_show(table);
  gtk_widget_show(frame);

  return(frame);
}

void
PREFIXclear_list_items(GtkList *list)
{
  gtk_list_clear_items(list,0,-1);
}

void
PREFIXbuild_list_items(GtkWidget *list)
{
  GList *tmp = gfig_list;
  GtkWidget *list_item;
  GtkWidget *list_pix;
  GFIGOBJ *g;

  while(tmp)
    {
      g = tmp->data;

      if(g->obj_status & 0x2)
	list_pix = gfig_new_pixmap(list,mini_cross_xpm);
      else
	list_pix = gfig_new_pixmap(list,blank_xpm);

      list_item = gfig_list_item_new_with_label_and_pixmap(g,g->draw_name,list_pix);
      gtk_object_set_user_data ((( (list_item) != 0 &&( (((GtkObject*) (list_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (list_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (list_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (list_item), GTK_TYPE_OBJECT) ), (gpointer)g);
      gtk_list_append_items ((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list)), ((gtk_list_get_type ()))))), g_list_append(0,list_item));

      gtk_signal_connect((( (list_item) != 0 &&( (((GtkObject*) (list_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (list_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (list_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (list_item), GTK_TYPE_OBJECT) ), "button_press_event",
			     (GtkSignalFunc) list_button_press,
			     (gpointer)g);
      gtk_widget_show (list_item);

      tmp = tmp->next;
    }
}

static GtkWidget *
PREFIXadd_objects_list ()
{
  GtkWidget *vbox;
  GtkWidget *table;
  GtkWidget *frame;
  GtkWidget *list_frame;
  GtkWidget *scrolled_win;
  GtkWidget *list;
  GtkWidget *button;

  frame = gtk_frame_new("Object");
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
  gtk_widget_show(frame);

  table = gtk_table_new (5, 4,(0));
  gtk_widget_show(table);

  delete_frame_to_freeze = list_frame = gtk_frame_new(0);
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
  gtk_widget_show(list_frame);

  scrolled_win = gtk_scrolled_window_new (0,0);
  gtk_scrolled_window_set_policy ((((GtkScrolledWindow *) gtk_type_check_object_cast ((GtkTypeObject*) ((scrolled_win)), ((gtk_scrolled_window_get_type ()))))),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_frame)), ((gtk_container_get_type ()))))), scrolled_win);
  gtk_widget_show (scrolled_win);

  gfig_gtk_list = list = gtk_list_new ();

  gtk_list_set_selection_mode ((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list)), ((gtk_list_get_type ()))))), GTK_SELECTION_BROWSE);
  gtk_scrolled_window_add_with_viewport ((((GtkScrolledWindow *) gtk_type_check_object_cast ((GtkTypeObject*) ((scrolled_win)), ((gtk_scrolled_window_get_type ()))))),						 list);
  gtk_widget_show (list);


  gfig_list_load_all(gfig_path_list);


  build_list_items(list);


  button = gtk_button_new_with_label ("Rescan");
  gtk_widget_show(button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) rescan_button_press,
 0);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Select directory and rescan Gfig object collection",0);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 2, 3, 0, 1, GTK_FILL, GTK_FILL,  0, 0);

  button = gtk_button_new_with_label ("Load");
  gtk_widget_show(button);
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) load_button_press,
		      list);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Load a single Gfig object collection",0);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 2, 3, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

  button = gtk_button_new_with_label ("New");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) new_button_press,
		      "New gfig obj");
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Create a new Gfig object collection for editing",0);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 2, 3, 2, 3, GTK_FILL, GTK_FILL, 0, 0);


  button = gtk_button_new_with_label ("Delete");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "button_press_event",
		      (GtkSignalFunc) gfig_delete_gfig_callback,
		      (gpointer)list);
  gtk_widget_show(button);
  gtk_tooltips_set_tip(gfig_tooltips,button,"Delete currently selected Gfig Object collection",0);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), button, 2, 3, 3, 4, GTK_FILL, GTK_FILL, 0, 0);



  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), list_frame, 1, 2, 0, 4, GTK_FILL|GTK_EXPAND , GTK_FILL|GTK_EXPAND, 1, 1);

  vbox = small_preview(list);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), vbox, 0, 1, 0, 4, 0, 0, 0, 0);

  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), table);
  return (frame);
}

static void
PREFIXgfig_pos_enable(GtkWidget *widget, gpointer data)
{
  gint enable = selvals.showpos;
  gtk_widget_set_sensitive((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((x_pos_label)), ((gtk_widget_get_type ()))))),enable);
  gtk_widget_set_sensitive((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((y_pos_label)), ((gtk_widget_get_type ()))))),enable);
}

static void
PREFIXgfig_pos_update_labels(gpointer data)
{
  static gchar buf[256];


  pos_tag = -1;

  if(x_pos_val < 0)
    sprintf(buf," X:%.3d ",x_pos_val);
  else
    sprintf(buf," X:  %.3d ",x_pos_val);

  gtk_label_set_text((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((x_pos_label)), ((gtk_label_get_type ()))))),buf);

  if(y_pos_val < 0)
    sprintf(buf," Y:%.3d ",y_pos_val);
  else
    sprintf(buf," Y:  %.3d ",y_pos_val);

  gtk_label_set_text((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((y_pos_label)), ((gtk_label_get_type ()))))),buf);
}

static void
PREFIXgfig_pos_update(gint x , gint y)
{
  gint update;

  if(x_pos_val != x || y_pos_val != y)
    update = 1;
  else
    update = 0;

  x_pos_val = x;
  y_pos_val = y;

  if(update && pos_tag == -1 && selvals.showpos)
    {

      gfig_pos_update_labels(0);
    }
}








































static GtkWidget *
PREFIXgfig_pos_labels(void)
{
  GtkWidget *label;
  GtkWidget *hbox;
  GtkWidget *vbox;
  gchar buf[256];

  hbox = gtk_hbox_new ((0), 0);
  vbox = gtk_vbox_new ((0), 0);


  label = gtk_label_new("XY Pos:- ");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))),label,(0),(0), 0);
  gtk_widget_show(label);

  x_pos_label = gtk_label_new("");
  gtk_widget_show(x_pos_label);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), x_pos_label);

  y_pos_label = gtk_label_new("");
  gtk_widget_show(y_pos_label);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), y_pos_label);

 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_container_get_type ()))))), 1);

  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), vbox,(0),(0), 0);
  gtk_widget_show(hbox);
  gtk_widget_show(vbox);

  sprintf(buf," X:  %.3d ",0);
  gtk_label_set_text((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((x_pos_label)), ((gtk_label_get_type ()))))),buf);
  sprintf(buf," Y:  %.3d ",0);
  gtk_label_set_text((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((y_pos_label)), ((gtk_label_get_type ()))))),buf);

  return(hbox);
}

static GtkWidget *
PREFIXmake_pos_info(void)
{
  GtkWidget * xframe;
  GtkWidget * hbox;
  GtkWidget * label;

  xframe = gtk_frame_new("Obj Details");
  hbox = gtk_hbox_new ((!(0)), 1);

  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((xframe)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((xframe)), ((gtk_container_get_type ()))))), hbox);


  label = gfig_pos_labels();
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gfig_pos_enable(0,0);






  gtk_widget_show(hbox);
  gtk_widget_show(xframe);
  return(xframe);
}


static GtkWidget *
PREFIXmake_status(void)
{
  GtkWidget * xframe;
  GtkWidget * table;
  GtkWidget * label;

  xframe = gtk_frame_new("Collection Details");

  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((xframe)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
  table = gtk_table_new (6, 6,(0));

  label = gtk_label_new("Draw name:");
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_widget_show(label);
  gtk_label_set_justify((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_label_get_type ()))))),GTK_JUSTIFY_RIGHT);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 1, 2, 0, 1, 0 , GTK_FILL, 0, 0);


  label = gtk_label_new("Filename:");
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_widget_show(label);
  gtk_label_set_justify((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_label_get_type ()))))),GTK_JUSTIFY_RIGHT);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), label, 1, 2, 1, 2, 0 , GTK_FILL, 0, 0);

  status_label_dname = gtk_label_new("<None>");
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_widget_show(status_label_dname);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), status_label_dname, 2, 4, 0, 1, GTK_FILL|GTK_EXPAND, 0, 0, 0);

  status_label_fname = gtk_label_new("<None>");
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_widget_show(status_label_fname);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), status_label_fname, 2, 4, 1, 2, GTK_FILL|GTK_EXPAND, 0, 0, 0);












  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((xframe)), ((gtk_container_get_type ()))))), table);

  gtk_widget_show(table);
  gtk_widget_show(xframe);
  return(xframe);
}

static GtkWidget *
PREFIXmake_preview(void)
{
  GtkWidget * xframe;
  GtkWidget * vbox;
  GtkWidget * hbox;
  GtkWidget * table;
  GtkWidget * ruler;


  gfig_preview = gtk_preview_new(GTK_PREVIEW_COLOR);
  gtk_widget_set_events((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_preview)), ((gtk_widget_get_type ()))))),GDK_EXPOSURE_MASK | GDK_MOTION_NOTIFY | GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_MOTION_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK );
  gfig_preview_exp_id =
    gtk_signal_connect_after((( (gfig_preview) != 0 &&( (((GtkObject*) (gfig_preview))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (gfig_preview))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (gfig_preview) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (gfig_preview), GTK_TYPE_OBJECT) ), "expose_event",
			      (GtkSignalFunc) gfig_preview_expose,
 0);

  gtk_signal_connect((( (gfig_preview) != 0 &&( (((GtkObject*) (gfig_preview))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (gfig_preview))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (gfig_preview) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (gfig_preview), GTK_TYPE_OBJECT) ), "event",
		      (GtkSignalFunc) gfig_preview_events,
 0);


  gtk_preview_size((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_preview)), ((gtk_preview_get_type ()))))), preview_width, preview_height);

  xframe = gtk_frame_new(0);

  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((xframe)), ((gtk_frame_get_type ()))))), GTK_SHADOW_IN);

  table = gtk_table_new (3, 3,(0));
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), gfig_preview, 1, 2, 1, 2, GTK_FILL ,GTK_FILL , 0, 0);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((xframe)), ((gtk_container_get_type ()))))), table);

  ruler = gtk_hruler_new ();
  gtk_ruler_set_range ((((GtkRuler *) gtk_type_check_object_cast ((GtkTypeObject*) ((ruler)), ((gtk_ruler_get_type ()))))), 0, preview_width, 0,650);
  gtk_signal_connect_object ((( (gfig_preview) != 0 &&( (((GtkObject*) (gfig_preview))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (gfig_preview))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (gfig_preview) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (gfig_preview), GTK_TYPE_OBJECT) ), "motion_notify_event",
			     (GtkSignalFunc)(((GtkWidgetClass *) gtk_type_check_class_cast ((GtkTypeClass*) (((( (ruler) != 0 &&( (((GtkObject*) (ruler))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (ruler))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (ruler) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (ruler), GTK_TYPE_OBJECT) )->klass)), ((gtk_widget_get_type ())))))->motion_notify_event,
(( (ruler) != 0 &&( (((GtkObject*) (ruler))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (ruler))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (ruler) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (ruler), GTK_TYPE_OBJECT) ));
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),ruler, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0,0);
  gtk_widget_show(ruler);

  ruler = gtk_vruler_new ();
  gtk_ruler_set_range ((((GtkRuler *) gtk_type_check_object_cast ((GtkTypeObject*) ((ruler)), ((gtk_ruler_get_type ()))))), 0, preview_height, 0,650);
  gtk_signal_connect_object ((( (gfig_preview) != 0 &&( (((GtkObject*) (gfig_preview))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (gfig_preview))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (gfig_preview) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (gfig_preview), GTK_TYPE_OBJECT) ), "motion_notify_event",
			     (GtkSignalFunc)(((GtkWidgetClass *) gtk_type_check_class_cast ((GtkTypeClass*) (((( (ruler) != 0 &&( (((GtkObject*) (ruler))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (ruler))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (ruler) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (ruler), GTK_TYPE_OBJECT) )->klass)), ((gtk_widget_get_type ())))))->motion_notify_event,
(( (ruler) != 0 &&( (((GtkObject*) (ruler))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (ruler))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (ruler) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (ruler), GTK_TYPE_OBJECT) ));
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),ruler, 0, 1, 1, 2, GTK_FILL , GTK_FILL, 0,0);
  gtk_widget_show(ruler);


  gtk_widget_show(xframe);
  gtk_widget_show(table);

  vbox = gtk_vbox_new ((0), 0);
  hbox = gtk_hbox_new ((0), 0);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), xframe,(0),(0), 0);

 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 2);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), hbox,(0),(0), 0);

  xframe = make_pos_info();
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), xframe,(!(0)),(!(0)), 0);

  xframe = make_status();
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), xframe,(!(0)),(!(0)), 0);

  gtk_widget_show(vbox);
  gtk_widget_show(hbox);

  return(vbox);
}










static void
PREFIXgfig_grid_colours(GtkWidget *w,GdkColormap *cmap)
{
  GdkGCValues values;
  GdkColor new_col1;
  GdkColor new_col2;
  unsigned char stipple[8] =
  {
    0xAA,
    0x55,
    0xAA,
    0x55,
    0xAA,
    0x55,
    0xAA,
    0x55,
  };

  gdk_color_parse("gray50",&new_col1);
  gdk_color_alloc(xxx,&new_col1);
  gdk_color_parse("gray80",&new_col2);
  gdk_color_alloc(xxx,&new_col2);
  values.background.pixel = new_col1.pixel;
  values.foreground.pixel = new_col2.pixel;
  values.fill = GDK_OPAQUE_STIPPLED;
  values.stipple = gdk_bitmap_create_from_data (w->window, (char*) stipple, 4, 4);
  grid_hightlight_drawgc = gdk_gc_new_with_values (w->window, &values,
						   GDK_GC_FOREGROUND |
						   GDK_GC_BACKGROUND |
						   GDK_GC_FILL |
						   GDK_GC_STIPPLE);
}


static gint
PREFIXgfig_dialog ()
{
  GtkWidget *button;
  GtkWidget *frame;
  GtkWidget *xframe;
  GtkWidget *oframe;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *notebook;
  GtkWidget *page;
  GtkWidget *top_level_dlg;
  GdkColor color;

  guchar     *color_cube;
  int argc;
  char ** argv;

  argc = 1;
  argv =((gchar * *) g_malloc ((unsigned) sizeof (gchar *) * (1)));
  argv[0] = g_strdup ("gfig");

  gtk_init (&argc, &argv);
  gtk_rc_parse (gimp_gtkrc ());



  plug_in_parse_gfig_path();



  gtk_preview_set_gamma(gimp_gamma());
  gtk_preview_set_install_cmap(gimp_install_cmap());
  color_cube = gimp_color_cube();
  gtk_preview_set_color_cube(color_cube[0], color_cube[1], color_cube[2], color_cube[3]);

  gtk_widget_set_default_visual(yyy = gtk_preview_get_visual());

  gtk_widget_set_default_colormap(xxx = gtk_preview_get_cmap());



  img_width  = gimp_drawable_width(gfig_select_drawable->id);
  img_height = gimp_drawable_height(gfig_select_drawable->id);


  top_level_dlg = gtk_dialog_new ();
  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (top_level_dlg), ((gtk_window_get_type ()))))), "Gfig");
 gtk_window_set_position ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (top_level_dlg), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);
  gtk_signal_connect ((( (top_level_dlg) != 0 &&( (((GtkObject*) (top_level_dlg))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (top_level_dlg))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (top_level_dlg) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (top_level_dlg), GTK_TYPE_OBJECT) ), "destroy",
		      (GtkSignalFunc) gfig_close_callback,
 0);


  gfig_tooltips = gtk_tooltips_new();


  button = gtk_button_new_with_label ("Done");
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) gfig_ok_callback,
                      top_level_dlg);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Paint");
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) gfig_paint_callback,
                      top_level_dlg);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  save_button = button = gtk_button_new_with_label ("Save");
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) save_button_press,
                      top_level_dlg);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Clear");
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) gfig_clear_callback,
                      top_level_dlg);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show (button);


  undo_widget = button = gtk_button_new_with_label ("Undo");
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) gfig_undo_callback,
                      top_level_dlg);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_set_sensitive(button,(0));
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_signal_connect((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
			     (GtkSignalFunc) gfig_cancel_callback,
			     top_level_dlg);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show (button);


  frame = gtk_frame_new ("preview");
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), 1);
  table = gtk_table_new (6, 6,(0));
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_container_get_type ()))))), 1);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), table);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((top_level_dlg)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), frame,(!(0)),(!(0)), 0);


  xframe = make_preview();
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), xframe, 1, 2, 0, 2, GTK_FILL, GTK_FILL, 0, 0);

  gtk_widget_show(table);
  gtk_widget_show(frame);
  gtk_widget_show(gfig_preview);


  xframe = draw_buttons(top_level_dlg);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), xframe, 0,1, 0, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(xframe);

  frame = gtk_frame_new ("Settings");
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), 1);

  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), frame, 2, 3, 0, 2, GTK_FILL , GTK_FILL, 0, 0);
  table = gtk_table_new (7, 7,(0));

 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_container_get_type ()))))), 1);
  gtk_table_set_row_spacings((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))),1);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((frame)), ((gtk_container_get_type ()))))), table);


  oframe = add_objects_list();
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), oframe, 0,6, 0 , 1, GTK_EXPAND|GTK_FILL, GTK_FILL, 0, 0);



  xframe = grid_frame();
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), xframe, 0,6, 3, 4, GTK_EXPAND|GTK_FILL, GTK_FILL, 0, 0);


  notebook = gtk_notebook_new();
  gtk_notebook_set_tab_pos((((GtkNotebook *) gtk_type_check_object_cast ((GtkTypeObject*) ((notebook)), ((gtk_notebook_get_type ()))))), GTK_POS_TOP);
  gtk_widget_show(notebook);
  gtk_table_attach((((GtkTable *) gtk_type_check_object_cast ((GtkTypeObject*) ((table)), ((gtk_table_get_type ()))))), notebook, 0, 6, 5, 6, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND, 0, 0);

  page = paint_page();
  label = gtk_label_new("Paint");
  gtk_widget_show(label);
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_notebook_append_page((((GtkNotebook *) gtk_type_check_object_cast ((GtkTypeObject*) ((notebook)), ((gtk_notebook_get_type ()))))), page, label);
  gtk_widget_show(page);

  brush_page_widget = brush_page();
  label = gtk_label_new("Brush");
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_widget_show(label);
  gtk_notebook_append_page((((GtkNotebook *) gtk_type_check_object_cast ((GtkTypeObject*) ((notebook)), ((gtk_notebook_get_type ()))))), brush_page_widget, label);
  gtk_widget_show(brush_page_widget);



#line 4911 "gfig.c"
  select_page_widget = select_page();
  label = gtk_label_new("Select");
  gtk_widget_show(label);
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_notebook_append_page((((GtkNotebook *) gtk_type_check_object_cast ((GtkTypeObject*) ((notebook)), ((gtk_notebook_get_type ()))))), select_page_widget, label);
  gtk_widget_show(select_page_widget);
  gtk_widget_set_sensitive(select_page_widget,(0));


  page = options_page();
  label = gtk_label_new("Options");
  gtk_widget_show(label);
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))),0.5,0.5);
  gtk_notebook_append_page((((GtkNotebook *) gtk_type_check_object_cast ((GtkTypeObject*) ((notebook)), ((gtk_notebook_get_type ()))))), page, label);
  gtk_widget_show(page);


  gtk_widget_show(frame);
  gtk_widget_show(table);

  gtk_widget_show (top_level_dlg);
  dialog_update_preview();
  gfig_new_gc();
  gfig_update_stat_labels();


  gfig_grid_colours(gfig_preview,xxx);

  gfig_op_menu_create(top_level_dlg);


  if(gdk_color_parse("bisque",&color) ==(0) || !gdk_color_alloc(xxx,&color))
    gtk_tooltips_set_colors(gfig_tooltips,&top_level_dlg->style->white,&top_level_dlg->style->black);
  else
    gtk_tooltips_set_colors(gfig_tooltips,&color,&top_level_dlg->style->black);



  gtk_main ();
  gdk_flush ();

  return gfig_run;
}

static void
PREFIXgfig_close_callback (GtkWidget *widget,
			 gpointer   data)
{
  gfig_brush_img_del();
  gtk_main_quit ();
}

static void
PREFIXdone_ok_window(GtkWidget * widget,
		    gpointer   data)
{
  gfig_run =(!(0));
  gtk_widget_destroy ((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((data)), ((gtk_widget_get_type ()))))));
}

void
PREFIXdone_warn_dialog (GtkWidget *w,gint count)
{
  GtkWidget *window = 0;
  GtkWidget *label;
  GtkWidget *button;
  gchar buf[128];

  window = gtk_dialog_new ();

  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), "Warning");
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_container_get_type ()))))), 0);

  button = gtk_button_new_with_label ("OK");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) done_ok_window,
                      (gpointer)w);

 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) ok_warn_window,
                      (gpointer)window);

 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show (button);

  label = gtk_label_new("Unsaved Gfig objects - continue with exiting?");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 10, 10);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), label,(!(0)),(!(0)), 0);
  gtk_widget_show (label);

  sprintf(buf,"Number objects unsaved = %d\n",count);
  label = gtk_label_new(buf);
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 10, 10);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), label,(!(0)),(!(0)), 0);
  gtk_widget_show (label);
  gtk_widget_show (window);
}

static void
PREFIXgfig_ok_callback (GtkWidget *widget,
		      gpointer   data)
{

  GList * list;
  GFIGOBJ * gfig;
  gint count = 0;

  list = gfig_list;
  while (list)
    {
      gfig = (GFIGOBJ *) list->data;
      if(gfig->obj_status & 0x1)
	count++;
      list = list->next;
    }

  if(count)
    done_warn_dialog((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((data)), ((gtk_widget_get_type ()))))),count);
  else
    {
      gfig_run =(!(0));
      gtk_widget_destroy ((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((data)), ((gtk_widget_get_type ()))))));
    }
  gfig_brush_img_del();
}

static void
PREFIXgfig_cancel_callback(GtkWidget *widget,
		      gpointer   data)
{
  gtk_widget_destroy((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((data)), ((gtk_widget_get_type ()))))));
  gfig_brush_img_del();
}



static void
PREFIXupdate_draw_area(GtkWidget *widget,GdkEvent *event)
{
  if (!((((((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) )->flags)) & GTK_VISIBLE) != 0) &&(((((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) )->flags)) & GTK_MAPPED) != 0)))
    return;

  gtk_signal_handler_block((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ),gfig_preview_exp_id);
  gtk_widget_draw(widget,0);
  gtk_signal_handler_unblock((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ),gfig_preview_exp_id );

  draw_grid(widget,0);
  draw_objects(current_obj->obj_list,(!(0)));
}

static gint
PREFIXgfig_preview_expose( GtkWidget *widget,
			    GdkEvent *event )
{
  GdkCursor *preview_cursor;
  static gint changed_cursor = 0;

  if(!changed_cursor && gfig_preview->window)
    {
      changed_cursor = 1;
      preview_cursor = gdk_cursor_new(GDK_CROSSHAIR);
      gdk_window_set_cursor(gfig_preview->window,preview_cursor);
    }
  update_draw_area(widget,event);
  return(0);
}

static gint
PREFIXpic_preview_expose( GtkWidget *widget,
			    GdkEvent *event )
{
  if(pic_obj)
    {
      drawing_pic =(!(0));
      draw_objects(pic_obj->obj_list,(0));
      drawing_pic =(0);
    }
  return(0);
}

static gint
PREFIXadjust_pic_coords(gint coord,gint ratio)
{

  static gint pratio = -1;

  if(pratio == -1)
    {
      pratio =(((preview_width) > (preview_height)) ? (preview_width) : (preview_height));
    }

  return((48 * coord)/pratio);
}

static gint
PREFIXgfig_preview_events ( GtkWidget *widget,
			     GdkEvent *event )
{
  GdkEventButton *bevent;
  GdkEventMotion *mevent;
  GdkPoint point;
  static gint tmp_show_single = 0;

  switch (event->type)
    {
    case GDK_EXPOSE:
      break;

    case GDK_BUTTON_PRESS:
      bevent = (GdkEventButton *) event;
      point.x = bevent->x;
      point.y = bevent->y;

 if (1){ if (!(need_to_scale == 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 5131,"need_to_scale == 0"); }else (void)0;


      if(selvals.otype >= MOVE_OBJ)
	{
	  if(!selvals.scaletoimage)
	    {
	      point.x = gfig_invscale_x(point.x);
	      point.y = gfig_invscale_y(point.y);
	    }
	  object_operation_start(&point,bevent->state & GDK_SHIFT_MASK);


	  if(selvals.opts.snap2grid)
	    {

	      if(bevent->button == 3)
		{
		  find_grid_pos(&point,&point,(0));
		}
	    }
	}
      else
	{
	  if(selvals.opts.snap2grid)
	    {
	      if(bevent->button == 3)
		{
		  find_grid_pos(&point,&point,(0));
		}
	      else
		{
		  find_grid_pos(&point,&point,(0));
		}
	    }
	  object_start(&point,bevent->state & GDK_SHIFT_MASK);
	}

      break;
    case GDK_BUTTON_RELEASE:
      bevent = (GdkEventButton *) event;
      point.x = bevent->x;
      point.y = bevent->y;


      if(selvals.opts.snap2grid)
	find_grid_pos(&point,&point,bevent->button == 3);


      if(selvals.otype >= MOVE_OBJ)
	{
	  if(!selvals.scaletoimage)
	    {
	      point.x = gfig_invscale_x(point.x);
	      point.y = gfig_invscale_y(point.y);
	    }
	  object_operation_end(&point,bevent->state & GDK_SHIFT_MASK);
	}
      else
	{
	  if(obj_creating)
	    {
	      object_end(&point,bevent->state & GDK_SHIFT_MASK);
	    }
	  else
	    break;
	}


      list_button_update(current_obj);

      break;
    case GDK_MOTION_NOTIFY:

      mevent = (GdkEventMotion *) event;
      point.x = mevent->x;
      point.y = mevent->y;

      if(selvals.opts.snap2grid)
	find_grid_pos(&point,&point,mevent->state & GDK_BUTTON3_MASK);

      if(selvals.otype >= MOVE_OBJ)
	{

	  if(!selvals.scaletoimage)
	    {
	      point.x = gfig_invscale_x(point.x);
	      point.y = gfig_invscale_y(point.y);
	    }
	  object_operation(&point,mevent->state & GDK_SHIFT_MASK);
	  gfig_pos_update(point.x,point.y);
	  return(0);
	}

      if(obj_creating)
	{
	  object_update(&point);
	}
      gfig_pos_update(point.x,point.y);
      break;
    case GDK_KEY_PRESS:
      if((tmp_show_single = obj_show_single) != -1)
	{
	  obj_show_single = -1;
	  draw_grid_clear(0,0);
	}
      break;
    case GDK_KEY_RELEASE:
      if(tmp_show_single != -1)
	{
	  obj_show_single = tmp_show_single;
	  draw_grid_clear(0,0);
	}
      break;
    default:
      break;
    }
  return(0);
}



#line 5256 "gfig.c"

static GtkWidget *
PREFIXgfig_list_add(GFIGOBJ *obj)
{
  GList *list;
  gint pos;
  GtkWidget *list_item;
  GtkWidget *list_pix;

  list_pix = gfig_new_pixmap(gfig_gtk_list,Floppy6_xpm);
  list_item = gfig_list_item_new_with_label_and_pixmap(obj,obj->draw_name,list_pix);

  gtk_object_set_user_data ((( (list_item) != 0 &&( (((GtkObject*) (list_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (list_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (list_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (list_item), GTK_TYPE_OBJECT) ), (gpointer)obj);

  pos = gfig_list_insert(obj);

  list = g_list_append(0, list_item);
  gtk_list_insert_items((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))), list, pos);
  gtk_widget_show (list_item);
  gtk_list_select_item((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))), pos);

  gtk_signal_connect((( (list_item) != 0 &&( (((GtkObject*) (list_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (list_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (list_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (list_item), GTK_TYPE_OBJECT) ), "button_press_event",
		     (GtkSignalFunc) list_button_press,
		     (gpointer)obj);

  return(list_item);
}

static void
PREFIXgfig_list_ok_callback (GtkWidget *w,
		       gpointer   client_data)
{
  GfigListOptions *options;
  GtkWidget *list;
  gint pos;

  options = (GfigListOptions *) client_data;
  list = options->list_entry;





  if (options->obj->draw_name)
    {
      g_free(options->obj->draw_name);
    }
  options->obj->draw_name = g_strdup (gtk_entry_get_text ((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((options->name_entry)), ((gtk_entry_get_type ())))))));







  pos = gtk_list_child_position((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))),list);




  gtk_list_clear_items((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))),pos,pos+1);


  gfig_list = g_list_remove(gfig_list,options->obj);
  gfig_list_add(options->obj);

  options->obj->obj_status |= 0x1;

  gtk_widget_destroy (options->query_box);
  g_free (options);

  gfig_update_stat_labels();
}

static void
PREFIXgfig_list_cancel_callback (GtkWidget *w,
				  gpointer   client_data)
{
  GfigListOptions *options;

  options = (GfigListOptions *) client_data;
  if(options->created)
    {

      delete_button_press_ok(w,gfig_gtk_list);
    }

  gtk_widget_destroy (options->query_box);
  g_free (options);
}

static void
PREFIXgfig_dialog_edit_list (GtkWidget *lwidget,GFIGOBJ *obj,gint created)
{
  GfigListOptions *options;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *button;
  GtkWidget *label;


  options = (GfigListOptions *) g_malloc (sizeof (GfigListOptions));
  options->list_entry = lwidget;
  options->obj = obj;
  options->created = created;


  options->query_box = gtk_dialog_new ();
  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (options->query_box), ((gtk_window_get_type ()))))), "Edit Gfig entry name");
 gtk_window_set_position ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (options->query_box), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);


  vbox = gtk_vbox_new ((0), 1);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 2);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((options->query_box)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), vbox,(!(0)),(!(0)), 0);


  hbox = gtk_hbox_new ((0), 1);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), hbox,(0),(0), 0);
  label = gtk_label_new ("Gfig object name:");
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show (label);
  options->name_entry = gtk_entry_new ();
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), options->name_entry,(!(0)),(!(0)), 0);
  gtk_entry_set_text ((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((options->name_entry)), ((gtk_entry_get_type ()))))),obj->draw_name);

  gtk_widget_show (options->name_entry);
  gtk_widget_show (hbox);

  button = gtk_button_new_with_label ("OK");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc)gfig_list_ok_callback,
                      options);
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((options->query_box)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc)gfig_list_cancel_callback,
                      options);
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((options->query_box)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show (button);

  gtk_widget_show (vbox);
  gtk_widget_show (options->query_box);
}

static void
PREFIXgfig_rescan_cancel_callback (GtkWidget *w,
				  gpointer   client_data)
{
  gtk_widget_destroy ((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((client_data)), ((gtk_widget_get_type ()))))));
}

static void
PREFIXgfig_rescan_ok_callback (GtkWidget *w,
		       gpointer   client_data)
{
  GList *list;

  list = rescan_list;
  while (list)
    {



      list = list->next;
    }
  list = gfig_path_list;
  while (list)
    {



      rescan_list = g_list_append(rescan_list,g_strdup(list->data));
      list = list->next;
    }
  clear_list_items((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))));
  gfig_list_load_all(rescan_list);
  build_list_items(gfig_gtk_list);
  list_button_update(current_obj);
  gtk_widget_destroy ((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((client_data)), ((gtk_widget_get_type ()))))));
}

static void
PREFIXgfig_rescan_file_selection_ok(GtkWidget *w,
		   GtkFileSelection *fs,
		   gpointer data)
{
  GtkWidget *list_item;
  GtkWidget *lw = (GtkWidget *)gtk_object_get_user_data((( (fs) != 0 &&( (((GtkObject*) (fs))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (fs))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (fs) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (fs), GTK_TYPE_OBJECT) ));
  gchar * filenamebuf;
  struct stat	filestat;
  gint		err;

  filenamebuf = gtk_file_selection_get_filename ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((fs)), ((gtk_file_selection_get_type ()))))));

  err = stat(filenamebuf, &filestat);

  if (!(((filestat.st_mode)&0xF000) == 0x4000))
    {
     g_warning("Entry %.100s is not a directory\n",filenamebuf);
    }
  else
    {

      list_item = gtk_list_item_new_with_label(filenamebuf);
      gtk_widget_show(list_item);

      gtk_list_prepend_items((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((lw)), ((gtk_list_get_type ()))))),g_list_append(0, list_item));

      rescan_list = g_list_prepend(rescan_list,g_strdup(filenamebuf));
    }

  gtk_widget_destroy((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((fs)), ((gtk_widget_get_type ()))))));
}

static void
PREFIXgfig_rescan_add_entry_callback (GtkWidget *w,
		       gpointer   client_data)
{
  static GtkWidget *window = 0;


  window = gtk_file_selection_new ("Add Gfig path");
 gtk_window_set_position ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);
  gtk_object_set_user_data((( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ),(gpointer)client_data);


  gtk_signal_connect ((( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ), "destroy",
		      (GtkSignalFunc) destroy_window,
		      &window);

  gtk_signal_connect ((( ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button) != 0 &&( (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button), GTK_TYPE_OBJECT) ),
		      "clicked", (GtkSignalFunc) gfig_rescan_file_selection_ok,
		      (gpointer)window);

  gtk_signal_connect_object((( ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button) != 0 &&( (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button), GTK_TYPE_OBJECT) ),
			    "clicked", (GtkSignalFunc) gtk_widget_destroy,
(( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ));
  gtk_widget_show(window);
}

static void
PREFIXgfig_rescan_list (void)
{
  GtkWidget *vbox;
  GtkWidget *button;
  GtkWidget *dlg;
  GtkWidget *list_frame;
  GtkWidget *scrolled_win;
  GtkWidget *list_widget;
  GList *list;


  dlg = gtk_dialog_new ();
  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (dlg), ((gtk_window_get_type ()))))), "Rescan for Gfig objects");
 gtk_window_set_position ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (dlg), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);


  vbox = gtk_vbox_new ((0), 1);
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 2);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((dlg)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), vbox,(!(0)),(!(0)), 0);


  list_frame = gtk_frame_new(0);
  gtk_frame_set_shadow_type ((((GtkFrame *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_frame)), ((gtk_frame_get_type ()))))), GTK_SHADOW_ETCHED_IN);
  gtk_widget_show(list_frame);

  scrolled_win = gtk_scrolled_window_new (0,0);
  gtk_scrolled_window_set_policy ((((GtkScrolledWindow *) gtk_type_check_object_cast ((GtkTypeObject*) ((scrolled_win)), ((gtk_scrolled_window_get_type ()))))),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_frame)), ((gtk_container_get_type ()))))), scrolled_win);
  gtk_widget_show (scrolled_win);

  list_widget = gtk_list_new ();
  gtk_list_set_selection_mode ((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_widget)), ((gtk_list_get_type ()))))), GTK_SELECTION_BROWSE);
  gtk_scrolled_window_add_with_viewport ((((GtkScrolledWindow *) gtk_type_check_object_cast ((GtkTypeObject*) ((scrolled_win)), ((gtk_scrolled_window_get_type ()))))),
					 list_widget);
  gtk_widget_show (list_widget);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), list_frame,(!(0)),(!(0)), 0);

  list = gfig_path_list;
  while (list)
    {
      GtkWidget *list_item;
      list_item = gtk_list_item_new_with_label(list->data);
      gtk_widget_show(list_item);
      gtk_container_add ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((list_widget)), ((gtk_container_get_type ()))))), list_item);
      list = list->next;
    }

  button = gtk_button_new_with_label ("OK");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc)gfig_rescan_ok_callback,
                      (gpointer)dlg);
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);


  if((list = rescan_list))
    {
      while (list)
	{
	  g_free(list->data);
	  list = list->next;
	}

      g_list_free(rescan_list);
      rescan_list = 0;
    }

  button = gtk_button_new_with_label ("Add Dir");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc)gfig_rescan_add_entry_callback,
                      (gpointer)list_widget);

  gtk_object_set_user_data((( (dlg) != 0 &&( (((GtkObject*) (dlg))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (dlg))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (dlg) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (dlg), GTK_TYPE_OBJECT) ),(gpointer)list_widget);


 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc)gfig_rescan_cancel_callback,
                      (gpointer)dlg);
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((dlg)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_show (button);

  gtk_widget_show (vbox);
  gtk_widget_show (dlg);
}


void
PREFIXlist_button_update(GFIGOBJ *obj)
{
 if (1){ if (!(obj != 0)) { g_log (((gchar*) 0), G_LOG_LEVEL_CRITICAL, "file %s: line %d: assertion `%s' failed.", "gfig.c", 5611,"obj != NULL"); return; }; }else (void)0;
  pic_obj = (GFIGOBJ *)obj;
  gtk_widget_draw(pic_preview,0);
  drawing_pic =(!(0));
  draw_objects(pic_obj->obj_list,(0));
  drawing_pic =(0);
}


static void
PREFIXgfig_load_file_selection_ok(GtkWidget *w,
		   GtkFileSelection *fs,
		   gpointer data)
{
  gchar * filename;
  struct stat	filestat;
  gint		err;
  GFIGOBJ * gfig;
  GFIGOBJ * current_saved;

  filename = gtk_file_selection_get_filename ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((fs)), ((gtk_file_selection_get_type ()))))));





  err = stat (filename, &filestat);

  if (!err &&(((filestat.st_mode)&0xF000) == 0x8000))
    {

#line 5644 "gfig.c"
      current_saved = current_obj;
      current_obj = 0;
      gfig = gfig_load (filename,filename);
      current_obj = current_saved;

      if (gfig)
	{

	  if(access(filename,2))
	    gfig->obj_status |= 0x2;

	  gfig_list_add(gfig);
	  new_obj_2edit(gfig);
	}
    }

  gtk_widget_destroy((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((fs)), ((gtk_widget_get_type ()))))));
}

static gint
PREFIXload_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  static GtkWidget *window = 0;


  window = gtk_file_selection_new ("Load Gfig obj");
 gtk_window_set_position ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);



  gtk_signal_connect ((( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ), "destroy",
		      (GtkSignalFunc) destroy_window,
		      &window);

  gtk_signal_connect ((( ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button) != 0 &&( (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->ok_button), GTK_TYPE_OBJECT) ),
		      "clicked", (GtkSignalFunc) gfig_load_file_selection_ok,
		      (gpointer)window);

  gtk_signal_connect_object((( ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button) != 0 &&( (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((((GtkFileSelection *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_file_selection_get_type ())))))->cancel_button), GTK_TYPE_OBJECT) ),
			    "clicked", (GtkSignalFunc) gtk_widget_destroy,
(( (window) != 0 &&( (((GtkObject*) (window))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (window))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (window) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (window), GTK_TYPE_OBJECT) ));
  gtk_widget_show(window);

  return((0));
}
















static gint32
PREFIXmygimp_layer_copy (gint32 layer_ID)
{
  GParam *return_vals;
  int nreturn_vals;

  return_vals = gimp_run_procedure ("gimp_layer_copy",
                                    &nreturn_vals,
                                    PARAM_LAYER, layer_ID,
				    PARAM_INT32, 0,
                                    PARAM_END);

  layer_ID = -1;
  if (return_vals[0].data.d_status == STATUS_SUCCESS)
    layer_ID = return_vals[1].data.d_layer;

  gimp_destroy_params (return_vals, nreturn_vals);

  return layer_ID;
}

static void
PREFIXpaint_layer_copy(gchar *new_name)
{
  gint32 old_drawable = gfig_drawable;
  if((gfig_drawable = mygimp_layer_copy(gfig_drawable)) < 0)
    {
      g_warning("Error in copy layer for onlayers\n");
      gfig_drawable = old_drawable;
      return;
    }

  gimp_layer_set_name(gfig_drawable,new_name);
  gimp_image_add_layer(gfig_image,gfig_drawable,-1);
}

static void
PREFIXpaint_layer_new(gchar *new_name)
{
  gint32 layer_id;
  gint32 fill_type;
  int isgrey = 0;

  switch ( gimp_drawable_type (gfig_select_drawable->id) )
  {
  case GRAYA_IMAGE:
  case GRAY_IMAGE:
    isgrey = 2;
  default:
    break;
  }

  if((layer_id = gimp_layer_new(gfig_image,
				new_name,
				img_width,
				img_height,
				1 + isgrey,
				100.0,
				0 )) < 0)
    printf("Error in creating\n");
  else
    {
      gimp_image_add_layer(gfig_image,layer_id,-1);
      gimp_drawable_fill(layer_id,1);
    }

  gfig_drawable = layer_id;

  switch(selvals.onlayerbg)
    {
    case LAYER_TRANS_BG:
      fill_type = 2;
      break;
    case LAYER_BG_BG:
      fill_type = 0;
      break;
    case LAYER_WHITE_BG:
      fill_type = 1;
      break;
    case LAYER_COPY_BG:
    default:
      fill_type = 1;
      g_warning("Paint layer new internal error %d\n",selvals.onlayerbg);
      break;
    }

#line 5795 "gfig.c"

  gimp_drawable_fill(layer_id,fill_type);

}

static void
PREFIXpaint_layer_fill()
{
  GParam *return_vals;
  int nreturn_vals;

  return_vals = gimp_run_procedure ("gimp_bucket_fill",
                                    &nreturn_vals,
				    PARAM_DRAWABLE, gfig_drawable,
				    PARAM_INT32, selopt.fill_type,
				    PARAM_INT32, 0,
				    PARAM_FLOAT, (gdouble)selopt.fill_opacity,
				    PARAM_FLOAT, (gdouble)0.0,
				    PARAM_INT32, 0,
				    PARAM_FLOAT, (gdouble)0.0,
				    PARAM_FLOAT, (gdouble)0.0,
                                    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);
}

static void
PREFIXgfig_paint_callback(GtkWidget *widget,
		  gpointer   data)
{
  DALLOBJS * objs;
  gint layer_count = 0;
  gchar buf[128];
  gint count;
  gint ccount = 0;
  BRUSHDESC *bdesc;

  objs = current_obj->obj_list;

  count = gfig_obj_counts(objs);





  bdesc = gtk_object_get_user_data ((( (brush_page_pw) != 0 &&( (((GtkObject*) (brush_page_pw))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (brush_page_pw))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (brush_page_pw) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (brush_page_pw), GTK_TYPE_OBJECT) ));

  if(bdesc)
    mygimp_brush_set(bdesc->bname);

  while(objs)
    {

      if(ccount == obj_show_single || obj_show_single == -1)
	{
	  sprintf(buf,"Gfig Layer %d",layer_count++);

	  if(selvals.painttype != PAINT_SELECTION_TYPE)
	    {
	      switch(selvals.onlayers)
		{
		case SINGLE_LAYER:
		  if(layer_count == 1)
		    {
		      if(selvals.onlayerbg == LAYER_COPY_BG)
			paint_layer_copy(buf);
		      else
			paint_layer_new(buf);
		    }
		  break;
		case MULTI_LAYER:
		  if(selvals.onlayerbg == LAYER_COPY_BG)
		    paint_layer_copy(buf);
		  else
		    paint_layer_new(buf);
		  break;
		case ORIGINAL_LAYER:

		  break;
		default:
		  g_warning("Error in onlayers val %d\n",selvals.onlayers);
		  break;
		}
	    }

	  objs->obj->paintfunc(objs->obj);


	  if(selvals.painttype == PAINT_SELECTION_FILL_TYPE
	     && selopt.fill_when == FILL_EACH)
	    paint_layer_fill();
	}

      objs = objs->next;

      ccount++;




    }


  if(selvals.painttype == PAINT_SELECTION_FILL_TYPE
     && selopt.fill_when == FILL_AFTER)
    paint_layer_fill();

  gimp_displays_flush();
}

static gint
PREFIXreload_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  refill_cache();
  draw_grid_clear(widget,data);

  return((0));
}

static gint
PREFIXabout_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{

  GtkWidget *window = 0;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *hbox;
  GtkWidget *vbox;
  GtkWidget *pm;

  window = gtk_dialog_new ();

  gtk_window_set_title ((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (window), ((gtk_window_get_type ()))))), "About");
 gtk_container_set_border_width ((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_container_get_type ()))))), 0);

  button = gtk_button_new_with_label ("OK");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) ok_warn_window,
                      window);

 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);


  pm = gfig_new_pixmap(window,rulers_comp_xpm);
  gtk_widget_show(pm);

  hbox = gtk_hbox_new((0),1);
  gtk_widget_show(hbox);

  vbox = gtk_vbox_new((0),1);
  gtk_widget_show(vbox);

  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), pm,(!(0)),(!(0)), 0);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((hbox)), ((gtk_box_get_type ()))))), vbox,(!(0)),(!(0)), 0);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((window)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), hbox,(!(0)),(!(0)), 0);

  label = gtk_label_new("Gfig - GIMP plug-in");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 2, 2);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  label = gtk_label_new("Release 1.3");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 2, 2);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  label = gtk_label_new("Andy Thomas");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 2, 2);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  label = gtk_label_new("Email alt@picnic.demon.co.uk");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 2, 2);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  label = gtk_label_new("http://www.picnic.demon.co.uk/");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 2, 2);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  label = gtk_label_new("Isometric grid By Rob Saunders");
  gtk_misc_set_padding ((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 2, 2);
  gtk_widget_show (label);
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);

  gtk_widget_show (window);

  return((0));
}


static gint
PREFIXsave_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  gfig_save();

  return((0));
}

static gint
PREFIXrescan_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  gfig_rescan_list();
  return((0));
}

static GtkWidget *
PREFIXnew_gfig_obj(gchar * name)
{
  GFIGOBJ * gfig;
  GtkWidget * new_list_item;


  gfig = gfig_new();

  if(!name)
    name = "New gfig obj";

  gfig->draw_name = g_strdup(name);


  pic_obj = current_obj = gfig;

  new_list_item = gfig_list_add(gfig);

  tmp_bezier = obj_creating = tmp_line = 0;


  update_draw_area(gfig_preview,0);
  list_button_update(gfig);
  return(new_list_item);
}

static gint
PREFIXnew_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  GtkWidget * new_list_item;

  new_list_item = new_gfig_obj((gchar*)data);
  gfig_dialog_edit_list(new_list_item,current_obj,(!(0)));

  return((0));
}

static gint
PREFIXdelete_button_press_cancel(GtkWidget *widget,
		  gpointer   data)
{
  gtk_widget_destroy(delete_dialog);
  gtk_widget_set_sensitive(delete_frame_to_freeze,(!(0)));
  delete_dialog = 0;

  return((0));
}

static gint
PREFIXdelete_button_press_ok(GtkWidget *widget,
		  gpointer   data)
{
  gint pos;
  GList * sellist;
  GFIGOBJ * sel_obj;
  GtkWidget *list = (GtkWidget *)data;








  sellist =(((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list)), ((gtk_list_get_type ())))))->selection;

  sel_obj = (GFIGOBJ *)gtk_object_get_user_data((( ((GtkWidget *)(sellist->data)) != 0 &&( (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((GtkWidget *)(sellist->data)) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((GtkWidget *)(sellist->data)), GTK_TYPE_OBJECT) ));

  pos = gtk_list_child_position((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))),sellist->data);





  gtk_list_clear_items((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))),pos,pos+1);

  gfig_list = g_list_remove(gfig_list,sel_obj);

  if(sel_obj == current_obj)
    {
      clear_undo();
    }


  gfig_free_everything(sel_obj);



  pos--;

  if(pos < 0 && g_list_length(gfig_list) == 0)
    {

#line 6114 "gfig.c"
      new_button_press(0,0,0);
      pos = 0;
    }

  if(delete_dialog)
  {
    gtk_widget_destroy(delete_dialog);
    gtk_widget_set_sensitive(delete_frame_to_freeze,(!(0)));
  }

  delete_dialog = 0;

  gtk_list_select_item((((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_gtk_list)), ((gtk_list_get_type ()))))), pos);

  current_obj = g_list_nth(gfig_list,pos)->data;

  update_draw_area(gfig_preview,0);

  list_button_update(current_obj);

  gfig_update_stat_labels();

  return((0));
}

static gint
PREFIXgfig_delete_gfig_callback(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  GtkWidget *vbox;
  GtkWidget *label;
  GtkWidget *button;
  char      *str;
  GtkWidget *list = (GtkWidget *)data;
  GList * sellist;
  GFIGOBJ * sel_obj;


  sellist =(((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list)), ((gtk_list_get_type ())))))->selection;

  sel_obj = (GFIGOBJ *)gtk_object_get_user_data((( ((GtkWidget *)(sellist->data)) != 0 &&( (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((GtkWidget *)(sellist->data)) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((GtkWidget *)(sellist->data)), GTK_TYPE_OBJECT) ));
  if(delete_dialog)
    return((0));

  delete_dialog = gtk_dialog_new();
  gtk_window_set_title((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (delete_dialog), ((gtk_window_get_type ()))))), "Delete gfig drawing");
 gtk_window_set_position((((GtkWindow *) gtk_type_check_object_cast ((GtkTypeObject*) (delete_dialog), ((gtk_window_get_type ()))))), GTK_WIN_POS_MOUSE);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((delete_dialog)), ((gtk_container_get_type ()))))), 0);

  vbox = gtk_vbox_new((0), 0);
 gtk_container_set_border_width((((GtkContainer *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_container_get_type ()))))), 8);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((delete_dialog)), ((gtk_dialog_get_type ())))))->vbox)), ((gtk_box_get_type ()))))), vbox,
(0),(0), 0);
  gtk_widget_show(vbox);



  label = gtk_label_new("Are you sure you want to delete");
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.0);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show(label);

  str = g_malloc((strlen(sel_obj->draw_name) + 32 * sizeof(char)));

  sprintf(str, "\"%s\" from the list and from disk?", sel_obj->draw_name);

  label = gtk_label_new(str);
  gtk_misc_set_alignment((((GtkMisc *) gtk_type_check_object_cast ((GtkTypeObject*) ((label)), ((gtk_misc_get_type ()))))), 0.0, 0.0);
  gtk_box_pack_start((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) ((vbox)), ((gtk_box_get_type ()))))), label,(0),(0), 0);
  gtk_widget_show(label);

  g_free(str);


  button = gtk_button_new_with_label ("Delete");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) delete_button_press_ok,
                      data);
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((delete_dialog)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_object_set_user_data((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ),widget);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect ((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ), "clicked",
                      (GtkSignalFunc) delete_button_press_cancel,
                      data);
 if (1){ ((((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) )->flags)) |= (GTK_CAN_DEFAULT)); }else (void)0;
  gtk_box_pack_start ((((GtkBox *) gtk_type_check_object_cast ((GtkTypeObject*) (((((GtkDialog *) gtk_type_check_object_cast ((GtkTypeObject*) ((delete_dialog)), ((gtk_dialog_get_type ())))))->action_area)), ((gtk_box_get_type ()))))), button,(!(0)),(!(0)), 0);
  gtk_widget_grab_default (button);
  gtk_object_set_user_data((( (button) != 0 &&( (((GtkObject*) (button))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (button))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (button) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (button), GTK_TYPE_OBJECT) ),widget);
  gtk_widget_show (button);



  gtk_widget_set_sensitive((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((delete_frame_to_freeze)), ((gtk_widget_get_type ()))))),(0));
  gtk_widget_show(delete_dialog);

  return((0));
}

static void
PREFIXgfig_update_stat_labels()
{
  gchar str[45];

  if(current_obj->draw_name)
    sprintf(str,"%.34s",current_obj->draw_name);
  else
    sprintf(str,"<NONE>");

  gtk_label_set_text((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((status_label_dname)), ((gtk_label_get_type ()))))),str);

  if(current_obj->filename)
    {
      gint slen;
      gchar *hm = getenv("HOME");
      gchar *dfn = g_strdup(current_obj->filename);

      if(!strncmp(dfn,hm,strlen(hm)-1))
	 {
	   strcpy(dfn,"~");
	   strcat(dfn,&dfn[strlen(hm)]);
	 }
      if((slen = strlen(dfn)) > 40)
	{
	  strncpy(str,dfn,19);
	  str[19] = '\0';
	  strcat(str,"...");
	  strncat(str,&dfn[slen - 21],19);
	  str[40] ='\0';
	}
      else
	sprintf(str,"%.40s",dfn);
      g_free(dfn);
    }
  else
    sprintf(str,"<NONE>");

  gtk_label_set_text((((GtkLabel *) gtk_type_check_object_cast ((GtkTypeObject*) ((status_label_fname)), ((gtk_label_get_type ()))))),str);

}

static void
PREFIXnew_obj_2edit(GFIGOBJ *obj)
{
  GFIGOBJ * old_current = current_obj;





  clear_undo();


  current_obj = obj;


  obj_show_single = -1;


  update_options(old_current);


#line 6282 "gfig.c"

  gfig_update_stat_labels();


  update_draw_area(gfig_preview,0);

  list_button_update(current_obj);

  if(obj->obj_status & 0x2)
    {
      create_warn_dialog("Editing read-only object - you will not be able to save it");
      gtk_widget_set_sensitive(save_button,(0));
    }
  else
    {
      gtk_widget_set_sensitive(save_button,(!(0)));
    }
}

static gint
PREFIXedit_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer data)
{
  GList * sellist;
  GFIGOBJ * sel_obj;
  GtkWidget *list = (GtkWidget *)data;








  sellist =(((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list)), ((gtk_list_get_type ())))))->selection;

  sel_obj = (GFIGOBJ *)gtk_object_get_user_data((( ((GtkWidget *)(sellist->data)) != 0 &&( (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((GtkWidget *)(sellist->data)) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((GtkWidget *)(sellist->data)), GTK_TYPE_OBJECT) ));

  if(sel_obj)
    new_obj_2edit(sel_obj);
  else
    g_warning("Internal error - list item has null object!");

  return((0));
}

static gint
PREFIXmerge_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  GList * sellist;
  GFIGOBJ * sel_obj;
  DALLOBJS * obj_copies;
  GtkWidget *list = (GtkWidget *)data;








  sellist =(((GtkList *) gtk_type_check_object_cast ((GtkTypeObject*) ((list)), ((gtk_list_get_type ())))))->selection;

  sel_obj = (GFIGOBJ *)gtk_object_get_user_data((( ((GtkWidget *)(sellist->data)) != 0 &&( (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) ((GtkWidget *)(sellist->data)))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) ((GtkWidget *)(sellist->data)) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) ((GtkWidget *)(sellist->data)), GTK_TYPE_OBJECT) ));
  if(sel_obj && sel_obj->obj_list && sel_obj != current_obj)
    {

      obj_copies = copy_all_objs(sel_obj->obj_list);
      prepend_to_all_obj(current_obj,obj_copies);


      update_draw_area(gfig_preview,0);

      list_button_update(current_obj);
    }
  return((0));
}


static void
PREFIXgfig_save_menu_callback(GtkWidget *widget, gpointer data)
{
  GFIGOBJ * real_current = current_obj;



  current_obj = gfig_obj_for_menu;

  gfig_save();

  current_obj = real_current;
}

static void
PREFIXgfig_edit_menu_callback(GtkWidget *widget, gpointer data)
{
  new_obj_2edit(gfig_obj_for_menu);
}

static void
PREFIXgfig_rename_menu_callback(GtkWidget *widget, gpointer data)
{
  create_file_selection(gfig_obj_for_menu,gfig_obj_for_menu->filename);
}

static void
PREFIXgfig_copy_menu_callback(GtkWidget *widget, gpointer data)
{

  gchar *new_name = g_malloc(strlen(gfig_obj_for_menu->draw_name) + 6);

  sprintf(new_name,"%s copy",gfig_obj_for_menu->draw_name);
  new_gfig_obj(new_name);
  g_free(new_name);


  current_obj->obj_list = copy_all_objs(gfig_obj_for_menu->obj_list);
  current_obj->opts = gfig_obj_for_menu->opts;


  update_draw_area(gfig_preview,0);

  list_button_update(current_obj);
}

static void
PREFIXgfig_op_menu_create(GtkWidget *window)
{
  GtkWidget *menu_item;




  gfig_op_menu = gtk_menu_new();








  save_menu_item = menu_item = gtk_menu_item_new_with_label("Save");
  gtk_menu_append((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_op_menu)), ((gtk_menu_get_type ()))))),menu_item);
  gtk_widget_show(menu_item);

  gtk_signal_connect((( (menu_item) != 0 &&( (((GtkObject*) (menu_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menu_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menu_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menu_item), GTK_TYPE_OBJECT) ),"activate",
		     (GtkSignalFunc)gfig_save_menu_callback,
 0);







  menu_item = gtk_menu_item_new_with_label("Save as...");
  gtk_menu_append((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_op_menu)), ((gtk_menu_get_type ()))))),menu_item);
  gtk_widget_show(menu_item);
  gtk_signal_connect((( (menu_item) != 0 &&( (((GtkObject*) (menu_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menu_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menu_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menu_item), GTK_TYPE_OBJECT) ),"activate",
		     (GtkSignalFunc)gfig_rename_menu_callback,
 0);







  menu_item = gtk_menu_item_new_with_label("Copy");
  gtk_menu_append((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_op_menu)), ((gtk_menu_get_type ()))))),menu_item);
  gtk_widget_show(menu_item);
  gtk_signal_connect((( (menu_item) != 0 &&( (((GtkObject*) (menu_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menu_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menu_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menu_item), GTK_TYPE_OBJECT) ),"activate",
		     (GtkSignalFunc)gfig_copy_menu_callback,
 0);







  menu_item = gtk_menu_item_new_with_label("Edit");
  gtk_menu_append((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_op_menu)), ((gtk_menu_get_type ()))))),menu_item);
  gtk_widget_show(menu_item);
  gtk_signal_connect((( (menu_item) != 0 &&( (((GtkObject*) (menu_item))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (menu_item))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (menu_item) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (menu_item), GTK_TYPE_OBJECT) ),"activate",
		     (GtkSignalFunc)gfig_edit_menu_callback,
 0);







}

static void
PREFIXgfig_op_menu_popup(gint button, guint32 activate_time,GFIGOBJ *obj)
{
  gfig_obj_for_menu = obj;

  if(obj->obj_status & 0x2)
    {
      gtk_widget_set_sensitive(save_menu_item,(0));
    }
  else
    {
      gtk_widget_set_sensitive(save_menu_item,(!(0)));
    }

  gtk_menu_popup((((GtkMenu *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_op_menu)), ((gtk_menu_get_type ()))))),0,0,0,0,button,activate_time);
}


static gint
PREFIXlist_button_press(GtkWidget *widget,
		  GdkEventButton *event,
		  gpointer   data)
{
  switch (event->type)
    {
    case GDK_BUTTON_PRESS:



      if(event->button == 3)
	{



	  gfig_op_menu_popup(event->button,event->time,(GFIGOBJ *)data);
	  return((0));
	}
      list_button_update((GFIGOBJ *)data);
      break;
    case GDK_2BUTTON_PRESS:



      gfig_dialog_edit_list(widget,data,(0));
      break;
    default:
      printf("Unknown event\n");
      break;
    }

  return((0));
}

static void
PREFIXgfig_entry_update(GtkWidget *widget, gint *value)
{
  GtkAdjustment *adjustment;
  gdouble        new_value;

  new_value = atoi(gtk_entry_get_text((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((widget)), ((gtk_entry_get_type ())))))));

  if (*value != new_value) {
    adjustment = gtk_object_get_user_data((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));

    if ((new_value >= adjustment->lower) &&
	(new_value <= adjustment->upper)) {
      *value            = new_value;
      adjustment->value = new_value;

      gtk_signal_emit_by_name((( (adjustment) != 0 &&( (((GtkObject*) (adjustment))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (adjustment))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (adjustment) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (adjustment), GTK_TYPE_OBJECT) ), "value_changed");


    }
  }
}

static void
PREFIXgfig_scale_update(GtkAdjustment *adjustment, gint *value)
{
  GtkWidget *entry;
  char       buf[256];

  if (*value != adjustment->value) {
    *value = adjustment->value;

    entry = gtk_object_get_user_data((( (adjustment) != 0 &&( (((GtkObject*) (adjustment))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (adjustment))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (adjustment) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (adjustment), GTK_TYPE_OBJECT) ));
    sprintf(buf,"%d",*value);

    gtk_signal_handler_block_by_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), value);
    gtk_entry_set_text((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((entry)), ((gtk_entry_get_type ()))))), buf);
    gtk_signal_handler_unblock_by_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), value);


  }
}


static void
PREFIXgfig_scale_update_scale(GtkAdjustment *adjustment, gdouble *value)
{

  if (*value != adjustment->value) {
    *value = adjustment->value;
    if(!selvals.scaletoimage)
      {
	scale_x_factor = (1/(*value)) * org_scale_x_factor;
	scale_y_factor = (1/(*value)) * org_scale_y_factor;
	update_draw_area(gfig_preview,0);
      }
  }
}


static void
PREFIXgfig_scale_update_fp(GtkAdjustment *adjustment, gdouble *value)
{
  GtkWidget *entry;
  char       buf[256];

  if (*value != adjustment->value) {
    *value = adjustment->value;

    entry = gtk_object_get_user_data((( (adjustment) != 0 &&( (((GtkObject*) (adjustment))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (adjustment))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (adjustment) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (adjustment), GTK_TYPE_OBJECT) ));
    if(entry)
      {
	sprintf(buf,"%0.3f",*value);

	gtk_signal_handler_block_by_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), value);
	gtk_entry_set_text((((GtkEntry *) gtk_type_check_object_cast ((GtkTypeObject*) ((entry)), ((gtk_entry_get_type ()))))), buf);
	gtk_signal_handler_unblock_by_data((( (entry) != 0 &&( (((GtkObject*) (entry))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (entry))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (entry) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (entry), GTK_TYPE_OBJECT) ), value);
      }


  }
}




























static void
PREFIXgfig_scale2img_update(GtkWidget *widget,
                      gpointer   data)
{
  gint *val = (gint *)data;
  GtkAdjustment *adjustment;
  GtkWidget *sw;

  adjustment = gtk_object_get_user_data((( (widget) != 0 &&( (((GtkObject*) (widget))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (widget))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (widget) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (widget), GTK_TYPE_OBJECT) ));
  sw = gtk_object_get_user_data((( (adjustment) != 0 &&( (((GtkObject*) (adjustment))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (adjustment))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (adjustment) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (adjustment), GTK_TYPE_OBJECT) ));

  if(*val)
    {
      *val = 0;
      gtk_widget_set_sensitive((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((sw)), ((gtk_widget_get_type ()))))),(!(0)));
    }
  else
    {
      scale_x_factor = org_scale_x_factor;
      scale_y_factor = org_scale_y_factor;
      adjustment->value = 1.0;
      gtk_signal_emit_by_name((( (adjustment) != 0 &&( (((GtkObject*) (adjustment))->klass) != 0 &&((GtkFundamentalType) ((((GtkObjectClass*) (((GtkObject*) (adjustment))->klass))->type) & 0xFF)) == GTK_TYPE_OBJECT ) ) ? (GtkObject*) (adjustment) : (GtkObject*) gtk_type_check_object_cast ((GtkTypeObject*) (adjustment), GTK_TYPE_OBJECT) ), "value_changed");
      *val = 1;
      gtk_widget_set_sensitive((((GtkWidget *) gtk_type_check_object_cast ((GtkTypeObject*) ((sw)), ((gtk_widget_get_type ()))))),(0));
    }
  update_draw_area(gfig_preview,0);
}



static void
PREFIXgfig_toggle_update(GtkWidget *widget,
                      gpointer   data)
{
  gint *val = (gint *)data;

  if(*val)
    *val = 0;
  else
    *val = 1;
}


static void
PREFIXdo_gfig_preview(guchar *dest_row,
	    guchar *src_row,
	    gint width,
	    gint dh,
	    gint height,
	    gint bpp)
{
  memcpy(dest_row,src_row,width*bpp);
}

static void
PREFIXdialog_update_preview(void)
{
  gint y;
  gint check,check_0,check_1;

  if(!selvals.showimage)
    {
      memset(preview_row,-1,preview_width*4);
      for (y = 0; y < preview_height; y++) {
	gtk_preview_draw_row((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_preview)), ((gtk_preview_get_type ()))))), preview_row, 0, y, preview_width);
      }
      return;
    }

  if(!pv_cache)
    {
      refill_cache();
    }

  for (y = 0; y < preview_height; y++) {

    if ((y / 8) & 1) {
      check_0 =((int) (1.0 / 3.0 * 255));
      check_1 =((int) (2.0 / 3.0 * 255));
    } else {
      check_0 =((int) (2.0 / 3.0 * 255));
      check_1 =((int) (1.0 / 3.0 * 255));
    }

    do_gfig_preview(preview_row,
		pv_cache+y*preview_width*img_bpp,
		preview_width,
		y,
		preview_height,
		img_bpp);

    if(img_bpp > 3)
      {
	int i,j;
	for (i = 0, j = 0 ; i < sizeof(preview_row); i += 4, j += 3 )
	  {
	    gint alphaval;
	    if (((i/4) / 8) & 1)
	      check = check_0;
	    else
	      check = check_1;

	    alphaval = preview_row[i + 3];

	    preview_row[j] =
	      check + (((preview_row[i] - check)*alphaval)/255);
	    preview_row[j + 1] =
	      check + (((preview_row[i + 1] - check)*alphaval)/255);
	    preview_row[j + 2] =
	      check + (((preview_row[i + 2] - check)*alphaval)/255);
	  }
      }

    gtk_preview_draw_row((((GtkPreview *) gtk_type_check_object_cast ((GtkTypeObject*) ((gfig_preview)), ((gtk_preview_get_type ()))))), preview_row, 0, y, preview_width);
  }
}

static void
PREFIXgfig_new_gc()
{
 GdkColor fg, bg;


 gfig_gc = gdk_gc_new (gfig_preview->window);
 gdk_gc_set_function (gfig_gc, GDK_INVERT);
 fg.pixel = 0xFFFFFFFF;
 bg.pixel = 0x00000000;
 gdk_gc_set_foreground (gfig_gc, &fg);
 gdk_gc_set_background (gfig_gc, &bg);
 gdk_gc_set_line_attributes (gfig_gc, 1, GDK_LINE_SOLID,GDK_CAP_BUTT,GDK_JOIN_MITER);
}

static gint
PREFIXget_num_radials()
{
  gint gridsp = 50 + 10;




  return(gridsp - selvals.opts.gridspacing);
}



static gint
PREFIXinside_sqr(GdkPoint *cpnt, GdkPoint *testpnt)
{

  gint16 x = cpnt->x;
  gint16 y = cpnt->y;
  gint16 tx = testpnt->x;
  gint16 ty = testpnt->y;




  return(abs(x - tx) <= 8 && abs (y - ty) < 8 );
}




static void
PREFIXfind_grid_pos(GdkPoint *p,GdkPoint *gp,guint is_butt3)
{
  gint16 x = p->x;
  gint16 y = p->y;
  static GdkPoint cons_pnt;
  static gdouble cons_radius;
  static gdouble cons_ang;
  static gboolean cons_center;

  if(selvals.opts.gridtype == RECT_GRID)
    {
      if(p->x % selvals.opts.gridspacing > selvals.opts.gridspacing/2)
	x += selvals.opts.gridspacing;

      if(p->y % selvals.opts.gridspacing > selvals.opts.gridspacing/2)
	y += selvals.opts.gridspacing;

      gp->x = (x/selvals.opts.gridspacing)*selvals.opts.gridspacing;
      gp->y = (y/selvals.opts.gridspacing)*selvals.opts.gridspacing;

      if(is_butt3)
	{
	  if(abs(gp->x - cons_pnt.x) < abs(gp->y - cons_pnt.y))
	    gp->x = cons_pnt.x;
	  else
	    gp->y = cons_pnt.y;
	}
      else
	{

	  cons_pnt = *gp;
	}
    }
  else if(selvals.opts.gridtype == POLAR_GRID)
    {
      gdouble ang_grid;
      gdouble ang_radius;
      gdouble real_radius;
      gdouble real_angle;
      gdouble rounded_angle;
      gint rounded_radius;
      gint16 shift_x = x - preview_width/2;
      gint16 shift_y = -y + preview_height/2;

      real_radius = ang_radius = sqrt((shift_y*shift_y) + (shift_x*shift_x));


      rounded_radius = (gint)(rint(ang_radius/selvals.opts.gridspacing))*selvals.opts.gridspacing;
      if(rounded_radius <= 0 || real_radius <=0)
	{

	  gp->x = preview_width/2;
	  gp->y = preview_height/2;
	  if (!is_butt3) cons_center =(!(0));



	  return;
	}

      ang_grid = 2* 3.14159265358979323846 /get_num_radials();

      real_angle = atan2(shift_y,shift_x);
      if(real_angle < 0)
	real_angle += 2* 3.14159265358979323846;

      rounded_angle = (rint((real_angle/ang_grid)))*ang_grid;






      gp->x = (gint)rint((rounded_radius*cos(rounded_angle))) + preview_width/2;
      gp->y = -(gint)rint((rounded_radius*sin(rounded_angle))) + preview_height/2;

      if(is_butt3)
	{
	  if(!cons_center)
	    {
	      if(fabs(rounded_angle - cons_ang) > ang_grid/2)
		{
		  gp->x = (gint)rint((cons_radius*cos(rounded_angle))) + preview_width/2;
		  gp->y = -(gint)rint((cons_radius*sin(rounded_angle))) + preview_height/2;
		}
	      else
		{
		  gp->x = (gint)rint((rounded_radius*cos(cons_ang))) + preview_width/2;
		  gp->y = -(gint)rint((rounded_radius*sin(cons_ang))) + preview_height/2;
		}
	    }
	}
      else
	{
	  cons_radius = rounded_radius;
	  cons_ang = rounded_angle;
	  cons_center =(0);
	}
    }
   else if(selvals.opts.gridtype == ISO_GRID)
     {
	if(is_butt3)
	  {
	     static GdkPoint b_pnt;
	     static GdkPoint i_pnt;
	     static GdkPoint ii_pnt;
	     gint d;
	     gint dd;

	     b_pnt.x = cons_pnt.x;
	     b_pnt.y = cons_pnt.y + preview_width;
	     d = calculate_point_to_line_distance(p, &cons_pnt, &b_pnt, &i_pnt);

	     b_pnt.x = cons_pnt.x;
	     b_pnt.y = cons_pnt.y - preview_width;
	     dd = calculate_point_to_line_distance(p, &cons_pnt, &b_pnt, &ii_pnt);
	     if (dd < d)
	     {
		i_pnt.x = ii_pnt.x;
		i_pnt.y = ii_pnt.y;
		d = dd;
	     }

	     b_pnt.x = cons_pnt.x + preview_width;
	     b_pnt.y = cons_pnt.y + preview_width/2;
	     dd = calculate_point_to_line_distance(p, &cons_pnt, &b_pnt, &ii_pnt);
	     if (dd < d)
	     {
		i_pnt.x = ii_pnt.x;
		i_pnt.y = ii_pnt.y;
		d = dd;
	     }

	     b_pnt.x = cons_pnt.x + preview_width;
	     b_pnt.y = cons_pnt.y - preview_width/2;
	     dd = calculate_point_to_line_distance(p, &cons_pnt, &b_pnt, &ii_pnt);
	     if (dd < d)
	     {
		i_pnt.x = ii_pnt.x;
		i_pnt.y = ii_pnt.y;
		d = dd;
	     }

	     b_pnt.x = cons_pnt.x - preview_width;
	     b_pnt.y = cons_pnt.y + preview_width/2;
	     dd = calculate_point_to_line_distance(p, &cons_pnt, &b_pnt, &ii_pnt);
	     if (dd < d)
	     {
		i_pnt.x = ii_pnt.x;
		i_pnt.y = ii_pnt.y;
		d = dd;
	     }

	     b_pnt.x = cons_pnt.x - preview_width;
	     b_pnt.y = cons_pnt.y - preview_width/2;
	     dd = calculate_point_to_line_distance(p, &cons_pnt, &b_pnt, &ii_pnt);
	     if (dd < d)
	     {
		i_pnt.x = ii_pnt.x;
		i_pnt.y = ii_pnt.y;
		d = dd;
	     }

	     x = i_pnt.x;
	     y = i_pnt.y;
	  }

	if(x % selvals.opts.gridspacing > selvals.opts.gridspacing/2)
	  x += selvals.opts.gridspacing;

	gp->x = (x/selvals.opts.gridspacing)*selvals.opts.gridspacing;

	if (((gp->x/selvals.opts.gridspacing) % 2) != 0)
	  {
	     y -= selvals.opts.gridspacing/2;

	     if(y % selvals.opts.gridspacing > selvals.opts.gridspacing/2)
	       y += selvals.opts.gridspacing;

	     gp->y = (selvals.opts.gridspacing/2) + ((y/selvals.opts.gridspacing)*selvals.opts.gridspacing);
	  }
	else
	  {
	     if(y % selvals.opts.gridspacing > selvals.opts.gridspacing/2)
	       y += selvals.opts.gridspacing;

	     gp->y = (y/selvals.opts.gridspacing)*selvals.opts.gridspacing;
	  }

	if (!is_butt3)
	  {

	     cons_pnt = *gp;
	  }
     }
}


#line 7008 "gfig.c"
static int
PREFIXcalculate_point_to_line_distance(GdkPoint *p, GdkPoint *A, GdkPoint *B, GdkPoint *I)
{
   gint L2;
   gint L;

   L2 = ((B->x - A->x)*(B->x - A->x)) + ((B->y - A->y)*(B->y - A->y));
   L = (gint) sqrt(L2);








   I->x = A->x + (((A->y - p->y)*(A->y - B->y) - (A->x - p->x)*(B->x - A->x))*(B->x - A->x))/L2;
   I->y = A->y + (((A->y - p->y)*(A->y - B->y) - (A->x - p->x)*(B->x - A->x))*(B->y - A->y))/L2;

   return abs((((A->y - p->y)*(B->x - A->x)) - ((A->x - p->x)*(B->y - A->y)))*L);
}


static void
PREFIXdraw_circle(GdkPoint *p)
{
  if(!selvals.opts.showcontrol || drawing_pic)
    return;

  gdk_draw_arc (gfig_preview->window,
		gfig_gc,
		0,
		p->x - 8 /2,
		p->y - 8 /2,
 8,
 8,
		0,
		360*64);
}



static void
PREFIXdraw_sqr(GdkPoint *p)
{
  if(!selvals.opts.showcontrol || drawing_pic)
    return;

  gdk_draw_rectangle(gfig_preview->window,
		     gfig_gc,
		     0,
		     gfig_scale_x((gint)p->x) - 8 /2,
		     gfig_scale_y((gint)p->y) - 8 /2,
		     (gint)8,
		     (gint)8);
}


#line 7067 "gfig.c"

static void
PREFIXdraw_grid_clear(GtkWidget *widget,
	  gpointer   data)
{

  dialog_update_preview();
  draw_grid(widget,data);
  draw_objects(current_obj->obj_list,(!(0)));
  gtk_widget_draw(gfig_preview,0);
  gdk_flush();
}

static void
PREFIXtoggle_tooltips(GtkWidget *widget,
	  gpointer   data)
{
  gint tt = *((gint *)data);

  if(tt)
    gtk_tooltips_disable(gfig_tooltips);
  else
    gtk_tooltips_enable(gfig_tooltips);
}


static void
PREFIXtoggle_show_image(GtkWidget *widget,
	  gpointer   data)
{

  draw_grid_clear(widget,data);
}


static void
PREFIXtoggle_obj_type(GtkWidget *widget,
	  gpointer   data)
{
  GdkCursorType ctype = GDK_LAST_CURSOR;
  static GdkCursor* p_cursors[DEL_OBJ + 1];


  if(selvals.otype != (DOBJTYPE)data)
    {

      obj_creating = 0;
      tmp_line = 0;
      tmp_bezier = 0;

      if((DOBJTYPE)data < MOVE_OBJ)
	{
	  obj_show_single = -1;
	}

      update_draw_area(gfig_preview,0);

      list_button_update(current_obj);
    }

  selvals.otype = (DOBJTYPE)data;

  switch(selvals.otype)
    {
    case LINE:
    case CIRCLE:
    case ELLIPSE:
    case ARC:
    case POLY:
    case STAR:
    case SPIRAL:
    case BEZIER:
    default:
      ctype = GDK_CROSSHAIR;
      break;
    case MOVE_OBJ:
    case MOVE_POINT:
    case COPY_OBJ:
    case MOVE_COPY_OBJ:
      ctype = GDK_DIAMOND_CROSS;
      break;
    case DEL_OBJ:
      ctype = GDK_PIRATE;
      break;
    }

  if(!p_cursors[selvals.otype])
    p_cursors[selvals.otype] = gdk_cursor_new(ctype);

  gdk_window_set_cursor(gfig_preview->window,p_cursors[selvals.otype]);
}

static void
PREFIXdraw_grid_polar(GdkGC *drawgc)
{
  gint step;
  gint loop;
  gint radius;
  gint max_rad;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble ang_radius;


  gint grid_x_center = preview_width/2;
  gint grid_y_center = preview_height/2;

  step = selvals.opts.gridspacing;
  max_rad = sqrt(preview_width*preview_width + preview_height*preview_height)/2;

  for(loop = 0 ; loop < max_rad ; loop += step)
    {
      radius = loop;

      gdk_draw_arc (gfig_preview->window,
		    drawgc,
		    0,
		    grid_x_center - radius,
		    grid_y_center - radius,
		    radius*2,
		    radius*2,
		    0,
		    360*64);
    }


  ang_grid = 2* 3.14159265358979323846 /get_num_radials();
  ang_radius = sqrt((preview_width*preview_width) + (preview_height*preview_height))/2;

  for(loop = 0 ; loop <= get_num_radials() ; loop++)
    {
      gint lx,ly;

      ang_loop = loop * ang_grid;

      lx = (gint)rint(ang_radius * cos(ang_loop));
      ly = (gint)rint(ang_radius * sin(ang_loop));

      gdk_draw_line(gfig_preview->window,
		    drawgc,
		    (gint)lx + (preview_width)/2,
		    -(gint)ly + (preview_height)/2,
		    (gint)(preview_width)/2,
		    (gint)(preview_height)/2);
    }
}

static void
PREFIXdraw_grid_sq(GdkGC *drawgc)
{
  gint step;
  gint loop;


  step = selvals.opts.gridspacing;

  for(loop = 0 ; loop < preview_height ; loop += step)
    {
        gdk_draw_line(gfig_preview->window,
		drawgc,
		(gint)0,
		(gint)loop,
		(gint)preview_width,
		(gint)loop);
    }



  for(loop = 0 ; loop < preview_width ; loop += step)
    {
        gdk_draw_line(gfig_preview->window,
		drawgc,
		(gint)loop,
		(gint)0,
		(gint)loop,
		(gint)preview_height);
    }
}

static void
PREFIXdraw_grid_iso(GdkGC *drawgc)
{
   gint step;
   gint loop;

   gint diagonal_start;
   gint diagonal_end;
   gint diagonal_width;
   gint diagonal_height;

   step = selvals.opts.gridspacing;


   for (loop = 0 ; loop < preview_width ; loop += step)
     {
	gdk_draw_line(gfig_preview->window,
		      drawgc,
		      (gint)loop,
		      (gint)0,
		      (gint)loop,
		      (gint)preview_height);
     }

   diagonal_start = preview_width/2;
   diagonal_start = diagonal_start - (diagonal_start % step);
   diagonal_start = -diagonal_start;

   diagonal_end = preview_height + (preview_width/2);
   diagonal_end = diagonal_end - (diagonal_end % step);

   diagonal_width = preview_width;
   diagonal_height = diagonal_width/2;


   for (loop = diagonal_start ; loop < diagonal_end ; loop += step)
     {
	gdk_draw_line(gfig_preview->window,
		      drawgc,
		      (gint)0,
		      (gint)loop,
		      (gint)diagonal_width,
		      (gint)loop + diagonal_height);

	gdk_draw_line(gfig_preview->window,
		      drawgc,
		      (gint)0,
		      (gint)loop,
		      (gint)diagonal_width,
		      (gint)loop - diagonal_height);
     }
}

static GdkGC *
PREFIXgfig_get_grid_gc(GtkWidget *w, gint gctype)
{
  switch(gctype)
    {
    case -2 :
      return(w->style->black_gc);
    case -3 :
      return(w->style->white_gc);
    case -4 :
      return(grid_hightlight_drawgc);
    case GTK_STATE_NORMAL:
      return(w->style->bg_gc[GTK_STATE_NORMAL]);
    case GTK_STATE_ACTIVE:
      return(w->style->bg_gc[GTK_STATE_ACTIVE]);
    case GTK_STATE_PRELIGHT:
      return(w->style->bg_gc[GTK_STATE_PRELIGHT]);
    case GTK_STATE_SELECTED:
      return(w->style->bg_gc[GTK_STATE_SELECTED]);
    case GTK_STATE_INSENSITIVE:
      return(w->style->bg_gc[GTK_STATE_INSENSITIVE]);
    default:
      g_warning("Unknown type for grid colouring\n");
      return(w->style->bg_gc[GTK_STATE_PRELIGHT]);
    }
}

static void
PREFIXdraw_grid(GtkWidget *widget,
	  gpointer   data)
{
  GdkGC *drawgc;



#line 7335 "gfig.c"

  if((preview_width < selvals.opts.gridspacing &&
     preview_height < selvals.opts.gridspacing ) ||
     drawing_pic)
    {

      return;
    }

  if(selvals.opts.drawgrid)
    drawgc = gfig_get_grid_gc(gfig_preview,grid_gc_type);
  else
    return;

  if(selvals.opts.gridtype == RECT_GRID)
    draw_grid_sq(drawgc);
  else if(selvals.opts.gridtype == POLAR_GRID)
    draw_grid_polar(drawgc);
  else if(selvals.opts.gridtype == ISO_GRID)
    draw_grid_iso(drawgc);
}

static void
PREFIXdo_gfig(void)
{

}


#line 7366 "gfig.c"




#line 7376 "gfig.c"

static void
PREFIXfree_one_obj(DOBJECT *obj)
{
  d_delete_dobjpoints(obj->points);
  g_free(obj);
}

static void
PREFIXfree_all_objs(DALLOBJS * objs)
{

  DALLOBJS * next;

  while(objs)
    {
      free_one_obj(objs->obj);
      next = objs->next;
      g_free(objs);
      objs = next;
    }
}

char *
PREFIXget_line(gchar *buf,gint s,FILE * from,gint init)
{
  gint slen;
  char * ret;

  if(init)
    line_no = 1;
  else
    line_no++;

  do
    {
      ret = fgets(buf,s,from);
    } while(!((from)->_flag& 040) && buf[0] == '#');

  slen = strlen(buf);


  if(slen > 0)
    buf[slen - 1] = '\0';

  if(((from)->_flag& 040))
    {
      g_warning("Error reading file");
      return(0);
    }





  return(ret);
}

static void
PREFIXgfig_clear_callback (GtkWidget *widget,
		      gpointer   data)
{

  if(!current_obj->obj_list)
    return;

  setup_undo();

  free_all_objs(current_obj->obj_list);
  current_obj->obj_list = 0;
  obj_creating = 0;
  tmp_line = 0;
  tmp_bezier = 0;
  update_draw_area(gfig_preview,0);

  list_button_update(current_obj);
}


static void
PREFIXgfig_undo_callback (GtkWidget *widget,
			  gpointer   data)
{
  if(undo_water_mark >= 0)
    {

      free_all_objs(current_obj->obj_list);
      current_obj->obj_list = 0;
      tmp_bezier = tmp_line = obj_creating = 0;
      current_obj->obj_list = undo_table[undo_water_mark];
      undo_water_mark--;

      update_draw_area(gfig_preview,0);

      list_button_update(current_obj);
      gfig_obj_modified(current_obj,0x1);
      current_obj->obj_status |= 0x1;
  }

  if(undo_water_mark < 0)
    gtk_widget_set_sensitive(widget,(0));
}

static void
PREFIXclear_undo()
{
  int lv;

  for(lv = undo_water_mark; lv >= 0; lv--)
    {
      if(undo_table[lv])
	free_all_objs(undo_table[lv]);
      undo_table[lv] = 0;
    }

  undo_water_mark = -1;
  gtk_widget_set_sensitive(undo_widget,(0));
}

static void
PREFIXsetup_undo()
{





  if(!current_obj)
    {

      return;
    }

  if(undo_water_mark >= selvals.maxundo - 1)
    {
      int loop;

      if(undo_table[0])
	free_one_obj(undo_table[0]->obj);
      for(loop = 0 ; loop < undo_water_mark ; loop++)
	{
	  undo_table[loop] = undo_table[loop + 1];
	}
    }
  else
    {
      undo_water_mark++;
    }
  undo_table[undo_water_mark] = copy_all_objs(current_obj->obj_list);
  gtk_widget_set_sensitive(undo_widget,(!(0)));

  gfig_obj_modified(current_obj,0x1);
  current_obj->obj_status |= 0x1;
}





void
PREFIXscale_to_orginal_x(gdouble *list)
{
  *list *= scale_x_factor;
}

static gint
PREFIXgfig_scale_x(gint x)
{
  if(!selvals.scaletoimage)
    return((gint)(x*(1/scale_x_factor)));
  else
    return(x);
}

static gint
PREFIXgfig_invscale_x(gint x)
{
  if(!selvals.scaletoimage)
    return((gint)(x*(scale_x_factor)));
  else
    return(x);
}

void
PREFIXscale_to_orginal_y(gdouble *list)
{
  *list *= scale_y_factor;
}

static gint
PREFIXgfig_scale_y(gint y)
{
  if(!selvals.scaletoimage)
    return((gint)(y*(1/scale_y_factor)));
  else
    return(y);
}

static gint
PREFIXgfig_invscale_y(gint y)
{
  if(!selvals.scaletoimage)
    return((gint)(y*(scale_y_factor)));
  else
    return(y);
}



void
PREFIXscale_to_original_xy(gdouble *list,gint size)
{
  int i;
  for (i = 0 ; i < size*2 ; i +=2)
    {
      scale_to_orginal_x(&list[i]);
      scale_to_orginal_y(&list[i+1]);
    }
}


void
PREFIXscale_to_xy(gdouble *list,gint size)
{
  int i;
  for (i = 0 ; i < size*2 ; i +=2)
    {
      list[i] *= (org_scale_x_factor/scale_x_factor);
      list[i + 1] *= (org_scale_y_factor/scale_y_factor);
    }
}




void
PREFIXreverse_pairs_list(gdouble *list, gint size)
{
  int i;
  struct cs {
    gdouble i1;
    gdouble i2;
  } copyit,*orglist;

  orglist = (struct cs *)list;


  for(i = 0 ; i < size/2 ; i++)
    {
      copyit = orglist[i];
      orglist[i] = orglist[size - 1 - i];
      orglist[size - 1 - i] = copyit;
    }
}



void
PREFIXd_delete_dobjpoints(DOBJPOINTS * pnts)
{
  DOBJPOINTS * next;
  DOBJPOINTS * pnt2del = pnts;

  while(pnt2del)
    {
      next = pnt2del->next;
      g_free(pnt2del);
      pnt2del = next;
    }
}

DOBJPOINTS *
PREFIXd_copy_dobjpoints(DOBJPOINTS * pnts)
{
  DOBJPOINTS * ret = 0;
  DOBJPOINTS * head = 0;
  DOBJPOINTS * newpnt;
  DOBJPOINTS * pnt2copy = pnts;

  while(pnt2copy)
    {
      newpnt = g_malloc0(sizeof(DOBJPOINTS));
      newpnt->pnt.x = pnt2copy->pnt.x;
      newpnt->pnt.y = pnt2copy->pnt.y;

      if(!ret)
	head = ret = newpnt;
      else
	{
	  head->next = newpnt;
	  head = newpnt;
	}
      pnt2copy = pnt2copy->next;
    }

  return(ret);
}

static gint
PREFIXscan_obj_points(DOBJPOINTS *opnt,GdkPoint *pnt)
{
  while(opnt)
    {
      if(inside_sqr(&opnt->pnt,pnt))
	{
	  opnt->found_me =(!(0));
	  return((!(0)));
	}
      opnt->found_me =(0);
      opnt = opnt->next;
    }
  return((0));
}

static DOBJECT *
PREFIXget_nearest_objs(GFIGOBJ * obj,GdkPoint *pnt)
{

  DALLOBJS *all;
  DOBJECT  *test_obj;
  gint count = 0;

  if(!obj)
    return(0);

  all = obj->obj_list;

  while(all)
    {
      test_obj = all->obj;

      if(count == obj_show_single || obj_show_single == -1)
	if(scan_obj_points(test_obj->points,pnt))
	  {
	    return(test_obj);
	  }
      all = all->next;
      count++;
    }
  return(0);
}

static void
PREFIXscale_obj_points(DOBJPOINTS *opnt,gdouble scale_x,gdouble scale_y)
{
  while(opnt)
    {
      opnt->pnt.x = (gint)(opnt->pnt.x * scale_x);
      opnt->pnt.y = (gint)(opnt->pnt.y * scale_y);
      opnt = opnt->next;
    }
}

static void
PREFIXremove_obj_from_list(GFIGOBJ *obj,DOBJECT *del_obj)
{

  DALLOBJS *all;
  DALLOBJS  *prev_all = 0;

 if (1){ if (!(del_obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 7736,"del_obj != NULL"); }else (void)0;

  all = obj->obj_list;

  while(all)
    {
      if(all->obj == del_obj)
	{





	  if(prev_all)
	    prev_all->next = all->next;
	  else
	    obj->obj_list = all->next;


	  del_obj->drawfunc(del_obj);

	  free_one_obj(del_obj);
	  g_free(all);

	  if(obj_show_single != -1)
	    {

	      draw_grid_clear(0,0);
	      obj_show_single = -1;
	    }
	  return;
	}
      prev_all = all;
      all = all->next;
    }
  g_warning("Hey where has the object gone ?");
}

static DOBJPOINTS *
PREFIXget_diffs(DOBJECT *obj,gint16 *xdiff, gint16 *ydiff, GdkPoint *to_pnt)
{
  DOBJPOINTS *spnt;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 7779,"obj != NULL"); }else (void)0;

  spnt = obj->points;

  if(!spnt)
    return(0);


  while(spnt)
    {
      if(spnt->found_me)
	{
	  *xdiff = spnt->pnt.x - to_pnt->x;
	  *ydiff = spnt->pnt.y - to_pnt->y;
	  return(spnt);
	}
      spnt = spnt->next;
    }
  return(0);
}

static void
PREFIXupdate_pnts(DOBJECT *obj,gint16 xdiff, gint16 ydiff)
{
  DOBJPOINTS *spnt;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 7805,"obj != NULL"); }else (void)0;


  spnt = obj->points;

  if(!spnt)
    return;


  while(spnt)
    {
      spnt->pnt.x = spnt->pnt.x - xdiff;
      spnt->pnt.y = spnt->pnt.y - ydiff;
      spnt = spnt->next;
    }
}


static void
PREFIXdo_move_all_obj(GdkPoint *to_pnt)
{


  DALLOBJS *all;
  DOBJECT *obj;
  gint16 xdiff = 0;
  gint16 ydiff = 0;

  xdiff = move_all_pnt->x - to_pnt->x;
  ydiff = move_all_pnt->y - to_pnt->y;

  if(!xdiff && !ydiff)
    return;

  all = current_obj->obj_list;

  while(all)
    {
      obj = all->obj;


      draw_one_obj(obj);

      update_pnts(obj,xdiff,ydiff);


      draw_one_obj(obj);

      all = all->next;
    }

  *move_all_pnt = *to_pnt;
}


static void
PREFIXdo_move_obj(DOBJECT *obj,GdkPoint *to_pnt)
{


  gint16 xdiff = 0;
  gint16 ydiff = 0;

  get_diffs(obj,&xdiff,&ydiff,to_pnt);

  if(!xdiff && !ydiff)
    return;


  draw_one_obj(obj);

  update_pnts(obj,xdiff,ydiff);


  draw_one_obj(obj);

}

static void
PREFIXdo_move_obj_pnt(DOBJECT *obj,GdkPoint *to_pnt)
{


  DOBJPOINTS *spnt;
  gint16 xdiff = 0;
  gint16 ydiff = 0;

  spnt = get_diffs(obj,&xdiff,&ydiff,to_pnt);

  if((!xdiff && !ydiff) || !spnt)
    return;


  draw_one_obj(obj);

  spnt->pnt.x = spnt->pnt.x - xdiff;
  spnt->pnt.y = spnt->pnt.y - ydiff;


  draw_one_obj(obj);

}



void
PREFIXd_save_line(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<LINE>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"</LINE>\n");
}



DOBJECT *
PREFIXd_load_line(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(strcmp("</LINE>",buf))
	    {
	      g_warning("[%d] Internal load error while loading line",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_line(xpnt,ypnt);
      else
	d_pnt_add_line(new_obj,xpnt,ypnt,-1);
    }

  return(new_obj);
}

DOBJECT *
PREFIXd_copy_line(DOBJECT *obj)
{
  DOBJECT *nl;

  if(!obj)
    return(0);

 if (1){ if (!(obj->type == LINE)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 7978,"obj->type == LINE"); }else (void)0;

  nl = d_new_line(obj->points->pnt.x,obj->points->pnt.y);

  nl->points->next = d_copy_dobjpoints(obj->points->next);

  return(nl);
}


void
PREFIXd_draw_line(DOBJECT * obj)
{
  DOBJPOINTS * spnt;
  DOBJPOINTS * epnt;

  spnt = obj->points;

  if(!spnt)
    return;

  epnt = spnt->next;

  while(spnt && epnt)
    {








      draw_sqr(&spnt->pnt);

      if(drawing_pic)
	{
	  gdk_draw_line(pic_preview->window,
			pic_preview->style->black_gc,
			adjust_pic_coords((gint)spnt->pnt.x,preview_width),
			adjust_pic_coords((gint)spnt->pnt.y,preview_height),
			adjust_pic_coords((gint)epnt->pnt.x,preview_width),
			adjust_pic_coords((gint)epnt->pnt.y,preview_height));
	}
      else
	{
	  gdk_draw_line(gfig_preview->window,
			gfig_gc,
			gfig_scale_x((gint)spnt->pnt.x),
			gfig_scale_y((gint)spnt->pnt.y),
			gfig_scale_x((gint)epnt->pnt.x),
			gfig_scale_y((gint)epnt->pnt.y));
	}
      spnt = epnt;
      epnt = epnt->next;
    }
  draw_sqr(&spnt->pnt);
}

static void
PREFIXd_paint_line(DOBJECT *obj)
{
  DOBJPOINTS * spnt;
  gdouble *line_pnts;
  GParam *return_vals = 0;
  gint nreturn_vals;
  gint seg_count = 0;
  gint i = 0;

  spnt = obj->points;



  while(spnt)
    {
      seg_count++;
      spnt = spnt->next;
    }

  spnt = obj->points;

  if(!spnt || !seg_count)
    return;


  line_pnts = g_malloc0(1 *(2*seg_count + 1)*sizeof(gdouble));


  while(spnt)
    {
      line_pnts[i++] = spnt->pnt.x;
      line_pnts[i++] = spnt->pnt.y;
      spnt = spnt->next;
    }


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.scaletoimage)
    scale_to_original_xy(&line_pnts[0],i/2);
  else
    scale_to_xy(&line_pnts[0],i/2);


  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,seg_count*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,seg_count*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,seg_count*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,seg_count*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {

      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,seg_count*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);
    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(line_pnts);
}



#line 8149 "gfig.c"

DOBJECT *
PREFIXd_new_line(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = LINE;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_line;
  nobj->loadfunc  = d_load_line;
  nobj->savefunc  = d_save_line;
  nobj->paintfunc = d_paint_line;
  nobj->copyfunc  = d_copy_line;

  return(nobj);
}


void
PREFIXd_delete_line(DOBJECT * obj)
{
 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 8184,"obj != NULL"); }else (void)0;

  d_delete_dobjpoints(obj->points);
  g_free(obj);
}



#line 8196 "gfig.c"

static void
PREFIXd_pnt_add_line(DOBJECT *obj, gint x, gint y, gint pos)
{
  DOBJPOINTS *npnts = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 8202,"obj != NULL"); }else (void)0;

  npnts->pnt.x = x;
  npnts->pnt.y = y;

  if(!pos)
    {

      npnts->next = obj->points;
      obj->points = npnts;
    }
  else
    {
      DOBJPOINTS *pnt = obj->points;


      while(pos < 0 || pos-- > 0)
	{
	  if(!(pnt->next) || !pos)
	    {
	      npnts->next = pnt->next;
	      pnt->next = npnts;
	      break;
	    }
	  else
	    {
	      pnt = pnt->next;
	    }
	}
    }
}


void
PREFIXd_update_line(GdkPoint *pnt)
{
  DOBJPOINTS *spnt, *epnt;

#line 8243 "gfig.c"


  spnt = obj_creating->points;

  if(!spnt)
    return;

  if((epnt = spnt->next))
    {


      draw_circle(&epnt->pnt);

      gdk_draw_line(gfig_preview->window,

			gfig_gc,
			(gint)spnt->pnt.x,
			(gint)spnt->pnt.y,
			(gint)epnt->pnt.x,
			(gint)epnt->pnt.y);
      g_free(epnt);
    }



  draw_circle(pnt);

  epnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));

  epnt->pnt.x = pnt->x;
  epnt->pnt.y = pnt->y;

  gdk_draw_line(gfig_preview->window,

		gfig_gc,
		(gint)spnt->pnt.x,
		(gint)spnt->pnt.y,
		(gint)epnt->pnt.x,
		(gint)epnt->pnt.y);
  spnt->next = epnt;
}

void
PREFIXd_line_start(GdkPoint *pnt,gint shift_down)
{
  if(!obj_creating || !shift_down)
    {


      obj_creating = d_new_line(pnt->x,pnt->y);
    }
  else
    {

      d_update_line(pnt);
    }
}

void
PREFIXd_line_end(GdkPoint *pnt,gint shift_down)
{

  draw_circle(pnt);

  if(shift_down)
    {
      if(tmp_line)
	{
	  GdkPoint tmp_pnt = *pnt;

	  if(need_to_scale)
	    {
	      tmp_pnt.x = (gint)(pnt->x * scale_x_factor);
	      tmp_pnt.y = (gint)(pnt->y * scale_y_factor);
	    }

	  d_pnt_add_line(tmp_line,tmp_pnt.x,tmp_pnt.y,-1);
	  free_one_obj(obj_creating);

	}
      else
	{
	  tmp_line = obj_creating;
	  add_to_all_obj(current_obj,obj_creating);
	}

      obj_creating = d_new_line(pnt->x,pnt->y);
    }
  else
    {
      if(tmp_line)
	{
	  GdkPoint tmp_pnt = *pnt;

	  if(need_to_scale)
	    {
	      tmp_pnt.x = (gint)(pnt->x * scale_x_factor);
	      tmp_pnt.y = (gint)(pnt->y * scale_y_factor);
	    }

	  d_pnt_add_line(tmp_line,tmp_pnt.x,tmp_pnt.y,-1);
	  free_one_obj(obj_creating);

	}
      else
	{
	  add_to_all_obj(current_obj,obj_creating);
	}
      obj_creating = 0;
      tmp_line = 0;
    }

}



void
PREFIXd_save_circle(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<CIRCLE>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"</CIRCLE>\n");
}



DOBJECT *
PREFIXd_load_circle(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(strcmp("</CIRCLE>",buf))
	    {
	      g_warning("[%d] Internal load error while loading circle",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_circle(xpnt,ypnt);
      else
	{
	  DOBJPOINTS *edge_pnt;

	  edge_pnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));

	  edge_pnt->pnt.x = xpnt;
	  edge_pnt->pnt.y = ypnt;

	  new_obj->points->next = edge_pnt;
	}
    }
  g_warning("[%d] Not enough points for circle",line_no);
  return(0);
}

static void
PREFIXd_draw_circle(DOBJECT * obj)
{
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * edge_pnt;
  gdouble radius;

  center_pnt = obj->points;

  if(!center_pnt)
    return;

  edge_pnt = center_pnt->next;

  if(!edge_pnt)
    {
      g_warning("Internal error - circle no edge pnt");
    }

  radius = sqrt(((center_pnt->pnt.x - edge_pnt->pnt.x) *
		 (center_pnt->pnt.x - edge_pnt->pnt.x)) +
		((center_pnt->pnt.y - edge_pnt->pnt.y) *
		 (center_pnt->pnt.y - edge_pnt->pnt.y)));

  draw_sqr(&center_pnt->pnt);
  draw_sqr(&edge_pnt->pnt);

  if(drawing_pic)
    {
      gdk_draw_arc (pic_preview->window,
		    pic_preview->style->black_gc,
		    0,
		    adjust_pic_coords(center_pnt->pnt.x - radius,
				      preview_width),
		    adjust_pic_coords(center_pnt->pnt.y - radius,
				      preview_height),
		    adjust_pic_coords(radius * 2,
				      preview_width),
		    adjust_pic_coords(radius * 2,
				      preview_height),
		    0,
		    360*64);
    }
  else
    {
      gdk_draw_arc (gfig_preview->window,
		    gfig_gc,
		    0,
		    gfig_scale_x(center_pnt->pnt.x - (gint)rint(radius)),
		    gfig_scale_y(center_pnt->pnt.y - (gint)rint(radius)),
		    gfig_scale_x((gint)rint(radius) * 2),
		    gfig_scale_y((gint)rint(radius) * 2),
		    0,
		    360*64);
    }
}

static void
PREFIXd_paint_circle(DOBJECT *obj)
{
  GParam *return_vals;
  gint nreturn_vals;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * edge_pnt;
  gint radius;
  gdouble dpnts[4];


 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 8496,"obj != NULL"); }else (void)0;

  if(selvals.approxcircles)
    {
      obj->type_data = (gpointer)600;



      d_paint_poly(obj);
      return;
    }



#line 8513 "gfig.c"
  center_pnt = obj->points;

  if(!center_pnt)
    return;

  edge_pnt = center_pnt->next;

  if(!edge_pnt)
    {
      g_error("Internal error - circle no edge pnt");
    }

  radius = (gint)sqrt(((center_pnt->pnt.x - edge_pnt->pnt.x) *
		 (center_pnt->pnt.x - edge_pnt->pnt.x)) +
		((center_pnt->pnt.y - edge_pnt->pnt.y) *
		 (center_pnt->pnt.y - edge_pnt->pnt.y)));

  dpnts[0] = (gdouble)center_pnt->pnt.x - radius;
  dpnts[1] = (gdouble)center_pnt->pnt.y - radius;
  dpnts[3] = dpnts[2] = (gdouble)radius*2;


  if(selvals.scaletoimage)
    scale_to_original_xy(&dpnts[0],2);
  else
    scale_to_xy(&dpnts[0],2);

  return_vals = gimp_run_procedure ("gimp_ellipse_select", &nreturn_vals,
				    PARAM_IMAGE, gfig_image,
				    PARAM_FLOAT,dpnts[0],
				    PARAM_FLOAT,dpnts[1],
				    PARAM_FLOAT,dpnts[2],
				    PARAM_FLOAT,dpnts[3],
				    PARAM_INT32,selopt.type,
				    PARAM_INT32,selopt.antia,
				    PARAM_INT32,selopt.feather,
				    PARAM_FLOAT,(gdouble)selopt.feather_radius,
				    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);


  if(selvals.painttype == PAINT_SELECTION_TYPE)
    return;

  return_vals = gimp_run_procedure ("gimp_edit_stroke", &nreturn_vals,
				    PARAM_DRAWABLE, gfig_drawable,
				    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);

  return_vals = gimp_run_procedure ("gimp_selection_clear", &nreturn_vals,
				    PARAM_IMAGE, gfig_image,
				    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);

}

DOBJECT *
PREFIXd_copy_circle(DOBJECT * obj)
{
  DOBJECT *nc;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == CIRCLE)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 8583,"obj->type == CIRCLE"); }else (void)0;

  nc = d_new_circle(obj->points->pnt.x,obj->points->pnt.y);

  nc->points->next = d_copy_dobjpoints(obj->points->next);







  return(nc);
}

DOBJECT *
PREFIXd_new_circle(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = CIRCLE;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_circle;
  nobj->loadfunc  = d_load_circle;
  nobj->savefunc  = d_save_circle;
  nobj->paintfunc = d_paint_circle;
  nobj->copyfunc  = d_copy_circle;

  return(nobj);
}

void
PREFIXd_update_circle(GdkPoint *pnt)
{
  DOBJPOINTS *center_pnt, *edge_pnt;
  gdouble radius;


  center_pnt = obj_creating->points;

  if(!center_pnt)
    return;

  if((edge_pnt = center_pnt->next))
    {

      draw_circle(&edge_pnt->pnt);
      radius = sqrt(((center_pnt->pnt.x - edge_pnt->pnt.x) *
			   (center_pnt->pnt.x - edge_pnt->pnt.x)) +
			  ((center_pnt->pnt.y - edge_pnt->pnt.y) *
			   (center_pnt->pnt.y - edge_pnt->pnt.y)));

      gdk_draw_arc (gfig_preview->window,
		    gfig_gc,
		    0,
		    center_pnt->pnt.x - (gint)rint(radius),
		    center_pnt->pnt.y - (gint)rint(radius),
		    (gint)rint(radius) * 2,
		    (gint)rint(radius) * 2,
		    0,
		    360*64);
    }

  draw_circle(pnt);

  edge_pnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));

  edge_pnt->pnt.x = pnt->x;
  edge_pnt->pnt.y = pnt->y;

  radius = sqrt(((center_pnt->pnt.x - edge_pnt->pnt.x) *
		       (center_pnt->pnt.x - edge_pnt->pnt.x)) +
		      ((center_pnt->pnt.y - edge_pnt->pnt.y) *
		       (center_pnt->pnt.y - edge_pnt->pnt.y)));

  gdk_draw_arc (gfig_preview->window,
		gfig_gc,
		0,
		center_pnt->pnt.x - (gint)rint(radius),
		center_pnt->pnt.y - (gint)rint(radius),
		(gint)rint(radius) * 2,
		(gint)rint(radius) * 2,
		0,
		360*64);

  center_pnt->next = edge_pnt;
}

void
PREFIXd_circle_start(GdkPoint *pnt,gint shift_down)
{
  obj_creating = d_new_circle(pnt->x, pnt->y);
}

void
PREFIXd_circle_end(GdkPoint *pnt, gint shift_down)
{

  if(!obj_creating->points->next)
    {

      free_one_obj(obj_creating);
    }
  else
    {
      draw_circle(pnt);
      add_to_all_obj(current_obj,obj_creating);
    }

  obj_creating = 0;
}



void
PREFIXd_save_ellipse(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<ELLIPSE>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"</ELLIPSE>\n");
}



DOBJECT *
PREFIXd_load_ellipse(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(strcmp("</ELLIPSE>",buf))
	    {
	      g_warning("[%d] Internal load error while loading ellipse",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_ellipse(xpnt,ypnt);
      else
	{
	  DOBJPOINTS *edge_pnt;

	  edge_pnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));

	  edge_pnt->pnt.x = xpnt;
	  edge_pnt->pnt.y = ypnt;

	  new_obj->points->next = edge_pnt;
	}
    }
  g_warning("[%d] Not enough points for ellipse",line_no);
  return(0);
}

static void
PREFIXd_draw_ellipse(DOBJECT * obj)
{
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * edge_pnt;
  gint bound_wx;
  gint bound_wy;
  gint top_x;
  gint top_y;

  center_pnt = obj->points;

  if(!center_pnt)
    return;

  edge_pnt = center_pnt->next;

  if(!edge_pnt)
    {
      g_warning("Internal error - ellipse no edge pnt");
    }

  draw_sqr(&center_pnt->pnt);
  draw_sqr(&edge_pnt->pnt);

  bound_wx = abs(center_pnt->pnt.x - edge_pnt->pnt.x)*2;
  bound_wy = abs(center_pnt->pnt.y - edge_pnt->pnt.y)*2;

  if(edge_pnt->pnt.x > center_pnt->pnt.x)
    top_x = 2*center_pnt->pnt.x - edge_pnt->pnt.x;
  else
    top_x = edge_pnt->pnt.x;

  if(edge_pnt->pnt.y > center_pnt->pnt.y)
    top_y = 2*center_pnt->pnt.y - edge_pnt->pnt.y;
  else
    top_y = edge_pnt->pnt.y;

  if(drawing_pic)
    {
      gdk_draw_arc (pic_preview->window,
		    pic_preview->style->black_gc,
		    0,
		    adjust_pic_coords(top_x,
				      preview_width),
		    adjust_pic_coords(top_y,
				      preview_height),
		    adjust_pic_coords(bound_wx,
				      preview_width),
		    adjust_pic_coords(bound_wy,
				      preview_height),
		    0,
		    360*64);
    }
  else
    {
      gdk_draw_arc (gfig_preview->window,
		    gfig_gc,
		    0,
		    gfig_scale_x(top_x),
		    gfig_scale_y(top_y),
		    gfig_scale_x(bound_wx),
		    gfig_scale_y(bound_wy),
		    0,
		    360*64);
    }
}

static void
PREFIXd_paint_approx_ellipse(DOBJECT *obj)
{


  gdouble *line_pnts;
  GParam *return_vals = 0;
  gint nreturn_vals;
  gint seg_count = 0;
  gint i = 0;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * radius_pnt;
  gdouble a_axis;
  gdouble b_axis;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gint loop;
  GdkPoint first_pnt,last_pnt;
  gint first = 1;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 8869,"obj != NULL"); }else (void)0;


  seg_count = 600;

  center_pnt = obj->points;

  if(!center_pnt || !seg_count)
    return;


  line_pnts = g_malloc0(1 *(2*seg_count + 1)*sizeof(gdouble));



  radius_pnt = center_pnt->next;


  a_axis = ((gdouble)(radius_pnt->pnt.x - center_pnt->pnt.x));
  b_axis = ((gdouble)(radius_pnt->pnt.y - center_pnt->pnt.y));


  ang_grid = 2* 3.14159265358979323846 /(gdouble)(gint)600;

  for(loop = 0 ; loop < (gint)600 ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid;

      radius = a_axis*b_axis/(sqrt(cos(ang_loop)*cos(ang_loop)*(b_axis*b_axis - a_axis*a_axis) + a_axis*a_axis));

      lx = radius * cos(ang_loop);
      ly = radius * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);


      if(!first)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      last_pnt.x = line_pnts[i++] = calc_pnt.x;
      last_pnt.y = line_pnts[i++] = calc_pnt.y;

      if(first)
	{
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	  first = 0;
	}
    }

  line_pnts[i++] = first_pnt.x;
  line_pnts[i++] = first_pnt.y;


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.scaletoimage)
    scale_to_original_xy(&line_pnts[0],i/2);
  else
    scale_to_xy(&line_pnts[0],i/2);


  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,(i/2)*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {

      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,(i/2)*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);

    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(line_pnts);
}



static void
PREFIXd_paint_ellipse(DOBJECT *obj)
{
  GParam *return_vals;
  gint nreturn_vals;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * edge_pnt;
  gint bound_wx;
  gint bound_wy;
  gint top_x;
  gint top_y;
  gdouble dpnts[4];


#line 9023 "gfig.c"

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 9024,"obj != NULL"); }else (void)0;

  if(selvals.approxcircles)
    {



      d_paint_approx_ellipse(obj);
      return;
    }

  center_pnt = obj->points;

  if(!center_pnt)
    return;

  edge_pnt = center_pnt->next;

  if(!edge_pnt)
    {
      g_error("Internal error - ellipse no edge pnt");
    }

  bound_wx = abs(center_pnt->pnt.x - edge_pnt->pnt.x)*2;
  bound_wy = abs(center_pnt->pnt.y - edge_pnt->pnt.y)*2;

  if(edge_pnt->pnt.x > center_pnt->pnt.x)
    top_x = 2*center_pnt->pnt.x - edge_pnt->pnt.x;
  else
    top_x = edge_pnt->pnt.x;

  if(edge_pnt->pnt.y > center_pnt->pnt.y)
    top_y = 2*center_pnt->pnt.y - edge_pnt->pnt.y;
  else
    top_y = edge_pnt->pnt.y;

  dpnts[0] = (gdouble)top_x;
  dpnts[1] = (gdouble)top_y;
  dpnts[2] = (gdouble)bound_wx;
  dpnts[3] = (gdouble)bound_wy;


  if(selvals.scaletoimage)
    scale_to_original_xy(&dpnts[0],2);
  else
    scale_to_xy(&dpnts[0],2);


  return_vals = gimp_run_procedure ("gimp_ellipse_select", &nreturn_vals,
				    PARAM_IMAGE, gfig_image,
				    PARAM_FLOAT,dpnts[0],
				    PARAM_FLOAT,dpnts[1],
				    PARAM_FLOAT,dpnts[2],
				    PARAM_FLOAT,dpnts[3],
				    PARAM_INT32,selopt.type,
				    PARAM_INT32,selopt.antia,
				    PARAM_INT32,selopt.feather,
				    PARAM_FLOAT,(gdouble)selopt.feather_radius,
				    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);


  if(selvals.painttype == PAINT_SELECTION_TYPE)
    return;

  return_vals = gimp_run_procedure ("gimp_edit_stroke", &nreturn_vals,
				    PARAM_DRAWABLE, gfig_drawable,
				    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);

  return_vals = gimp_run_procedure ("gimp_selection_clear", &nreturn_vals,
				    PARAM_IMAGE, gfig_image,
				    PARAM_END);

  gimp_destroy_params (return_vals, nreturn_vals);

}

DOBJECT *
PREFIXd_copy_ellipse(DOBJECT * obj)
{
  DOBJECT *nc;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == ELLIPSE)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 9115,"obj->type == ELLIPSE"); }else (void)0;

  nc = d_new_ellipse(obj->points->pnt.x,obj->points->pnt.y);

  nc->points->next = d_copy_dobjpoints(obj->points->next);







  return(nc);
}

DOBJECT *
PREFIXd_new_ellipse(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = ELLIPSE;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_ellipse;
  nobj->loadfunc  = d_load_ellipse;
  nobj->savefunc  = d_save_ellipse;
  nobj->paintfunc = d_paint_ellipse;
  nobj->copyfunc  = d_copy_ellipse;

  return(nobj);
}

void
PREFIXd_update_ellipse(GdkPoint *pnt)
{
  DOBJPOINTS *center_pnt, *edge_pnt;
  gint bound_wx;
  gint bound_wy;
  gint top_x;
  gint top_y;


  center_pnt = obj_creating->points;

  if(!center_pnt)
    return;


  if((edge_pnt = center_pnt->next))
    {

      bound_wx = abs(center_pnt->pnt.x - edge_pnt->pnt.x)*2;
      bound_wy = abs(center_pnt->pnt.y - edge_pnt->pnt.y)*2;

      if(edge_pnt->pnt.x > center_pnt->pnt.x)
	top_x = 2*center_pnt->pnt.x - edge_pnt->pnt.x;
      else
	top_x = edge_pnt->pnt.x;

      if(edge_pnt->pnt.y > center_pnt->pnt.y)
	top_y = 2*center_pnt->pnt.y - edge_pnt->pnt.y;
      else
	top_y = edge_pnt->pnt.y;

      draw_circle(&edge_pnt->pnt);

      gdk_draw_arc (gfig_preview->window,
		    gfig_gc,
		    0,
		    top_x,
		    top_y,
		    bound_wx,
		    bound_wy,
		    0,
		    360*64);
    }

  draw_circle(pnt);

  edge_pnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));

  edge_pnt->pnt.x = pnt->x;
  edge_pnt->pnt.y = pnt->y;

  bound_wx = abs(center_pnt->pnt.x - edge_pnt->pnt.x)*2;
  bound_wy = abs(center_pnt->pnt.y - edge_pnt->pnt.y)*2;

  if(edge_pnt->pnt.x > center_pnt->pnt.x)
    top_x = 2*center_pnt->pnt.x - edge_pnt->pnt.x;
  else
    top_x = edge_pnt->pnt.x;

  if(edge_pnt->pnt.y > center_pnt->pnt.y)
    top_y = 2* center_pnt->pnt.y - edge_pnt->pnt.y;
  else
    top_y = edge_pnt->pnt.y;

  gdk_draw_arc (gfig_preview->window,
		gfig_gc,
		0,
		top_x,
		top_y,
		bound_wx,
		bound_wy,
		0,
		360*64);

  center_pnt->next = edge_pnt;
}

void
PREFIXd_ellipse_start(GdkPoint *pnt,gint shift_down)
{
  obj_creating = d_new_ellipse(pnt->x, pnt->y);
}

void
PREFIXd_ellipse_end(GdkPoint *pnt, gint shift_down)
{

  if(!obj_creating->points->next)
    {

      free_one_obj(obj_creating);
    }
  else
    {
      draw_circle(pnt);
      add_to_all_obj(current_obj,obj_creating);
    }

  obj_creating = 0;
}



void
PREFIXd_save_poly(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<POLY>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"<EXTRA>\n");
  fprintf(to,"%d\n</EXTRA>\n",(gint)obj->type_data);
  fprintf(to,"</POLY>\n");

}



DOBJECT *
PREFIXd_load_poly(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(!strcmp("<EXTRA>",buf))
	    {
	      gint nsides = 3;

	      if(!new_obj)
		{
		  g_warning("[%d] Internal load error while loading poly (extra area)",
			    line_no);
		  return(0);
		}
	      get_line(buf,256,from,0);
	      if(sscanf(buf,"%d",&nsides) != 1)
		{
		  g_warning("[%d] Internal load error while loading poly (extra area scanf)",
			    line_no);
		  return(0);
		}
	      new_obj->type_data = (gpointer)nsides;
	      get_line(buf,256,from,0);
	      if(strcmp("</EXTRA>",buf))
		{
		  g_warning("[%d] Internal load error while loading poly",
			    line_no);
		  return(0);
		}

	      continue;
	    }
	  else if(strcmp("</POLY>",buf))
	    {
	      g_warning("[%d] Internal load error while loading poly",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_poly(xpnt,ypnt);
      else
	d_pnt_add_line(new_obj,xpnt,ypnt,-1);
    }
  return(new_obj);
}

static void
PREFIXd_draw_poly(DOBJECT *obj)
{
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gdouble offset_angle;
  gint loop;
  GdkPoint start_pnt;
  GdkPoint first_pnt;
  gint do_line = 0;

  center_pnt = obj->points;

  if(!center_pnt)
    return;




  draw_sqr(&center_pnt->pnt);


  radius_pnt = center_pnt->next;

  if(!radius_pnt)
    {



      return;
    }


  draw_sqr(&radius_pnt->pnt);



  shift_x = radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = radius_pnt->pnt.y - center_pnt->pnt.y;

  radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));


  ang_grid = 2* 3.14159265358979323846 /(gdouble)(gint)obj->type_data;
  offset_angle = atan2(shift_y,shift_x);

  for(loop = 0 ; loop < (gint)obj->type_data ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + offset_angle;

      lx = radius * cos(ang_loop);
      ly = radius * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);

      if(do_line)
	{


	  if(calc_pnt.x == start_pnt.x && calc_pnt.y == start_pnt.y)
	    continue;

	  if(drawing_pic)
	    {
	      gdk_draw_line(pic_preview->window,
			    pic_preview->style->black_gc,
			    adjust_pic_coords(calc_pnt.x,
					      preview_width),
			    adjust_pic_coords(calc_pnt.y,
					      preview_height),
			    adjust_pic_coords(start_pnt.x,
					      preview_width),
			    adjust_pic_coords(start_pnt.y,
					      preview_height));
	    }
	  else
	    {
	      gdk_draw_line(gfig_preview->window,
			    gfig_gc,
			    gfig_scale_x(calc_pnt.x),
			    gfig_scale_y(calc_pnt.y),
			    gfig_scale_x(start_pnt.x),
			    gfig_scale_y(start_pnt.y));
	    }
	}
      else
	{
	  do_line = 1;
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	}
      start_pnt.x = calc_pnt.x;
      start_pnt.y = calc_pnt.y;
    }


  if(drawing_pic)
    {
      gdk_draw_line(pic_preview->window,
		    pic_preview->style->black_gc,
		adjust_pic_coords(first_pnt.x,preview_width),
		adjust_pic_coords(first_pnt.y,preview_width),
		adjust_pic_coords(start_pnt.x,preview_width),
		adjust_pic_coords(start_pnt.y,preview_width));
    }
  else
    {
      gdk_draw_line(gfig_preview->window,
		gfig_gc,
		gfig_scale_x(first_pnt.x),
		gfig_scale_y(first_pnt.y),
		gfig_scale_x(start_pnt.x),
		gfig_scale_y(start_pnt.y));
    }
}

static void
PREFIXd_paint_poly(DOBJECT *obj)
{


  gdouble *line_pnts;
  GParam *return_vals = 0;
  gint nreturn_vals;
  gint seg_count = 0;
  gint i = 0;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gdouble offset_angle;
  gint loop;
  GdkPoint first_pnt,last_pnt;
  gint first = 1;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 9501,"obj != NULL"); }else (void)0;


  seg_count = (gint)obj->type_data + 1;

  center_pnt = obj->points;

  if(!center_pnt || !seg_count || !center_pnt->next)
    return;


  line_pnts = g_malloc0(1 *(2*seg_count + 1)*sizeof(gdouble));



  radius_pnt = center_pnt->next;


  shift_x = radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = radius_pnt->pnt.y - center_pnt->pnt.y;

  radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));


  ang_grid = 2* 3.14159265358979323846 /(gdouble)(gint)obj->type_data;
  offset_angle = atan2(shift_y,shift_x);

  for(loop = 0 ; loop < (gint)obj->type_data ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + offset_angle;

      lx = radius * cos(ang_loop);
      ly = radius * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);


      if(!first)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      last_pnt.x = line_pnts[i++] = calc_pnt.x;
      last_pnt.y = line_pnts[i++] = calc_pnt.y;

      if(first)
	{
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	  first = 0;
	}
    }

  line_pnts[i++] = first_pnt.x;
  line_pnts[i++] = first_pnt.y;


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.scaletoimage)
    scale_to_original_xy(&line_pnts[0],i/2);
  else
    scale_to_xy(&line_pnts[0],i/2);


  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,(i/2)*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {

      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,(i/2)*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);
    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(line_pnts);
}

static void
PREFIXd_poly2lines(DOBJECT *obj)
{


  gint seg_count = 0;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gdouble offset_angle;
  gint loop;
  GdkPoint first_pnt,last_pnt;
  gint first = 1;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 9654,"obj != NULL"); }else (void)0;






  seg_count = (gint)obj->type_data + 1;

  center_pnt = obj->points;

  if(!center_pnt)
    return;


  obj->drawfunc(obj);


  obj->points = 0;



  radius_pnt = center_pnt->next;


  shift_x = radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = radius_pnt->pnt.y - center_pnt->pnt.y;

  radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));


  ang_grid = 2* 3.14159265358979323846 /(gdouble)(gint)obj->type_data;
  offset_angle = atan2(shift_y,shift_x);

  for(loop = 0 ; loop < (gint)obj->type_data ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + offset_angle;

      lx = radius * cos(ang_loop);
      ly = radius * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);

      if(!first)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      d_pnt_add_line(obj,calc_pnt.x,calc_pnt.y,0);

      last_pnt.x = calc_pnt.x;
      last_pnt.y = calc_pnt.y;

      if(first)
	{
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	  first = 0;
	}
    }

  d_pnt_add_line(obj,first_pnt.x,first_pnt.y,0);

  d_delete_dobjpoints(center_pnt);


  obj->type = LINE;
  obj->drawfunc  = d_draw_line;
  obj->loadfunc  = d_load_line;
  obj->savefunc  = d_save_line;
  obj->paintfunc = d_paint_line;
  obj->copyfunc  = d_copy_line;


  obj->drawfunc(obj);
}

static void
PREFIXd_star2lines(DOBJECT *obj)
{


  gint seg_count = 0;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * outer_radius_pnt;
  DOBJPOINTS * inner_radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble outer_radius;
  gdouble inner_radius;
  gdouble offset_angle;
  gint loop;
  GdkPoint first_pnt,last_pnt;
  gint first = 1;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 9758,"obj != NULL"); }else (void)0;






  seg_count = 2*(gint)obj->type_data + 1;

  center_pnt = obj->points;

  if(!center_pnt)
    return;


  obj->drawfunc(obj);


  obj->points = 0;



  outer_radius_pnt = center_pnt->next;

  if(!outer_radius_pnt)
    {



      return;
    }

  inner_radius_pnt = outer_radius_pnt->next;

  if(!inner_radius_pnt)
    {



      return;
    }

  shift_x = outer_radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = outer_radius_pnt->pnt.y - center_pnt->pnt.y;

  outer_radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));


  ang_grid = 2* 3.14159265358979323846 /(2.0*(gdouble)(gint)obj->type_data);
  offset_angle = atan2(shift_y,shift_x);

  shift_x = inner_radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = inner_radius_pnt->pnt.y - center_pnt->pnt.y;

  inner_radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));

  for(loop = 0 ; loop < 2*(gint)obj->type_data ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + offset_angle;

      if(loop%2)
	{
	  lx = inner_radius * cos(ang_loop);
	  ly = inner_radius * sin(ang_loop);
	}
      else
	{
	  lx = outer_radius * cos(ang_loop);
	  ly = outer_radius * sin(ang_loop);
	}

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);

      if(!first)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      d_pnt_add_line(obj,calc_pnt.x,calc_pnt.y,0);

      last_pnt.x = calc_pnt.x;
      last_pnt.y = calc_pnt.y;

      if(first)
	{
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	  first = 0;
	}
    }

  d_pnt_add_line(obj,first_pnt.x,first_pnt.y,0);

  d_delete_dobjpoints(center_pnt);


  obj->type = LINE;
  obj->drawfunc  = d_draw_line;
  obj->loadfunc  = d_load_line;
  obj->savefunc  = d_save_line;
  obj->paintfunc = d_paint_line;
  obj->copyfunc  = d_copy_line;


  obj->drawfunc(obj);
}

DOBJECT *
PREFIXd_copy_poly(DOBJECT * obj)
{
  DOBJECT *np;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == POLY)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 9883,"obj->type == POLY"); }else (void)0;

  np = d_new_poly(obj->points->pnt.x,obj->points->pnt.y);

  np->points->next = d_copy_dobjpoints(obj->points->next);

  np->type_data = obj->type_data;




  return(np);
}

DOBJECT *
PREFIXd_new_poly(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = POLY;
  nobj->type_data = (gpointer)3;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_poly;
  nobj->loadfunc  = d_load_poly;
  nobj->savefunc  = d_save_poly;
  nobj->paintfunc = d_paint_poly;
  nobj->copyfunc  = d_copy_poly;

  return(nobj);
}

void
PREFIXd_update_poly(GdkPoint *pnt)
{
  DOBJPOINTS *center_pnt, *edge_pnt;
  gint saved_cnt_pnt = selvals.opts.showcontrol;


  center_pnt = obj_creating->points;

  if(!center_pnt)
    return;


#line 9944 "gfig.c"


#line 9949 "gfig.c"


  if((edge_pnt = center_pnt->next))
    {

      draw_circle(&edge_pnt->pnt);
      selvals.opts.showcontrol = 0;
      d_draw_poly(obj_creating);

      edge_pnt->pnt.x = pnt->x;
      edge_pnt->pnt.y = pnt->y;
    }
  else
    {


      d_pnt_add_line(obj_creating,pnt->x,pnt->y,-1);
      edge_pnt = center_pnt->next;
    }


  selvals.opts.showcontrol = 0;
  d_draw_poly(obj_creating);
  selvals.opts.showcontrol = saved_cnt_pnt;


  draw_circle(&edge_pnt->pnt);
}


#line 9981 "gfig.c"

void
PREFIXd_poly_start(GdkPoint *pnt,gint shift_down)
{
  gint16 x,y;

  obj_creating = d_new_poly(x = pnt->x, y = pnt->y);
  obj_creating->type_data = (gpointer)poly_num_sides;
}

void
PREFIXd_poly_end(GdkPoint *pnt, gint shift_down)
{
  draw_circle(pnt);
  add_to_all_obj(current_obj,obj_creating);
  obj_creating = 0;
}



static double
PREFIXdist(double x1,double y1, double x2, double y2)
{

  double s1 = x1 - x2;
  double s2 = y1 - y2;

  return(sqrt((s1*s1) + (s2*s2)));
}


static void
PREFIXmid_point(double x1,double y1,double x2, double y2,double *mx,double *my)
{
  *mx = ((double)(x1 - x2))/2.0 + (double)x2;
  *my = ((double)(y1 - y2))/2.0 + (double)y2;
}


static double
PREFIXline_grad(double x1,double y1, double x2, double y2)
{
  double dx,dy;

  dx = x1 - x2;
  dy = y1 - y2;

  if(dx == 0.0)
    return (0.0);

  return(dy/dx);
}


static double
PREFIXline_cons(double x, double y, double lgrad)
{
  return(y - lgrad*x);
}


static void
PREFIXline_definition(double x1, double y1, double x2, double y2, double *lgrad, double *lconst)
{
  double grad1;
  double midx,midy;

  grad1 = line_grad(x1,y1,x2,y2);

  if(grad1 == 0.0)
    {



      return;
    }

  mid_point(x1,y1,x2,y2,&midx,&midy);



  *lgrad = -1.0/grad1;

  *lconst = line_cons(midx,midy,*lgrad);
}


#line 10071 "gfig.c"

static void
PREFIXarc_details(GdkPoint *vert_a,GdkPoint *vert_b, GdkPoint *vert_c,GdkPoint *center_pnt, gdouble *radius)
{

  double ax,ay;
  double bx,by;
  double cx,cy;

  double len_a,len_b,len_c;
  double sum_sides2;
  double area;
  double circumcircle_R;
  double line1_grad,line1_const;
  double line2_grad,line2_const;
  double inter_x=0.0,inter_y=0.0;
  int got_x=0,got_y=0;

  ax = (double)(vert_a->x);
  ay = (double)(vert_a->y);
  bx = (double)(vert_b->x);
  by = (double)(vert_b->y);
  cx = (double)(vert_c->x);
  cy = (double)(vert_c->y);





  len_a = dist(ax,ay,bx,by);
  len_b = dist(bx,by,cx,cy);
  len_c = dist(cx,cy,ax,ay);





  sum_sides2 = (fabs(len_a) + fabs(len_b) + fabs(len_c))/2;





  area = sqrt(sum_sides2*(sum_sides2 - len_a)*(sum_sides2 - len_b)*(sum_sides2 - len_c));





  circumcircle_R = len_a*len_b*len_c/(4*area);
  *radius = circumcircle_R;






  if(ax == bx || ax == cx || cx == bx)
    {

      if( ax == bx && bx == cx)
	{

	  double miny = ay;
	  double maxy = ay;

	  if(by > maxy)
	    maxy = by;

	  if(by < miny)
	    miny = by;

	  if(cy > maxy)
	    maxy = cy;

	  if(cy < miny)
	    miny = cy;

	  inter_y = (maxy - miny)/2 + miny;
	}
      else if(ax == bx)
	{
	  inter_y = (ay - by)/2 + by;
	}
      else if(bx == cx)
	{
	  inter_y = (by - cy)/2 + cy;
	}
      else
	{
	  inter_y = (cy - ay)/2 + ay;
	}
      got_y = 1;
    }

  if(ay == by || by == cy || ay == cy)
    {

      if( ax == bx && bx == cx)
	{

	  double minx = ax;
	  double maxx = ax;

	  if(bx > maxx)
	    maxx = bx;

	  if(bx < minx)
	    minx = bx;

	  if(cx > maxx)
	    maxx = cx;

	  if(cx < minx)
	    minx = cx;

	  inter_x = (maxx - minx)/2 + minx;
	}
      else if(ay == by)
	{
	  inter_x = (ax - bx)/2 + bx;
	}
      else if(by == cy)
	{
	  inter_x = (bx - cx)/2 + cx;
	}
      else
	{
	  inter_x = (cx - ax)/2 + ax;
	}
      got_x = 1;
    }

  if(!got_x || !got_y)
    {


      if(ax != bx && ay != by)
	line_definition(ax,ay,bx,by,&line1_grad,&line1_const);
      else
	line_definition(ax,ay,cx,cy,&line1_grad,&line1_const);

      if(bx != cx && by != cy)
	line_definition(bx,by,cx,cy,&line2_grad,&line2_const);
      else
	line_definition(ax,ay,cx,cy,&line2_grad,&line2_const);
    }



  if(!got_x)
    inter_x = ((line2_const - line1_const)/(line1_grad - line2_grad));
  if(!got_y)
    inter_y = ((line1_grad * inter_x + line1_const));





  center_pnt->x = (gint16)inter_x;
  center_pnt->y = (gint16)inter_y;
}

static gdouble
PREFIXarc_angle(GdkPoint *pnt, GdkPoint *center)
{

  gint16 shift_x;
  gint16 shift_y;
  gdouble offset_angle;

  shift_x = pnt->x - center->x;
  shift_y = -pnt->y + center->y;
  offset_angle = atan2(shift_y,shift_x);



  if(offset_angle < 0)
    offset_angle += 2* 3.14159265358979323846;

  return(offset_angle*360/(2* 3.14159265358979323846));
}

void
PREFIXd_save_arc(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<ARC>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"</ARC>\n");
}



DOBJECT *
PREFIXd_load_arc(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];
  gint num_pnts = 0;





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(strcmp("</ARC>",buf) || num_pnts != 3)
	    {
	      g_warning("[%d] Internal load error while loading arc",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      num_pnts++;

      if(!new_obj)
	new_obj = d_new_arc(xpnt,ypnt);
      else
	{
	  d_pnt_add_line(new_obj,xpnt,ypnt,-1);
	}
    }
  g_warning("[%d] Not enough points for arc",line_no);
  return(0);
}

static void
PREFIXarc_drawing_details(DOBJECT *obj,
		    gdouble *minang,
		    GdkPoint *center_pnt,
		    gdouble *arcang,
		    gdouble *radius,
		    gint draw_cnts,
		    gint do_scale)
{
  DOBJPOINTS * pnt1 = 0;
  DOBJPOINTS * pnt2 = 0;
  DOBJPOINTS * pnt3 = 0;
  DOBJPOINTS dpnts[3];
  gdouble ang1,ang2,ang3;
  gdouble maxang;

  pnt1 = obj->points;

  if(!pnt1)
    return;

  pnt2 = pnt1->next;

  if(!pnt2)
    return;

  pnt3 = pnt2->next;

  if(!pnt3)
    return;

  if(draw_cnts)
    {
      draw_sqr(&pnt1->pnt);
      draw_sqr(&pnt2->pnt);
      draw_sqr(&pnt3->pnt);
    }

  if(do_scale)
    {



      gdouble xy[2];
      int j;

      dpnts[0] = *pnt1;
      dpnts[1] = *pnt2;
      dpnts[2] = *pnt3;

      pnt1 = &dpnts[0];
      pnt2 = &dpnts[1];
      pnt3 = &dpnts[2];

      for(j = 0 ; j < 3; j++)
	{
	  xy[0] = dpnts[j].pnt.x;
	  xy[1] = dpnts[j].pnt.y;
	  if(selvals.scaletoimage)
	    scale_to_original_xy(&xy[0],1);
	  else
	    scale_to_xy(&xy[0],1);
	  dpnts[j].pnt.x = xy[0];
	  dpnts[j].pnt.y = xy[1];
	}
    }

  arc_details(&pnt1->pnt,&pnt2->pnt,&pnt3->pnt,center_pnt,radius);

  ang1 = arc_angle(&pnt1->pnt,center_pnt);
  ang2 = arc_angle(&pnt2->pnt,center_pnt);
  ang3 = arc_angle(&pnt3->pnt,center_pnt);



  maxang = ang1;

  if(ang3 > maxang)
    maxang = ang3;

  *minang = ang1;

  if(ang3 < *minang)
    *minang = ang3;

  if (ang2 > *minang && ang2 < maxang)
    *arcang = maxang - *minang;
  else
    *arcang = maxang - *minang - 360;
}

static void
PREFIXd_draw_arc(DOBJECT * obj)
{
  GdkPoint center_pnt;
  gdouble radius,minang,arcang;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 10416,"obj != NULL"); }else (void)0;

  if(!obj)
    return;

  arc_drawing_details(obj,&minang,&center_pnt,&arcang,&radius,(!(0)),(0));





  if(drawing_pic)
    {
      gdk_draw_arc (pic_preview->window,
		    pic_preview->style->black_gc,
		    0,
		    adjust_pic_coords(center_pnt.x - (gint)radius,
				      preview_width),
		    adjust_pic_coords(center_pnt.y - (gint)radius,
				      preview_height),
		    adjust_pic_coords((gint)(radius * 2),
				      preview_width),
		    adjust_pic_coords((gint)(radius * 2),
				      preview_height),
		    (gint)(minang*64),
		    (gint)(arcang*64));
    }
  else
    {
      gdk_draw_arc (gfig_preview->window,
		    gfig_gc,
		    0,
		    gfig_scale_x(center_pnt.x - (gint)radius),
		    gfig_scale_y(center_pnt.y - (gint)radius),
		    gfig_scale_x((gint)(radius * 2)),
		    gfig_scale_y((gint)(radius * 2)),
		    (gint)(minang*64),
		    (gint)(arcang*64));
    }
}

static void
PREFIXd_paint_arc(DOBJECT *obj)
{


  gdouble *line_pnts;
  GParam *return_vals = 0;
  gint nreturn_vals;
  gint seg_count = 0;
  gint i = 0;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gint loop;
  GdkPoint first_pnt,last_pnt;
  gint first = 1;
  GdkPoint center_pnt;
  gdouble minang,arcang;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 10476,"obj != NULL"); }else (void)0;

  if(!obj)
    return;


  arc_drawing_details(obj,&minang,&center_pnt,&arcang,&radius,(0),(!(0)));





  seg_count = 360;



  line_pnts = g_malloc0(1 *(2*seg_count + 3)*sizeof(gdouble));


  ang_grid = 2* 3.14159265358979323846 /(gdouble)360;

  if(arcang < 0.0)
    {

      minang += arcang;
      arcang = -arcang;
    }

  minang = minang * (2* 3.14159265358979323846 /360);

  for(loop = 0 ; loop < abs((gint)arcang) ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + minang;

      lx = radius * cos(ang_loop);
      ly = -radius * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt.y);


      if(!first)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      last_pnt.x = line_pnts[i++] = calc_pnt.x;
      last_pnt.y = line_pnts[i++] = calc_pnt.y;

      if(first)
	{
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	  first = 0;
	}
    }


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,(i/2)*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {
      if(selopt.as_pie)
	{

	  line_pnts[i++] = center_pnt.x;
	  line_pnts[i++] = center_pnt.y;
	}


      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,(i/2)*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);
    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(line_pnts);
}

DOBJECT *
PREFIXd_copy_arc(DOBJECT * obj)
{
  DOBJECT *nc;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == ARC)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 10623,"obj->type == ARC"); }else (void)0;

  nc = d_new_arc(obj->points->pnt.x,obj->points->pnt.y);

  nc->points->next = d_copy_dobjpoints(obj->points->next);








  return(nc);
}

DOBJECT *
PREFIXd_new_arc(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = ARC;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_arc;
  nobj->loadfunc  = d_load_arc;
  nobj->savefunc  = d_save_arc;
  nobj->paintfunc = d_paint_arc;
  nobj->copyfunc  = d_copy_arc;

  return(nobj);
}

void
PREFIXd_update_arc(GdkPoint *pnt)
{
  DOBJPOINTS * pnt1 = 0;
  DOBJPOINTS * pnt2 = 0;
  DOBJPOINTS * pnt3 = 0;


#line 10679 "gfig.c"

  pnt1 = obj_creating->points;

  if(!pnt1 ||
     !(pnt2 = pnt1->next) ||
     !(pnt3 = pnt2->next))
    {
      d_update_line(pnt);
      return;
    }



}

void
PREFIXd_arc_start(GdkPoint *pnt,gint shift_down)
{

  if(!tmp_line)
    draw_sqr(pnt);
  d_line_start(pnt,(!(0)));
}

void
PREFIXd_arc_end(GdkPoint *pnt, gint shift_down)
{

  if(!tmp_line ||
     !tmp_line->points ||
     !tmp_line->points->next)
    {





      d_line_end(pnt,(!(0)));
    }
  else
    {


      tmp_line->type = ARC;
      tmp_line->drawfunc  = d_draw_arc;
      tmp_line->loadfunc  = d_load_arc;
      tmp_line->savefunc  = d_save_arc;
      tmp_line->paintfunc = d_paint_arc;
      tmp_line->copyfunc  = d_copy_arc;
      d_line_end(pnt,(0));

      if(need_to_scale)
	{
	  selvals.scaletoimage = 0;
	}

      update_draw_area(gfig_preview,0);
      if(need_to_scale)
	{
	  selvals.scaletoimage = 1;
	}

    }
}



void
PREFIXd_save_star(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<STAR>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"<EXTRA>\n");
  fprintf(to,"%d\n</EXTRA>\n",(gint)obj->type_data);
  fprintf(to,"</STAR>\n");
}



DOBJECT *
PREFIXd_load_star(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(!strcmp("<EXTRA>",buf))
	    {
	      gint nsides = 3;

	      if(!new_obj)
		{
		  g_warning("[%d] Internal load error while loading star (extra area)",
			    line_no);
		  return(0);
		}
	      get_line(buf,256,from,0);
	      if(sscanf(buf,"%d",&nsides) != 1)
		{
		  g_warning("[%d] Internal load error while loading star (extra area scanf)",
			    line_no);
		  return(0);
		}
	      new_obj->type_data = (gpointer)nsides;
	      get_line(buf,256,from,0);
	      if(strcmp("</EXTRA>",buf))
		{
		  g_warning("[%d] Internal load error while loading star",
			    line_no);
		  return(0);
		}

	      continue;
	    }
	  else if(strcmp("</STAR>",buf))
	    {
	      g_warning("[%d] Internal load error while loading star",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_star(xpnt,ypnt);
      else
	d_pnt_add_line(new_obj,xpnt,ypnt,-1);
    }
  return(new_obj);
}

static void
PREFIXd_draw_star(DOBJECT *obj)
{
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * outer_radius_pnt;
  DOBJPOINTS * inner_radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble outer_radius;
  gdouble inner_radius;
  gdouble offset_angle;
  gint loop;
  GdkPoint start_pnt;
  GdkPoint first_pnt;
  gint do_line = 0;

  center_pnt = obj->points;

  if(!center_pnt)
    return;




  draw_sqr(&center_pnt->pnt);


  outer_radius_pnt = center_pnt->next;

  if(!outer_radius_pnt)
    {



      return;
    }

  inner_radius_pnt = outer_radius_pnt->next;

  if(!inner_radius_pnt)
    {



      return;
    }


  draw_sqr(&outer_radius_pnt->pnt);
  draw_sqr(&inner_radius_pnt->pnt);



  shift_x = outer_radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = outer_radius_pnt->pnt.y - center_pnt->pnt.y;

  outer_radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));


  ang_grid = 2* 3.14159265358979323846 /(2.0*(gdouble)(gint)obj->type_data);
  offset_angle = atan2(shift_y,shift_x);

  shift_x = inner_radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = inner_radius_pnt->pnt.y - center_pnt->pnt.y;

  inner_radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));

  for(loop = 0 ; loop < 2*(gint)obj->type_data ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + offset_angle;

      if(loop%2)
	{
	  lx = inner_radius * cos(ang_loop);
	  ly = inner_radius * sin(ang_loop);
	}
      else
	{
	  lx = outer_radius * cos(ang_loop);
	  ly = outer_radius * sin(ang_loop);
	}

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);

      if(do_line)
	{


	  if(calc_pnt.x == start_pnt.x && calc_pnt.y == start_pnt.y)
	    continue;

	  if(drawing_pic)
	    {
	      gdk_draw_line(pic_preview->window,
			    pic_preview->style->black_gc,
			    adjust_pic_coords(calc_pnt.x,
					      preview_width),
			    adjust_pic_coords(calc_pnt.y,
					      preview_height),
			    adjust_pic_coords(start_pnt.x,
					      preview_width),
			    adjust_pic_coords(start_pnt.y,
					      preview_height));
	    }
	  else
	    {
      gdk_draw_line(gfig_preview->window,
			    gfig_gc,
			    gfig_scale_x(calc_pnt.x),
			    gfig_scale_y(calc_pnt.y),
			    gfig_scale_x(start_pnt.x),
			    gfig_scale_y(start_pnt.y));
	    }
	}
      else
	{
	  do_line = 1;
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	}
      start_pnt.x = calc_pnt.x;
      start_pnt.y = calc_pnt.y;
    }


  if(drawing_pic)
    {
      gdk_draw_line(pic_preview->window,
		    pic_preview->style->black_gc,
		adjust_pic_coords(first_pnt.x,preview_width),
		adjust_pic_coords(first_pnt.y,preview_width),
		adjust_pic_coords(start_pnt.x,preview_width),
		adjust_pic_coords(start_pnt.y,preview_width));
    }
  else
    {
      gdk_draw_line(gfig_preview->window,
		gfig_gc,
		gfig_scale_x(first_pnt.x),
		gfig_scale_y(first_pnt.y),
		gfig_scale_x(start_pnt.x),
		gfig_scale_y(start_pnt.y));
    }
}

static void
PREFIXd_paint_star(DOBJECT *obj)
{


  gdouble *line_pnts;
  GParam *return_vals = 0;
  gint nreturn_vals;
  gint seg_count = 0;
  gint i = 0;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * outer_radius_pnt;
  DOBJPOINTS * inner_radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble outer_radius;
  gdouble inner_radius;

  gdouble offset_angle;
  gint loop;
  GdkPoint first_pnt,last_pnt;
  gint first = 1;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 11011,"obj != NULL"); }else (void)0;


  seg_count = 2*(gint)obj->type_data + 1;

  center_pnt = obj->points;

  if(!center_pnt || !seg_count)
    return;


  line_pnts = g_malloc0(1 *(2*seg_count + 1)*sizeof(gdouble));



  outer_radius_pnt = center_pnt->next;

  if(!outer_radius_pnt)
    {



      return;
    }

  inner_radius_pnt = outer_radius_pnt->next;

  if(!inner_radius_pnt)
    {



      return;
    }

  shift_x = outer_radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = outer_radius_pnt->pnt.y - center_pnt->pnt.y;

  outer_radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));


  ang_grid = 2* 3.14159265358979323846 /(2.0*(gdouble)(gint)obj->type_data);
  offset_angle = atan2(shift_y,shift_x);

  shift_x = inner_radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = inner_radius_pnt->pnt.y - center_pnt->pnt.y;

  inner_radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));

  for(loop = 0 ; loop < 2*(gint)obj->type_data ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid + offset_angle;

      if(loop%2)
	{
	  lx = inner_radius * cos(ang_loop);
	  ly = inner_radius * sin(ang_loop);
	}
      else
	{
	  lx = outer_radius * cos(ang_loop);
	  ly = outer_radius * sin(ang_loop);
	}

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);


      if(!first)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      last_pnt.x = line_pnts[i++] = calc_pnt.x;
      last_pnt.y = line_pnts[i++] = calc_pnt.y;

      if(first)
	{
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	  first = 0;
	}
    }

  line_pnts[i++] = first_pnt.x;
  line_pnts[i++] = first_pnt.y;


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.scaletoimage)
    scale_to_original_xy(&line_pnts[0],i/2);
  else
    scale_to_xy(&line_pnts[0],i/2);



  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,(i/2)*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {

      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,(i/2)*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);

    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(line_pnts);
}

DOBJECT *
PREFIXd_copy_star(DOBJECT * obj)
{
  DOBJECT *np;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == STAR)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 11189,"obj->type == STAR"); }else (void)0;

  np = d_new_star(obj->points->pnt.x,obj->points->pnt.y);

  np->points->next = d_copy_dobjpoints(obj->points->next);

  np->type_data = obj->type_data;




  return(np);
}

DOBJECT *
PREFIXd_new_star(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = STAR;
  nobj->type_data = (gpointer)3;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_star;
  nobj->loadfunc  = d_load_star;
  nobj->savefunc  = d_save_star;
  nobj->paintfunc = d_paint_star;
  nobj->copyfunc  = d_copy_star;

  return(nobj);
}

void
PREFIXd_update_star(GdkPoint *pnt)
{
  DOBJPOINTS *center_pnt, *inner_pnt, *outer_pnt;
  gint saved_cnt_pnt = selvals.opts.showcontrol;


  center_pnt = obj_creating->points;

  if(!center_pnt)
    return;


#line 11250 "gfig.c"


#line 11255 "gfig.c"


  if((outer_pnt = center_pnt->next))
    {

      inner_pnt = outer_pnt->next;
      draw_circle(&inner_pnt->pnt);
      draw_circle(&outer_pnt->pnt);
      selvals.opts.showcontrol = 0;
      d_draw_star(obj_creating);
      outer_pnt->pnt.x = pnt->x;
      outer_pnt->pnt.y = pnt->y;
      inner_pnt->pnt.x = pnt->x + (2*(center_pnt->pnt.x - pnt->x))/3;
      inner_pnt->pnt.y = pnt->y + (2*(center_pnt->pnt.y - pnt->y))/3;
    }
  else
    {


      d_pnt_add_line(obj_creating,pnt->x,pnt->y,-1);
      outer_pnt = center_pnt->next;

      d_pnt_add_line(obj_creating,
		     pnt->x + (2*(center_pnt->pnt.x - pnt->x))/3,
		     pnt->y + (2*(center_pnt->pnt.y - pnt->y))/3,
		     -1);
      inner_pnt = outer_pnt->next;
    }


  selvals.opts.showcontrol = 0;
  d_draw_star(obj_creating);
  selvals.opts.showcontrol = saved_cnt_pnt;


  draw_circle(&outer_pnt->pnt);
  draw_circle(&inner_pnt->pnt);
}


#line 11297 "gfig.c"

void
PREFIXd_star_start(GdkPoint *pnt,gint shift_down)
{
  gint16 x,y;

  obj_creating = d_new_star(x = pnt->x, y = pnt->y);
  obj_creating->type_data = (gpointer)star_num_sides;
}

void
PREFIXd_star_end(GdkPoint *pnt, gint shift_down)
{
  draw_circle(pnt);
  add_to_all_obj(current_obj,obj_creating);
  obj_creating = 0;
}




void
PREFIXd_save_spiral(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<SPIRAL>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"<EXTRA>\n");
  fprintf(to,"%d\n</EXTRA>\n",(gint)obj->type_data);
  fprintf(to,"</SPIRAL>\n");

}



DOBJECT *
PREFIXd_load_spiral(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(!strcmp("<EXTRA>",buf))
	    {
	      gint nsides = 3;

	      if(!new_obj)
		{
		  g_warning("[%d] Internal load error while loading spiral (extra area)",
			    line_no);
		  return(0);
		}
	      get_line(buf,256,from,0);
	      if(sscanf(buf,"%d",&nsides) != 1)
		{
		  g_warning("[%d] Internal load error while loading spiral (extra area scanf)",
			    line_no);
		  return(0);
		}
	      new_obj->type_data = (gpointer)nsides;
	      get_line(buf,256,from,0);
	      if(strcmp("</EXTRA>",buf))
		{
		  g_warning("[%d] Internal load error while loading spiral",
			    line_no);
		  return(0);
		}

	      continue;
	    }
	  else if(strcmp("</SPIRAL>",buf))
	    {
	      g_warning("[%d] Internal load error while loading spiral",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_spiral(xpnt,ypnt);
      else
	d_pnt_add_line(new_obj,xpnt,ypnt,-1);
    }
  return(new_obj);
}

static void
PREFIXd_draw_spiral(DOBJECT *obj)
{
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gdouble offset_angle;
  gdouble sp_cons;
  gint loop;
  GdkPoint start_pnt;
  GdkPoint first_pnt;
  gint do_line = 0;
  gint clock_wise = 1;

  center_pnt = obj->points;

  if(!center_pnt)
    return;




  draw_sqr(&center_pnt->pnt);


  radius_pnt = center_pnt->next;

  if(!radius_pnt)
    {



      return;
    }


  draw_sqr(&radius_pnt->pnt);



  shift_x = radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = radius_pnt->pnt.y - center_pnt->pnt.y;

  radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));

  offset_angle = atan2(shift_y,shift_x);

  clock_wise = ((gint)obj->type_data)/(abs((gint)(obj->type_data)));

  if(offset_angle < 0)
    offset_angle += 2* 3.14159265358979323846;

  sp_cons = radius/((gint)obj->type_data * 2 * 3.14159265358979323846 + offset_angle);

  ang_grid = 2.0* 3.14159265358979323846 /(gdouble)180;


  for(loop = 0 ; loop <= abs((gint)(obj->type_data)*180) + clock_wise*(gint)rint(offset_angle/ang_grid) ; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid;

      lx = sp_cons * ang_loop * cos(ang_loop)*clock_wise;
      ly = sp_cons * ang_loop * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);

      if(do_line)
	{


	  if(calc_pnt.x == start_pnt.x && calc_pnt.y == start_pnt.y)
	    continue;

	  if(drawing_pic)
	    {
	      gdk_draw_line(pic_preview->window,
			    pic_preview->style->black_gc,
			    adjust_pic_coords(calc_pnt.x,
					      preview_width),
			    adjust_pic_coords(calc_pnt.y,
					      preview_height),
			    adjust_pic_coords(start_pnt.x,
					      preview_width),
			    adjust_pic_coords(start_pnt.y,
					      preview_height));
	    }
	  else
	    {
	      gdk_draw_line(gfig_preview->window,
			    gfig_gc,
			    gfig_scale_x(calc_pnt.x),
			    gfig_scale_y(calc_pnt.y),
			    gfig_scale_x(start_pnt.x),
			    gfig_scale_y(start_pnt.y));
	    }
	}
      else
	{
	  do_line = 1;
	  first_pnt.x = calc_pnt.x;
	  first_pnt.y = calc_pnt.y;
	}
      start_pnt.x = calc_pnt.x;
      start_pnt.y = calc_pnt.y;
    }
}

static void
PREFIXd_paint_spiral(DOBJECT *obj)
{


  gdouble *line_pnts;
  GParam *return_vals = 0;
  gint nreturn_vals;
  gint seg_count = 0;
  gint i = 0;
  DOBJPOINTS * center_pnt;
  DOBJPOINTS * radius_pnt;
  gint16 shift_x;
  gint16 shift_y;
  gdouble ang_grid;
  gdouble ang_loop;
  gdouble radius;
  gdouble offset_angle;
  gdouble sp_cons;
  gint loop;
  GdkPoint last_pnt;
  gint clock_wise = 1;

 if (1){ if (!(obj != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 11546,"obj != NULL"); }else (void)0;

  center_pnt = obj->points;

  if(!center_pnt || !center_pnt->next)
    return;



  radius_pnt = center_pnt->next;


  shift_x = radius_pnt->pnt.x - center_pnt->pnt.x;
  shift_y = radius_pnt->pnt.y - center_pnt->pnt.y;

  radius = sqrt((shift_x*shift_x) + (shift_y*shift_y));

  clock_wise = ((gint)obj->type_data)/(abs((gint)(obj->type_data)));

  offset_angle = atan2(shift_y,shift_x);

  if(offset_angle < 0)
    offset_angle += 2* 3.14159265358979323846;

  sp_cons = radius/((gint)obj->type_data * 2 * 3.14159265358979323846 + offset_angle);

  ang_grid = 2.0* 3.14159265358979323846 /(gdouble)180;



  seg_count = abs((gint)(obj->type_data)*180) + clock_wise*(gint)rint(offset_angle/ang_grid);


  line_pnts = g_malloc0(1 *(2*seg_count + 3)*sizeof(gdouble));

  for(loop = 0 ; loop <= seg_count; loop++)
    {
      gdouble lx,ly;
      GdkPoint calc_pnt;

      ang_loop = (gdouble)loop * ang_grid;

      lx = sp_cons * ang_loop * cos(ang_loop)*clock_wise;
      ly = sp_cons * ang_loop * sin(ang_loop);

      calc_pnt.x = (gint)rint(lx + center_pnt->pnt.x);
      calc_pnt.y = (gint)rint(ly + center_pnt->pnt.y);


      if(!loop)
	{
	  if(calc_pnt.x == last_pnt.x && calc_pnt.y == last_pnt.y)
	    {
	      continue;
	    }
	}

      last_pnt.x = line_pnts[i++] = calc_pnt.x;
      last_pnt.y = line_pnts[i++] = calc_pnt.y;
    }


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.scaletoimage)
    scale_to_original_xy(&line_pnts[0],i/2);
  else
    scale_to_xy(&line_pnts[0],i/2);



  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,(i/2)*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {

      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,(i/2)*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);

    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(line_pnts);
}

DOBJECT *
PREFIXd_copy_spiral(DOBJECT * obj)
{
  DOBJECT *np;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == SPIRAL)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 11692,"obj->type == SPIRAL"); }else (void)0;

  np = d_new_spiral(obj->points->pnt.x,obj->points->pnt.y);

  np->points->next = d_copy_dobjpoints(obj->points->next);

  np->type_data = obj->type_data;




  return(np);
}

DOBJECT *
PREFIXd_new_spiral(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = SPIRAL;
  nobj->type_data = (gpointer)4;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_spiral;
  nobj->loadfunc  = d_load_spiral;
  nobj->savefunc  = d_save_spiral;
  nobj->paintfunc = d_paint_spiral;
  nobj->copyfunc  = d_copy_spiral;

  return(nobj);
}

void
PREFIXd_update_spiral(GdkPoint *pnt)
{
  DOBJPOINTS *center_pnt, *edge_pnt;
  gint saved_cnt_pnt = selvals.opts.showcontrol;


  center_pnt = obj_creating->points;

  if(!center_pnt)
    return;


#line 11753 "gfig.c"


#line 11758 "gfig.c"


  if((edge_pnt = center_pnt->next))
    {

      draw_circle(&edge_pnt->pnt);
      selvals.opts.showcontrol = 0;
      d_draw_spiral(obj_creating);

      edge_pnt->pnt.x = pnt->x;
      edge_pnt->pnt.y = pnt->y;
    }
  else
    {


      d_pnt_add_line(obj_creating,pnt->x,pnt->y,-1);
      edge_pnt = center_pnt->next;
    }


  selvals.opts.showcontrol = 0;
  d_draw_spiral(obj_creating);
  selvals.opts.showcontrol = saved_cnt_pnt;


  draw_circle(&edge_pnt->pnt);
}


#line 11790 "gfig.c"

void
PREFIXd_spiral_start(GdkPoint *pnt,gint shift_down)
{
  gint16 x,y;

  obj_creating = d_new_spiral(x = pnt->x, y = pnt->y);
  obj_creating->type_data = (gpointer)(spiral_num_turns*((spiral_toggle == 0)?1:-1));
}

void
PREFIXd_spiral_end(GdkPoint *pnt, gint shift_down)
{
  draw_circle(pnt);
  add_to_all_obj(current_obj,obj_creating);
  obj_creating = 0;
}



void
PREFIXd_save_bezier(DOBJECT * obj, FILE *to)
{
  DOBJPOINTS * spnt;

  spnt = obj->points;

  if(!spnt)
    return;

  fprintf(to,"<BEZIER>\n");

  while(spnt)
    {
      fprintf(to,"%d %d\n",
	      (gint)spnt->pnt.x,
	      (gint)spnt->pnt.y);
      spnt = spnt->next;
    }

  fprintf(to,"<EXTRA>\n");
  fprintf(to,"%d\n</EXTRA>\n",(gint)obj->type_data);
  fprintf(to,"</BEZIER>\n");

}



DOBJECT *
PREFIXd_load_bezier(FILE *from)
{
  DOBJECT *new_obj = 0;
  gint xpnt;
  gint ypnt;
  gchar buf[256];





  while(get_line(buf,256,from,0))
    {
      if(sscanf(buf,"%d %d",&xpnt,&ypnt) != 2)
	{

	  if(!strcmp("<EXTRA>",buf))
	    {
	      gint nsides = 3;

	      if(!new_obj)
		{
		  g_warning("[%d] Internal load error while loading bezier (extra area)",
			    line_no);
		  return(0);
		}
	      get_line(buf,256,from,0);
	      if(sscanf(buf,"%d",&nsides) != 1)
		{
		  g_warning("[%d] Internal load error while loading bezier (extra area scanf)",
			    line_no);
		  return(0);
		}
	      new_obj->type_data = (gpointer)nsides;
	      get_line(buf,256,from,0);
	      if(strcmp("</EXTRA>",buf))
		{
		  g_warning("[%d] Internal load error while loading bezier",
			    line_no);
		  return(0);
		}

	      continue;
	    }
	  else if(strcmp("</BEZIER>",buf))
	    {
	      g_warning("[%d] Internal load error while loading bezier",
			line_no);
	      return(0);
	    }
	  return(new_obj);
	}

      if(!new_obj)
	new_obj = d_new_bezier(xpnt,ypnt);
      else
	d_pnt_add_line(new_obj,xpnt,ypnt,-1);
    }
  return(new_obj);
}




static void
PREFIXfp_pnt_start()
{
  fp_pnt_cnt = 0;
}


static void
PREFIXfp_pnt_add(gdouble p1, gdouble p2, gdouble p3, gdouble p4)
{
  if(!fp_pnt_pnts)
    {
      fp_pnt_pnts = g_malloc0(10 *sizeof(gdouble));
      fp_pnt_chunk = 1;
    }

  if(((fp_pnt_cnt+4)/ 10) >= fp_pnt_chunk)
    {

      fp_pnt_chunk++;
      fp_pnt_pnts = (gdouble *)g_realloc(fp_pnt_pnts,sizeof(gdouble)*fp_pnt_chunk* 10);
    }

  fp_pnt_pnts[fp_pnt_cnt++] = p1;
  fp_pnt_pnts[fp_pnt_cnt++] = p2;
  fp_pnt_pnts[fp_pnt_cnt++] = p3;
  fp_pnt_pnts[fp_pnt_cnt++] = p4;
}

static gdouble *
PREFIXd_bz_get_array(gint *sz)
{
  *sz = fp_pnt_cnt;
  return (fp_pnt_pnts);
}


static void
PREFIXd_bz_line()
{
  gint i,x0,y0,x1,y1;

 if (1){ if (!((fp_pnt_cnt%4) == 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 11950,"(fp_pnt_cnt%4) == 0"); }else (void)0;

  for(i = 0 ; i < fp_pnt_cnt; i+=4)
    {
      x0 = (gint)fp_pnt_pnts[i];
      y0 = (gint)fp_pnt_pnts[i+1];
      x1 = (gint)fp_pnt_pnts[i+2];
      y1 = (gint)fp_pnt_pnts[i+3];

      if(drawing_pic)
	{
	  gdk_draw_line(pic_preview->window,
			pic_preview->style->black_gc,
			adjust_pic_coords((gint)x0,
					  preview_width),
			adjust_pic_coords((gint)y0,
					  preview_height),
			adjust_pic_coords((gint)x1,
					  preview_width),
			adjust_pic_coords((gint)y1,
					  preview_height));
	}
      else
	{
	  gdk_draw_line(gfig_preview->window,
			gfig_gc,
			gfig_scale_x((gint)x0),
			gfig_scale_y((gint)y0),
			gfig_scale_x((gint)x1),
			gfig_scale_y((gint)y1));
	}
    }
}



void
PREFIXDrawBezier (gdouble (*points)[2], gint np, gdouble mid, gint depth)
{
  gint i,j,x0=0,y0=0,x1,y1;
  fp_pnt left;
  fp_pnt right;

    if (depth==0)
      {
	for (i=0; i<np; i++)
	  {
	    x1=(int) points[i][0];
	    y1=(int) points[i][1];
	    if(i > 0 && (x1 != x0 || y1 != y0))
	      {

		fp_pnt_add((gdouble)x0,(gdouble)y0,(gdouble)x1,(gdouble)y1);
	      }
	    x0=x1;
	    y0=y1;
	  }
      }
    else
      {
	left = (fp_pnt)((gdouble *) g_malloc ((unsigned) sizeof (gdouble) * (np*2)));
	right = (fp_pnt)((gdouble *) g_malloc ((unsigned) sizeof (gdouble) * (np*2)));
	for (i=0; i<np; i++)
	  {
	    right[i][0]=points[i][0];
	    right[i][1]=points[i][1];
	  }
	left[0][0]=right[0][0];
	left[0][1]=right[0][1];
	for (j=np-1; j>=1; j--)
	  {
	    for (i=0; i<j; i++)
	      {
		right[i][0]=(1-mid)*right[i][0]+mid*right[i+1][0];
		right[i][1]=(1-mid)*right[i][1]+mid*right[i+1][1];
	      }
	    left[np-j][0]=right[0][0];
	    left[np-j][1]=right[0][1];
	  }
	if (depth>0)
	  {
	    DrawBezier(left,np,mid,depth-1);
	    DrawBezier(right,np,mid,depth-1);
	    g_free(left);
	    g_free(right);
	  }
      }
}


static void
PREFIXd_draw_bezier(DOBJECT *obj)
{
  DOBJPOINTS * spnt;
  gint seg_count = 0;
  gint i = 0;
  gdouble (*line_pnts)[2];

  spnt = obj->points;




  while(spnt)
    {
      seg_count++;
      spnt = spnt->next;
    }

  spnt = obj->points;

  if(!spnt || !seg_count)
    return;


  line_pnts = (fp_pnt)g_malloc0(1 *(2*seg_count + 1)*sizeof(gdouble));


  while(spnt)
    {
      draw_sqr(&spnt->pnt);
      line_pnts[i][0] = spnt->pnt.x;
      line_pnts[i++][1] = spnt->pnt.y;
      spnt = spnt->next;
    }



  if(!drawing_pic && bezier_line_frame && tmp_bezier)
    {
      fp_pnt_start();
      DrawBezier(line_pnts,seg_count,0.5,0);
      d_bz_line();
    }

  fp_pnt_start();
  DrawBezier(line_pnts,seg_count,0.5,3);
  d_bz_line();


  g_free(line_pnts);
}

static void
PREFIXd_paint_bezier(DOBJECT *obj)
{
  gdouble *line_pnts;
  gdouble (*bz_line_pnts)[2];
  DOBJPOINTS * spnt;
  gint seg_count = 0;

  GParam *return_vals = 0;
  gint nreturn_vals;
  gint i=0;

  spnt = obj->points;




  while(spnt)
    {
      seg_count++;
      spnt = spnt->next;
    }

  spnt = obj->points;

  if(!spnt || !seg_count)
    return;


  bz_line_pnts = (fp_pnt)g_malloc0(1 *(2*seg_count + 1)*sizeof(gdouble));


  while(spnt)
    {
      bz_line_pnts[i][0] = spnt->pnt.x;
      bz_line_pnts[i++][1] = spnt->pnt.y;
      spnt = spnt->next;
    }

  fp_pnt_start();
  DrawBezier(bz_line_pnts,seg_count,0.5,5);
  line_pnts = d_bz_get_array(&i);


  if(selvals.reverselines)
    reverse_pairs_list(&line_pnts[0],i/2);


  if(selvals.scaletoimage)
    scale_to_original_xy(&line_pnts[0],i/2);
  else
    scale_to_xy(&line_pnts[0],i/2);

  if(selvals.painttype == PAINT_BRUSH_TYPE)
    {
      switch(selvals.brshtype)
	{
	case BRUSH_BRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_paintbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.brushfade,
					    PARAM_INT32,(i/2)*2* 1,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PENCIL_TYPE:
	  return_vals = gimp_run_procedure ("gimp_pencil", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_AIRBRUSH_TYPE:
	  return_vals = gimp_run_procedure ("gimp_airbrush", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_FLOAT,(gdouble)selvals.airbrushpressure,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	case BRUSH_PATTERN_TYPE:
	  return_vals = gimp_run_procedure ("gimp_clone", &nreturn_vals,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_DRAWABLE, gfig_drawable,
					    PARAM_INT32, 1,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_FLOAT,(gdouble)0.0,
					    PARAM_INT32,(i/2)*2,
					    PARAM_FLOATARRAY, &line_pnts[0],
					    PARAM_END);
	  break;
	default:
	  break;
	}
    }
  else
    {

      return_vals = gimp_run_procedure ("gimp_free_select", &nreturn_vals,
					PARAM_IMAGE, gfig_image,
					PARAM_INT32,(i/2)*2,
					PARAM_FLOATARRAY, &line_pnts[0],
					PARAM_INT32,selopt.type,
					PARAM_INT32,selopt.antia,
					PARAM_INT32,selopt.feather,
					PARAM_FLOAT,(gdouble)selopt.feather_radius,
					PARAM_END);

    }

  gimp_destroy_params (return_vals, nreturn_vals);

  g_free(bz_line_pnts);

}

DOBJECT *
PREFIXd_copy_bezier(DOBJECT * obj)
{
  DOBJECT *np;




  if(!obj)
    return(0);

 if (1){ if (!(obj->type == BEZIER)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 12222,"obj->type == BEZIER"); }else (void)0;

  np = d_new_bezier(obj->points->pnt.x,obj->points->pnt.y);

  np->points->next = d_copy_dobjpoints(obj->points->next);

  np->type_data = obj->type_data;




  return(np);
}

DOBJECT *
PREFIXd_new_bezier(gint x, gint y)
{
  DOBJECT *nobj;
  DOBJPOINTS *npnt;




  npnt = (DOBJPOINTS *)g_malloc0(sizeof(DOBJPOINTS));




  npnt->pnt.x = x;
  npnt->pnt.y = y;

  nobj = (DOBJECT *)g_malloc0(sizeof(DOBJECT));

  nobj->type = BEZIER;
  nobj->type_data = (gpointer)4;
  nobj->points = npnt;
  nobj->drawfunc  = d_draw_bezier;
  nobj->loadfunc  = d_load_bezier;
  nobj->savefunc  = d_save_bezier;
  nobj->paintfunc = d_paint_bezier;
  nobj->copyfunc  = d_copy_bezier;

  return(nobj);
}

void
PREFIXd_update_bezier(GdkPoint *pnt)
{
  DOBJPOINTS *s_pnt, *l_pnt;
  gint saved_cnt_pnt = selvals.opts.showcontrol;

 if (1){ if (!(tmp_bezier != 0)) g_log (((gchar*) 0), G_LOG_LEVEL_ERROR, "file %s: line %d: assertion failed: (%s)", "gfig.c", 12273,"tmp_bezier != NULL"); }else (void)0;


  s_pnt = tmp_bezier->points;

  if(!s_pnt)
    return;


#line 12283 "gfig.c"

  if((l_pnt = s_pnt->next))
    {

      while(l_pnt->next)
	{
	  l_pnt = l_pnt->next;
	}

      draw_circle(&l_pnt->pnt);
      selvals.opts.showcontrol = 0;
      d_draw_bezier(tmp_bezier);
      l_pnt->pnt.x = pnt->x;
      l_pnt->pnt.y = pnt->y;
    }
  else
    {


      d_pnt_add_line(tmp_bezier,pnt->x,pnt->y,-1);
      l_pnt = s_pnt->next;
    }


  selvals.opts.showcontrol = 0;
  d_draw_bezier(tmp_bezier);
  selvals.opts.showcontrol = saved_cnt_pnt;


  draw_circle(&l_pnt->pnt);
}


#line 12318 "gfig.c"

void
PREFIXd_bezier_start(GdkPoint *pnt,gint shift_down)
{
  gint16 x,y;

  if(!tmp_bezier)
    {

      tmp_bezier = obj_creating = d_new_bezier(x = pnt->x, y = pnt->y);
    }
}

void
PREFIXd_bezier_end(GdkPoint *pnt, gint shift_down)
{
  DOBJPOINTS *l_pnt;

  if(!tmp_bezier)
    {
      tmp_bezier = obj_creating;
    }

  l_pnt = tmp_bezier->points->next;

  if(!l_pnt)
    return;

  if(shift_down)
    {

      while(l_pnt->next)
	{
	  l_pnt = l_pnt->next;
	}

      if(l_pnt)
	{
	  draw_circle(&l_pnt->pnt);
	  draw_sqr(&l_pnt->pnt);

	  if(bezier_closed)
	    {
	      gint tmp_frame = bezier_line_frame;

	      d_draw_bezier(tmp_bezier);
	      d_pnt_add_line(tmp_bezier,
			     tmp_bezier->points->pnt.x,
			     tmp_bezier->points->pnt.y,-1);

	      bezier_line_frame = 0;
	      d_draw_bezier(tmp_bezier);
	      bezier_line_frame = tmp_frame;
	    }
	  else if(bezier_line_frame)
	    {
	      d_draw_bezier(tmp_bezier);
	      bezier_line_frame = 0;
	      d_draw_bezier(tmp_bezier);
	      bezier_line_frame = 1;
	    }

	  add_to_all_obj(current_obj,obj_creating);
	}


      tmp_bezier = 0;
      obj_creating = 0;
    }
  else
    {
      if(!tmp_bezier->points->next)
	{
	  draw_circle(&tmp_bezier->points->pnt);
	  draw_sqr(&tmp_bezier->points->pnt);
	}

      d_draw_bezier(tmp_bezier);
      d_pnt_add_line(tmp_bezier,pnt->x,pnt->y,-1);
      d_draw_bezier(tmp_bezier);
    }
}



DALLOBJS *
PREFIXcopy_all_objs(DALLOBJS *objs)
{
  DALLOBJS * nobj;
  DALLOBJS * new_all_objs = 0;
  DALLOBJS * ret = 0;

  while(objs)
    {
      nobj = g_malloc0(sizeof(DALLOBJS));

     if(!ret)
	{
	  ret = new_all_objs = nobj;
	}
      else
	{
	  new_all_objs->next = nobj;
	  new_all_objs = nobj;
	}

      nobj->obj = (DOBJECT *)objs->obj->copyfunc(objs->obj);

      objs = objs->next;
    }

  return(ret);
}


void
PREFIXdraw_one_obj(DOBJECT * obj)
{
  obj->drawfunc(obj);
}

static void
PREFIXdraw_objects(DALLOBJS * objs,gint show_single)
{

#line 12445 "gfig.c"

  gint count = 0;

  while(objs)
    {
      if(!show_single || count == obj_show_single || obj_show_single == -1)
	draw_one_obj(objs->obj);
      objs = objs->next;
      count++;
    }
}

static void
PREFIXprepend_to_all_obj(GFIGOBJ *fobj,DALLOBJS *nobj)
{
  DALLOBJS *cobj;

  setup_undo();

  if(!fobj->obj_list)
    {
      fobj->obj_list = nobj;
      return;
    }

  cobj = fobj->obj_list;

  while(cobj->next)
    {
      cobj = cobj->next;
    }

  cobj->next = nobj;
}

static void
PREFIXadd_to_all_obj(GFIGOBJ * fobj,DOBJECT *obj)
{
  DALLOBJS *nobj;

  nobj = g_malloc0(sizeof(DALLOBJS));

  nobj->obj = obj;

  if(need_to_scale)
    scale_obj_points(obj->points,scale_x_factor,scale_y_factor);

  prepend_to_all_obj(fobj,nobj);
}

void
PREFIXobject_operation_start(GdkPoint *pnt,gint shift_down)
{
  DOBJECT *new_obj;


  operation_obj = get_nearest_objs(current_obj,pnt);



  if(shift_down && selvals.otype == MOVE_OBJ)
    {
      move_all_pnt = g_malloc0(sizeof(*move_all_pnt));
      *move_all_pnt = *pnt;
      setup_undo();
      return;
    }

  if(!operation_obj)
    return;


  setup_undo();

  switch(selvals.otype)
    {
    case MOVE_OBJ:
      if(operation_obj->type == BEZIER)
	{
	  d_draw_bezier(operation_obj);
	  tmp_bezier = operation_obj;
	  d_draw_bezier(operation_obj);
	}
      break;
    case MOVE_POINT:
      if(operation_obj->type == BEZIER)
	{
	  d_draw_bezier(operation_obj);
	  tmp_bezier = operation_obj;
	  d_draw_bezier(operation_obj);
	}

      if((operation_obj->type == POLY
	  || operation_obj->type == STAR)
	 && shift_down)
	{
	  switch(operation_obj->type)
	    {
	    case POLY:
	      d_poly2lines(operation_obj);
	      break;
	    case STAR:
	      d_star2lines(operation_obj);
	      break;
	    default:
	      break;
	    }

	  scan_obj_points(operation_obj->points,pnt);
	}
      break;
    case COPY_OBJ:





      new_obj = (DOBJECT *)operation_obj->copyfunc(operation_obj);
      if(new_obj)
	{
	  scan_obj_points(new_obj->points,pnt);
	  add_to_all_obj(current_obj,new_obj);
	  operation_obj = new_obj;
	  selvals.otype = MOVE_COPY_OBJ;
	  new_obj->drawfunc(new_obj);
	}
      break;
    case DEL_OBJ:
      remove_obj_from_list(current_obj,operation_obj);
      break;
    case MOVE_COPY_OBJ:
    default:
      g_warning("Internal error selvals.otype object operation start");
      break;
    }
}

void
PREFIXobject_operation_end(GdkPoint *pnt,gint shift_down)
{
  if(selvals.otype != DEL_OBJ && operation_obj && operation_obj->type == BEZIER)
    {
      d_draw_bezier(operation_obj);
      tmp_bezier = 0;
      d_draw_bezier(operation_obj);
    }

  operation_obj = 0;

  if(move_all_pnt)
    {
      g_free(move_all_pnt);
      move_all_pnt = 0;
    }


  if(selvals.otype == MOVE_COPY_OBJ)
    selvals.otype = COPY_OBJ;


}


void
PREFIXobject_operation(GdkPoint *to_pnt,gint shift_down)
{




  if(move_all_pnt && shift_down && selvals.otype == MOVE_OBJ)
    {
      do_move_all_obj(to_pnt);
      return;
    }

  if(!operation_obj)
    return;

  switch(selvals.otype)
    {
    case MOVE_OBJ:
    case MOVE_COPY_OBJ:
      switch(operation_obj->type)
	{
	case LINE:
	case CIRCLE:
	case ELLIPSE:
	case POLY:
	case ARC:
	case STAR:
	case SPIRAL:
	case BEZIER:
	  do_move_obj(operation_obj,to_pnt);
	  break;
	default:

	  g_warning("Internal error in operation_obj->type");
	  break;
	}
      break;
    case MOVE_POINT:
      switch(operation_obj->type)
	{
	case LINE:
	case CIRCLE:
	case ELLIPSE:
	case POLY:
	case ARC:
	case STAR:
	case SPIRAL:
	case BEZIER:
	  do_move_obj_pnt(operation_obj,to_pnt);
	  break;
	default:

	  g_warning("Internal error in operation_obj->type");
	  break;
	}
      break;
    case DEL_OBJ:
      break;
    case COPY_OBJ:
    default:
      g_warning("Internal error selvals.otype");
      break;
    }
}


void
PREFIXobject_start(GdkPoint *pnt,gint shift_down)
{

  if(!selvals.scaletoimage)
    {
      need_to_scale = 1;
      selvals.scaletoimage = 1;
    }
  else
    {
      need_to_scale = 0;
    }

  switch(selvals.otype)
    {
    case LINE:

      if(!shift_down || !obj_creating)
	draw_sqr(pnt);
      d_line_start(pnt,shift_down);
      break;
    case CIRCLE:
      draw_sqr(pnt);
      d_circle_start(pnt,shift_down);
      break;
    case ELLIPSE:
      draw_sqr(pnt);
      d_ellipse_start(pnt,shift_down);
      break;
    case POLY:
      draw_sqr(pnt);
      d_poly_start(pnt,shift_down);
      break;
    case ARC:
      d_arc_start(pnt,shift_down);
      break;
    case STAR:
      draw_sqr(pnt);
      d_star_start(pnt,shift_down);
      break;
    case SPIRAL:
      draw_sqr(pnt);
      d_spiral_start(pnt,shift_down);
      break;
    case BEZIER:
      if(!tmp_bezier)
	draw_sqr(pnt);
      d_bezier_start(pnt,shift_down);
      break;
    default:

      break;
    }
}


void
PREFIXobject_end(GdkPoint *pnt,gint shift_down)
{




  switch(selvals.otype)
    {
    case LINE:
      d_line_end(pnt,shift_down);
      draw_sqr(pnt);
      break;
    case CIRCLE:
      draw_sqr(pnt);
      d_circle_end(pnt,shift_down);
      break;
    case ELLIPSE:
      draw_sqr(pnt);
      d_ellipse_end(pnt,shift_down);
      break;
    case POLY:
      draw_sqr(pnt);
      d_poly_end(pnt,shift_down);
      break;
    case STAR:
      draw_sqr(pnt);
      d_star_end(pnt,shift_down);
      break;
    case ARC:
      draw_sqr(pnt);
      d_arc_end(pnt,shift_down);
      break;
    case SPIRAL:
      draw_sqr(pnt);
      d_spiral_end(pnt,shift_down);
      break;
    case BEZIER:
      d_bezier_end(pnt,shift_down);
      break;
    default:

      break;
    }

  if(need_to_scale)
    {
      need_to_scale = 0;
      selvals.scaletoimage = 0;
    }
}

void
PREFIXobject_update(GdkPoint * pnt)
{


  switch(selvals.otype)
    {
    case LINE:
      d_update_line(pnt);
      break;
    case CIRCLE:
      d_update_circle(pnt);
      break;
    case ELLIPSE:
      d_update_ellipse(pnt);
      break;
    case POLY:
      d_update_poly(pnt);
      break;
    case STAR:
      d_update_star(pnt);
      break;
    case ARC:
      d_update_arc(pnt);
      break;
    case SPIRAL:
      d_update_spiral(pnt);
      break;
    case BEZIER:
      d_update_bezier(pnt);
      break;
    default:

      break;
    }
}

